#include <iostream>
#include "POSCC1_online_cycle_detector.hpp"
#include "topological_sort.hpp"

#if __GNUC__ >= 3
#include <ext/functional>
#define SELECT1ST __gnu_cxx::select1st
#define PROJECT1ST __gnu_cxx::project1st
#else 
// everything else for the moment.
#include <functional>
#define SELECT1ST std::select1st
#define PROJECT1ST std::project1st
#endif

using namespace std;

POSCC1_online_cycle_detector::POSCC1_online_cycle_detector(graph2_t &g,
						     vector<variable_info> &v)
  : _graph(g), _varmap(v), _c2i(size(g)), _data(2*size(g),false) {
  // pre: graph is a DAG
  std::vector<unsigned int> order(size(g));
  // topologically sort graph.
  topological_sort(_graph,order);
  unsigned int index(0);
  // remember topological sort is produced
  // in reverse order.
  std::vector<unsigned int>::reverse_iterator iend(order.rend());
  for(std::vector<unsigned int>::reverse_iterator i(order.rbegin());
      i!=iend;++i,++index) {
    _c2i[*i]=index;
  }  
}

class POSCC1_oto_comp {
private:
  std::vector<unsigned int> const &_n2i;
public:
  POSCC1_oto_comp(std::vector<unsigned int> const &p)
    : _n2i(p) {
  }

  bool operator()(unsigned int x, unsigned int y) const {
    return _n2i[x] < _n2i[y];
  }
};

#define VISITED(x) _data[(x<<1)]
#define IN_COMPONENT(x) _data[(x<<1)+1]
#define C2I(x) _c2i[x]

void POSCC1_online_cycle_detector::new_edge(unsigned int tail, unsigned int head,
					 std::vector<unsigned int> &q) {

  unsigned int lb(_c2i[head]);
  unsigned int ub(_c2i[tail]);	

  if(lb < ub) {
    // in this case we are forced
    // to update the topological
    // order
    static std::vector<unsigned int> reachable,reaching;
    VISITED(tail)=true;
    IN_COMPONENT(tail) = true;
    // should be able to replace these with
    // generic call to depth_first_visit
    dfs_f(head,ub,reachable);

    // this is annoying! It is necessary because we must visit
    // nodes which are in a component (hence have already been visited)
    // in order to identify and unvisited predecessors. The algorithm
    // knows not to place the already visited nodes onto the reaching
    // vector because the IN_COMPONENT flag is set for them
    std::vector<unsigned int>::const_iterator iend(reachable.end());
    for(std::vector<unsigned int>::const_iterator i(reachable.begin());
	i!=iend;++i) {
      VISITED(*i) = false;
    }

    IN_COMPONENT(tail)=false; // do this to ensure that
                              // the tail node is available
                              // to represent the new component.
    
    dfs_b(tail,lb,reaching);
    POSCC1_oto_comp pc(_c2i);

    std::sort(reachable.begin(),reachable.end(),pc);
    std::sort(reaching.begin(),reaching.end(),pc);	  

    reassign(tail, reachable, reaching, q);

    // clear them because they are static.
    reachable.clear();
    reaching.clear();
  }
}

void POSCC1_online_cycle_detector::dfs_f(unsigned int n, unsigned int ub, std::vector<unsigned int> &reachable) {	   
  assert(n < size(_graph));
  VISITED(n)=true;    
  reachable.push_back(n);
  
  graph1_t::const_siterator i,iend;
  for(boost::tie(i,iend) = successors(n,_graph);i!=iend;++i) {
    assert(C2I(*i) >= C2I(n) || C2I(n) == ub);
    if(C2I(*i) <= ub) {
      if(!VISITED(*i)) { dfs_f(*i,ub,reachable); }
      IN_COMPONENT(n) = IN_COMPONENT(n) | IN_COMPONENT(*i);
    }
  }
}

void POSCC1_online_cycle_detector::dfs_b(unsigned int n, unsigned int lb, 
				      std::vector<unsigned int> &reaching) {
  assert(n < size(_graph));
  VISITED(n)=true;
  if(!IN_COMPONENT(n)) { reaching.push_back(n); }

  graph1_t::const_piterator i,iend;
  for(boost::tie(i,iend) = predecessors(n,_graph);i!=iend;++i) {
    assert(C2I(*i) <= C2I(n) || C2I(n) == lb);
    if(C2I(*i) > lb && !VISITED(*i)) { dfs_b(*i,lb,reaching); }
  }
}

void POSCC1_online_cycle_detector::reassign(unsigned int t,
	      std::vector<unsigned int> &reachable, 
	      std::vector<unsigned int> &reaching, 
	      std::vector<unsigned int> &q) {
  
  std::vector<unsigned int> tmp;
  tmp.reserve(reachable.size() + reaching.size());  
  
  std::vector<unsigned int>::iterator iend(reaching.end());
  for(std::vector<unsigned int>::iterator i(reaching.begin());
      i!=iend;++i) {
    unsigned int n(*i);
    tmp.push_back(n);
    VISITED(n) = false;
    *i = _c2i[*i];
  }

  iend = reachable.end();
  for(std::vector<unsigned int>::iterator i(reachable.begin());
      i!=iend;++i) {
    unsigned int n(*i);
    tmp.push_back(n);      
    if(IN_COMPONENT(n)) { 
      q.push_back(n);
      _varmap[n].node = t;
      IN_COMPONENT(n) = false;
    }
    VISITED(n) = false;
    *i = _c2i[*i];
  }
  
  std::vector<unsigned int>::iterator i(reachable.begin());
  std::vector<unsigned int>::iterator j(reaching.begin());
  std::vector<unsigned int>::iterator jend(reaching.end());

  unsigned int index(0);
  while(i != iend || j != jend) {
    unsigned int w;
    if(j == jend || (i != iend && *i < *j)) {
      w=*i;++i;
    } else {
      w=*j;++j;
    }
    unsigned int n(tmp[index++]);
    // allocate n at w
    C2I(n)=w;
  }
}









