#include <iostream>
#include "POSCC2_online_cycle_detector.hpp"
#include "topological_sort.hpp"

#if __GNUC__ >= 3
#include <ext/functional>
#define SELECT1ST __gnu_cxx::select1st
#define PROJECT1ST __gnu_cxx::project1st
#define IDENTITY __gnu_cxx::identity
#else 
// everything else for the moment.
#include <functional>
#define SELECT1ST std::select1st
#define PROJECT1ST std::project1st
#define IDENTITY std::identity
#endif

using namespace std;

POSCC2_online_cycle_detector::POSCC2_online_cycle_detector(graph2_t &g,
						     vector<variable_info> &v)
  : _graph(g), _varmap(v), _i2c(size(g)),
    _c2i(size(g)), _data(2*size(g),false), _vindex(size(g)) {
  // pre: graph is a DAG
  std::vector<unsigned int> order(size(g));
  // topologically sort graph.
  topological_sort(_graph,order);
  unsigned int index(0);
  // remember topological sort is produced
  // in reverse order.
  std::vector<unsigned int>::reverse_iterator iend(order.rend());
  for(std::vector<unsigned int>::reverse_iterator i(order.rbegin());
      i!=iend;++i,++index) {
    _c2i[*i]=index;
    _i2c[index]=*i;
  }  
}

#define VISITED(x) _data[(x<<1)]
#define IN_COMPONENT(x) _data[(x<<1)+1]
#define C2I(x) _c2i[x]
#define I2C(x) _i2c[x]

template<class T, class S = IDENTITY<T> >
class my_less {
private:
  S _s;
public:
  inline bool operator()(T const &a, T const &b) const {
    return _s(a) < _s(b);
  }
};

void POSCC2_online_cycle_detector::new_edge(unsigned int tail, unsigned int head,
					 std::vector<unsigned int> &q) {
  assert(head < size(_graph));assert(tail < size(_graph));

  unsigned int lb(C2I(head));
  unsigned int ub(C2I(tail));	

  if(lb < ub) {
    assert(ub < size(_graph));
    // in this case we are forced
    // to update the topological
    // order
    //
    // notice that i push the index of the tail
    // and not the tail itself!
    _batch.push_back(make_pair(ub,head));
  }
}

void POSCC2_online_cycle_detector::finish(std::vector<unsigned int> &q) {
  // Notes: q is the unifyqueue!

  // First, sort back edges into increasing order by their tail
  sort(_batch.begin(),_batch.end(),my_less<pair<unsigned int, unsigned int>, 
       SELECT1ST<pair<unsigned int, unsigned int> > >());

  std::vector<pair<unsigned int, unsigned int> > reachables;
  // initially set lb to have index larger than any. Then proceed by decreasing until
  unsigned int lb(_i2c.size()); 
  _index = 0;
  // now, identify reachables
  for(std::vector<pair<unsigned int, unsigned int> >::reverse_iterator i(_batch.rbegin());
      i!=_batch.rend();++i) {
    if(i->first < lb && !reachables.empty()) {       
      // in this case, the edges remaining to process
      // do not overlap with any that have been processed.
      shift(lb,reachables);
      reachables.clear();
    }

    if(!VISITED(i->second)) { 
      assert(i->second == _varmap[i->second].node);
      dfs_f(i->second,i->first,reachables,q); 
    }

    lb = std::min(lb,C2I(i->second));
  }
  
  if(!reachables.empty()) { shift(lb,reachables); }
  // finally, empty the batch of processed edges!
  // The following statement is part of an idea to reduce
  // the visit count of the overall algorithm ...
  _batch.clear();
}

void POSCC2_online_cycle_detector::dfs_f(unsigned int n, unsigned int ub, 
				       vector<pair<unsigned int, unsigned int> > &reachables,
				       vector<unsigned int> &q) {	   
  assert(n < size(_graph));
  assert(n == _varmap[n].node);
  VISITED(n)=true;    
  IN_COMPONENT(n)=false;
  _vindex[n]=_index++;
  
  graph1_t::const_siterator i,iend;
  for(boost::tie(i,iend) = successors(n,_graph);i!=iend;++i) {
    unsigned int w(*i);
    if(C2I(w) <= ub) {
      if(!VISITED(w)) { dfs_f(w,ub,reachables,q); }
      if(!IN_COMPONENT(w)) {
	w = _varmap[w].node;
	if(_vindex[w] < _vindex[_varmap[n].node]) {
	  _varmap[n].node = w;
	}
      }
    }
  }

  // at this point we must check to see
  // if any components have been identified
  if(_varmap[n].node == n) {
    unsigned int t(_vindex[n]);
    IN_COMPONENT(n) = true;
    while(!_stack.empty() && t < _vindex[_stack.back()]) {
      unsigned int w(_stack.back());
      _varmap[w].node = n;
      IN_COMPONENT(w) = true;
      q.push_back(w); // place w onto the unifyqueue
      _stack.pop_back(); 
    }
  } else {
    _stack.push_back(n);
  }

  // place this pair into the shift queue. Notice that
  // it is crucial to the algorithm that this is performed
  // in "topological order" and hence must be done after
  // all successors of n are visited.
  reachables.push_back(make_pair(ub,n));
}

void POSCC2_online_cycle_detector::shift(unsigned int index,
				       std::vector<pair<unsigned int, unsigned int> > &reachables) {

  std::vector<pair<unsigned int, unsigned int> >::reverse_iterator i(reachables.rbegin());
  std::vector<pair<unsigned int, unsigned int> >::reverse_iterator iend(reachables.rend());
  unsigned int lindex(index);

  while(i != iend) {    
    assert(index < _i2c.size());
    unsigned int w(I2C(index));

    if(VISITED(w)) {
      // reset the visited relation.
      VISITED(w)=false;
    } else {
      assert(lindex < size(_graph));
      I2C(lindex)=w;
      C2I(w)=lindex;
      ++lindex;
    }

    // check for an end, like.
    while(i != iend && i->first == index) {
      // ok, we're at an end point
      // so allocate now.
      assert(lindex < size(_graph));
      I2C(lindex)=i->second;
      C2I(i->second)=lindex;
      ++i;
      ++lindex;
    }      

    ++index;
  }
}









