// A fast solver of constraints containing indirections
// ====================================================
// Author David J. Pearce, Feb 2003
// 
// Notes:
//
// This algorithm is an implementation of the algorithm
// by Nevin Heintze and Olivier Tardieu which was presented 
// in their paper:
//
// "Ultra-fast Aliasing Analysis using CLA: A Million Lines
//  of C Code in a Second"
//
// This paper can be found online at www.cs.bell-labs.com/who/nch/
//
// I have reformulated the algorithm so that it now uses a
// cycle detection scheme which is based on Tarjan's algorithm
// for detecting scc's. 
//
// The use of the STACK instead of the path relation comes
// from the paper:
//
// "On Finding the Strongly Connected Components in a Directed
//  Graph", Esko Nuutila, Eljas Soisalon-Soininen, 1994

#ifndef ALGORITHM_HT_HPP
#define ALGORITHM_HT_HPP

#include <boost/tuple/tuple.hpp>
#include "topological_sort.hpp"
#include "priority_comp.hpp"
#include "filter_graph.hpp"
#include "vector_set.hpp"
#include "set_union.hpp"
#include "store.hpp"
#include "pcs.hpp"

template<class G, class S>
class algorithm_HT {
private:
  G &_graph;
  std::vector<variable_info> &_varmap;
  store<S> &_sol;
  store<S> *_cache;   // the "cache"
  bool _change; 
  
  // the final few vectors are used
  // by the algorithm for detecting
  // strongly connected components.
  std::vector<unsigned int> _vindex;   
  std::vector<bool> _in_component; 
  std::vector<unsigned int> _stack; 
  // should use vector of vector here for
  // efficiency
  std::vector<unsigned int> _unifyqueue;
  unsigned int _index; 

  // this final property is just used to obtain
  // useful information on the algorithms operation
  checker_stats &_stats;
public:
  algorithm_HT(G &g, 
	       std::vector<variable_info> &v,
	       store<S> &s, 
	       unsigned int d,
	       checker_stats &stats) :
    _graph(g),
    _varmap(v),
    _sol(s),
    _cache(s.clone()),
    _vindex(size(g)), 
    _in_component(size(g),true), 
    _stats(stats) {
    
    // empty the cache to start with!
    _cache->clear();
  }
  
  ~algorithm_HT(void) {
    delete _cache; // free memory
  }
  
  void solve(void) {
    do {
      _change = false;

      _stats.iter_count++;

      // this loop goes through each variable p which 
      // has a constraint of the form *p = y, y=*p
      // *p=&y   

      _index=0;
      for(unsigned int i=0;i!=_varmap.size();++i) {

	if(_varmap[i].node == i && !_varmap[i].complicates.empty()) {
	  S solution(_varmap.size());

	  if(!_cache->is_stored(i)) {
	    solution =  getLvals(i);
	  } else {	    
	    solution = (*_cache)[i];
	  }

	  // collapse any detected cycles
	  process_unifyqueue();

	  assert(_varmap[i].node == i);
	  
	  // ASSERT: i is always a component here. this is because
	  //         the search begins from i and, hence, any 
	  //         new compomnent involving i must be named i

	  // process complex constraints      
	  vector_set<constraint> complicates(_varmap[i].complicates);
	  for(vector_set<constraint>::const_iterator j(complicates.begin());
	      j!=complicates.end();++j) {      
	    if(!redundant_constraint(i,*j)) {
	      do_complex_constraint(*j,solution);	  
	    }
	  }
	}	  
      }
      // the following corresponds to flushing the cache
      // in the paper.
      
      _cache->clear(); // empty the cache completely          
      
    } while(_change);    
    
    // OK, condense graph
    std::vector<unsigned int> order;
    std::vector<bool> visited(size(_graph),false);
    _index=0;
    for(unsigned int i=0;i!=size(_graph);++i) {
      if(!visited[i] && _varmap[i].node == i) { visit(i,visited,order); }
    }
    process_unifyqueue();

    _stats.iter_count++;
    
    // at this point we can compute the entire solution using a 
    // topological traversal of the (acyclic)     

    for(std::vector<unsigned int>::reverse_iterator i(order.rbegin());
	i!=order.rend();++i) {
      if(_varmap[*i].node == *i) {
	_stats.visit_count++;
	// proceed through predecessors of i
	S sol(_sol[*i]);
	bool flag(false);
	typename G::const_piterator j,jend;
	for(boost::tie(j,jend) = predecessors(*i,_graph);j!=jend;++j) {
	  flag |= int_set_union(sol,_sol[*j]);
	}
	// write the final solution for *i
	if(flag) { _sol.update(*i,sol); }
      }
    }
  }

protected:

  /////////////////////////////////////////////////////////////////
  
  S getLvals(unsigned int n) {
    S sol(_sol[n]);

    _in_component[n] = false;
    _vindex[n]=_index++;
    _stats.visit_count++;

    // now process all predecessors
    typename G::const_piterator i,iend;
    for(boost::tie(i,iend) = predecessors(n,_graph);i!=iend;++i) {
      unsigned int w(*i);

      if(!_cache->is_stored(w)) { 
	if(_in_component[w]) {
	  int_set_union(sol,getLvals(w));
	}
      } else { int_set_union(sol,(*_cache)[w]); }

      if(!_in_component[w]) {
	unsigned int t(_varmap[w].node);
	if(_vindex[t] < _vindex[_varmap[n].node]) {
	  _varmap[n].node = t;
	}
      }      
    }

    // at this point we must check to see
    // if any components have been identified

    if(_varmap[n].node == n) {
      unsigned int t(_vindex[n]);
      _in_component[n] = true;
      while(!_stack.empty() && t < _vindex[_stack.back()]) {
       unsigned int w(_stack.back());
       _varmap[w].node = n;
       _in_component[w] = true;
       _unifyqueue.push_back(w);
       _stack.pop_back(); 
     }
      // write the new solution for n into the cache
      _cache->update(n,sol);
    } else {
      _cache->update(n,S(_varmap.size()));
      _stack.push_back(n);
    }   

    return sol;
  }

  void visit(unsigned int n, std::vector<bool> &visited,
	    std::vector<unsigned int> &order) {
    assert(_varmap[n].node == n);
    visited[n]=true;
    _in_component[n]=false;
    _vindex[n]=_index++;
   
    // visit all successors
    typename G::const_siterator i,iend;
    for(boost::tie(i,iend) = successors(n,_graph);i!=iend;++i) {
      // only traverse zero weighted edges
      unsigned int w(*i);
      if(!visited[w]) { visit(w,visited,order); }
      if(!_in_component[w]) {
	unsigned int t(_varmap[w].node);
	if(_vindex[t] < _vindex[_varmap[n].node]) {
	  _varmap[n].node = t;
	}
      }
    }
    
    // at this point we must check to see
    // if any components have been identified
    if(_varmap[n].node == n) {
      order.push_back(n);
      unsigned int t(_vindex[n]);
      _in_component[n] = true;
      while(!_stack.empty() && t < _vindex[_stack.back()]) {
	unsigned int w(_stack.back());
	_varmap[w].node = n;
	_in_component[w] = true;
	_unifyqueue.push_back(w);
	_stack.pop_back(); 
      }
    } else {
      _stack.push_back(n);
    }
  }

  void do_da_constraint(constraint const &c, 
			S const &delta) {
    // this function deals with constraints of the
    // form *x = &y
    unsigned int rhs(c.rhs().var());
    unsigned int off(c.lhs().offset());

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {
      if(type_safe(*j,off)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+off].node);
	S tmp(_sol[t]);
	
	if(tmp.insert(rhs).second) { 
	  _sol.update(t,tmp); 
	  _change |= _cache->is_stored(t);
	}
      }
    }
  }

  void do_ds_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form *x = y
    unsigned int rhs(_varmap[c.rhs().var()].node);
    unsigned int loff(c.lhs().offset());
    
    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {      
      if(type_safe(*j,loff)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+loff].node);
	if(t != rhs) {
	  if(add_edge(diedge(rhs,t),_graph)) {
	    _change=true;
	    _stats.new_edge_count++;
	  }
	}
      }
    }
  }

  void do_sd_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form x = *y
    unsigned int lhs(_varmap[c.lhs().var()].node);
    unsigned int roff(c.rhs().offset());

    for(typename S::const_iterator j(delta.begin()); j!=delta.end();++j) {
      if(type_safe(*j,roff)) {
	// *j != NULL && *j != UNKNOWN
	unsigned int t(_varmap[*j+roff].node);       
	if(t != lhs) {
	  if(add_edge(diedge(t,lhs),_graph)) {
	    _change=true;
	    _stats.new_edge_count++;
	  }
	}
      } 
    }
  }

  void do_complex_constraint(constraint c,
			     S const &delta) {
    _stats.complex_count++;

    if(c.lhs().dereferenced()) {
      if(c.rhs().addressof()) {
	// *x=&y
	do_da_constraint(c,delta);
      } else {
	// *x=y
	do_ds_constraint(c,delta);
      }
    } else {
      // x=*y
      do_sd_constraint(c,delta);
    }
  }

  void process_unifyqueue(void) {    
    for(std::vector<unsigned int>::const_iterator i(_unifyqueue.begin());
	i!=_unifyqueue.end();++i) {

      unsigned int np(*i);
      unsigned int n(_varmap[np].node);
      merge_simple(n,np,_graph);
      condense_nodes(n,np,_varmap);
      unifynodes(n,np);
      // prevent self loops
      erase_edge(diedge(n,n),_graph);
    }
    // Finally, clear the unifyqueue.
    _unifyqueue.clear();
  }

  void unifynodes(unsigned int n, unsigned int np) {
    S tmp(_sol[n]);    
    S const &snp(_sol[np]);

    if(int_set_union(tmp,snp)) { _sol.update(n,tmp); }    
    _sol.clear(np); // save memory
  }

  bool redundant_constraint(unsigned int n, constraint const &c) {	    
    if(c.rhs().dereferenced()) {	     
      assert(c.rhs().var() == n);
      unsigned int lhs(c.lhs().var());
      // a = *n
      if(_varmap[lhs].node != lhs) {
	constraint tmp(expression(SCALAR,_varmap[lhs].node,0),c.rhs());
	_varmap[n].complicates.erase(c);
	if(!_varmap[n].complicates.insert(tmp).second) {
	  return true;
	}
      }
    } else {
      assert(c.lhs().var() == n);
      if(!c.rhs().addressof()) {
	unsigned int rhs(c.rhs().var());
	if(_varmap[rhs].node != rhs) {
	  // *n = a
	  constraint tmp(c.lhs(),expression(SCALAR,_varmap[rhs].node,c.rhs().offset()));
	  _varmap[n].complicates.erase(c);
	  if(!_varmap[n].complicates.insert(tmp).second) {
	    return true;
	  }
	}
      }
    }
    return false;
  }

  bool type_safe(unsigned int n, unsigned int o) {
    return n > 1 && (n+o) < _varmap[n].end;
  }

  inline bool int_set_union(S &to, S const &from) {
    _stats.union_count++;
    _stats.union_size += to.size();
    _stats.union_size += from.size();
    return set_union(to,from);
  }
};

template<class G, class S>
void solve_HT(G &g, 
	      std::vector<variable_info> &v,
	      store<S> &s,
	      unsigned int d,
	      checker_stats &cs) {
  
  algorithm_HT<G,S> sc(g,v,s,d,cs);
  sc.solve();
}

#endif






















