// A fast solver of constraints containing indirections
// ====================================================
// Author David J. Pearce, Feb 2003
// 
// Notes:
//
// This algorithm is an implementation of the algorithm
// by Nevin Heintze and Olivier Tardieu which was presented 
// in their paper:
//
// "Ultra-fast Aliasing Analysis using CLA: A Million Lines
//  of C Code in a Second"
//
// This paper can be found online at www.cs.bell-labs.com/who/nch/
//
// I have reformulated the algorithm so that it now uses a
// cycle detection scheme which is based on Tarjan's algorithm
// for detecting scc's. 
//
// The use of the STACK instead of the path relation comes
// from the paper:
//
// "On Finding the Strongly Connected Components in a Directed
//  Graph", Esko Nuutila, Eljas Soisalon-Soininen, 1994

#ifndef ALGORITHM_HT_FDS_HPP
#define ALGORITHM_HT_FDS_HPP

#include <boost/tuple/tuple.hpp>
#include "topological_sort.hpp"
#include "priority_comp.hpp"
#include "filter_graph.hpp"
#include "vector_set.hpp"
#include "set_union.hpp"
#include "store.hpp"
#include "pcs.hpp"

template<class G, class S>
class algorithm_HT_FDS {
private:
  G &_graph;
  std::vector<variable_info> &_varmap;
  store<S> &_sol;
  store<S> *_cache;   // the "cache"
  bool _change; 

  // the final few vectors are used
  // by the algorithm for detecting
  // strongly connected components.
  std::vector<unsigned int> _vindex;   
  std::vector<bool> _in_component; 
  std::vector<unsigned int> _stack; 
  std::vector<bool> _visited;
  // should use vector of vector here for
  // efficiency
  std::vector<unsigned int> _unifyqueue;
  unsigned int _index; 
  // the followng two are only used by the worklist solver
  unsigned int _changed_count;
  std::vector<bool> _changed;

  // this final property is just used to obtain
  // useful information on the algorithms operation
  checker_stats &_stats;
public:
  algorithm_HT_FDS(G &g, 
		   std::vector<variable_info> &v,
		   store<S> &s, 
		   unsigned int d,
		   checker_stats &stats) :
    _graph(g),
    _varmap(v),
    _sol(s),
    _cache(s.clone()),
    _vindex(size(g)), 
    _in_component(size(g),true), 
    _visited(size(g),false),
    _changed_count(size(g)),
    _changed(size(g),true),    
    _stats(stats){    
    // empty the cache to start with!
    _cache->clear();
    // make sure the changed count is correct.
    for(unsigned int i=0;i!=_varmap.size();++i) {
      if(_varmap[i].node != i) {
	_changed[i]=false;
	_changed_count--;	    	
      }            
    }    
  }
  
  ~algorithm_HT_FDS(void) {
    delete _cache; // free memory
  }
  
  void solve(void) {
    do {
      _change = false;

      _stats.iter_count++;

      // this loop goes through each variable p which 
      // has a constraint of the form *p = y, y=*p
      // *p=&y   
      
      _index=0;
      for(unsigned int i=0;i!=_varmap.size();++i) {

	if(_varmap[i].node == i && !_varmap[i].complicates.empty()) {
	  S solution(_varmap.size());

	  if(!_cache->is_stored(i)) {
	    scc_visit(i);
	    // collapse any detected cycles
	    process_unifyqueue();

	    solution =  getLvals(i);
	  } else {	    
	    solution = (*_cache)[i];
	  }

	  assert(_varmap[i].node == i);
	  
	  // ASSERT: i is always a component here. this is because
	  //         the search begins from i and, hence, any 
	  //         new compomnent involving i must be named i

	  vector_set<constraint> complicates(_varmap[i].complicates);
	  for(vector_set<constraint>::const_iterator j(complicates.begin());
	      j!=complicates.end();++j) {      
	    if(!redundant_constraint(i,*j)) {
	      do_complex_constraint(*j,solution);	  
	    }
	  }
	}
      }

      // the following corresponds to flushing the cache
      // in the paper.
      
      _cache->clear(); // empty the cache completely          

      // clear the visited relation
      fill(_visited.begin(),_visited.end(),false);

    } while(_change);    

    // at this point we can compute the entire solution using a 
    // worklist style iteration of the graph.
    // the graph may not be acyclic because of positive weight
    // cycles

    // must prevent nodes which  from
    // being visited. Otherwise they will be
    // regarded as components and remarked
    // such that varmap[x].node = x.

    // first step compute the strongly connected components
    // of the filtered graph.
    std::vector<unsigned int> order;    
    // this stage basically corresponds to a worklist style algorithm    
    do {
      _stats.iter_count++;
      _stats.post_iter_count++;

      _index = 0;
      _change = false;
      for(unsigned int i=0;i!=_varmap.size();++i) {
	if(!_visited[i] && _varmap[i].node == i) { scc_visit(i); }
      }
      // collapse any cycles
      process_unifyqueue();
      // generate topological order
      order.clear();
      for(unsigned int i=0;i!=_varmap.size();++i) {
	if(_visited[i] && _varmap[i].node == i) { topo_visit(i,order); }
      }

      for(std::vector<unsigned int>::iterator i(order.begin());i!=order.end();++i) {	
	_stats.visit_count++;
	if(_changed[*i]) {
	  _changed[*i]=false;
	  assert(_changed_count > 0);
	  _changed_count--;
	  S solution(_sol[*i]);
	  
	  // process complex constraints
	  vector_set<constraint>::const_iterator kend(_varmap[*i].complicates.end());
	  for(vector_set<constraint>::const_iterator k(_varmap[*i].complicates.begin());
	      k!=kend;++k) {      
	    do_complex_constraint2(*k,solution);
	  }
	  
	  // now propagate solution to all successors
	  assert(_varmap[*i].node == *i);
	  typename G::const_siterator j,jend;
	  for(boost::tie(j,jend) = successors(*i,_graph);j!=jend;++j) {	
	    S tmp(_sol[*j]);
	    bool flag(false);
	    // process weighted edges
	    vector_set<unsigned int> &labels(*(j.base()->second()));
	    for(vector_set<unsigned int>::const_iterator k(labels.begin());
		k!=labels.end();++k) {
	      flag |= int_set_union(tmp,solution,*k);
	    }
	    if(flag) {
	      _sol.update(*j,tmp);
	      if(!_changed[*j]) { _changed[*j]=true;_changed_count++; }
	    }
	  }
	}
      }     
    } while(_changed_count > 0);
  }

protected:

  /////////////////////////////////////////////////////////////////
  void scc_visit(unsigned int n) {
    assert(_varmap[n].node == n);
    _visited[n]=true;
    _in_component[n]=false;
    _vindex[n]=_index++;
   
    // visit all successors
    typename G::const_piterator i,iend;
    for(boost::tie(i,iend) = predecessors(n,_graph);i!=iend;++i) {
      // only traverse zero weighted edges
      if(*(i.base()->second()->begin()) == 0) {
	unsigned int w(*i);
	if(!_visited[w]) { scc_visit(w); }
	if(!_in_component[w]) {
	  unsigned int t(_varmap[w].node);
	  if(_vindex[t] < _vindex[_varmap[n].node]) {
	    _varmap[n].node = t;
	  }
	}
      }
    }
    
    // at this point we must check to see
    // if any components have been identified
    if(_varmap[n].node == n) {
      unsigned int t(_vindex[n]);
      _in_component[n] = true;
      while(!_stack.empty() && t < _vindex[_stack.back()]) {
	unsigned int w(_stack.back());
	_varmap[w].node = n;
	_in_component[w] = true;
	_unifyqueue.push_back(w);
	_stack.pop_back(); 
      }
    } else {
      _stack.push_back(n);
    }
  }

  void topo_visit(unsigned int n, 
		  std::vector<unsigned int> &order) {
    assert(_varmap[n].node == n);
    _visited[n]=false;
   
    // visit all successors
    typename G::const_piterator i,iend;
    for(boost::tie(i,iend) = predecessors(n,_graph);i!=iend;++i) {
      // only traverse zero weighted edges
      unsigned int w(*i);
      if(_visited[w]) { topo_visit(w,order); }
    }

    order.push_back(n);
  }

  S getLvals(unsigned int n) {
    S sol(_sol[n]);
    _in_component[n]=false;

    // now process all predecessors
    typename G::const_piterator i,iend;
    for(boost::tie(i,iend) = predecessors(n,_graph);i!=iend;++i) {
      unsigned int w(*i);
      S tmp(_varmap.size());
      if(!_cache->is_stored(w)) {	  
	if(_in_component[w]) {
	  tmp = getLvals(w);
	} else {
	  tmp = _sol[w];
	}
      } else { tmp = (*_cache)[w]; }
      
      vector_set<unsigned int> &labels(*(i.base()->second()));
      for(vector_set<unsigned int>::iterator j(labels.begin());
	  j!=labels.end();++j) {
	int_set_union(sol,tmp,*j);
      }      
    }

    _cache->update(n,sol);
    _in_component[n]=true;

    return sol;
  }

  void do_da_constraint(constraint const &c, 
			S const &delta) {
    // this function deals with constraints of the
    // form *x = &y
    unsigned int rhs(c.rhs().var());
    unsigned int off(c.lhs().offset());

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {
      if(type_safe(*j,off)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+off].node);
	S tmp(_sol[t]);
	
	if(tmp.insert(rhs).second) { 
	  _sol.update(t,tmp); 
	  _change |= _cache->is_stored(t);
	}
      }
    }
  }

  void do_ds_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form *x = y
    unsigned int rhs(_varmap[c.rhs().var()].node);
    unsigned int loff(c.lhs().offset());
    unsigned int roff(c.rhs().offset());

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {      
      if(type_safe(*j,loff)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+loff].node);
	_change |= int_add_edge(rhs,t,roff);
      }
    }
  }

  void do_sd_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form x = *y
    unsigned int lhs(_varmap[c.lhs().var()].node);
    unsigned int roff(c.rhs().offset());

    for(typename S::const_iterator j(delta.begin()); j!=delta.end();++j) {
      if(type_safe(*j,roff)) {
	// *j != NULL && *j != UNKNOWN
	unsigned int t(_varmap[*j+roff].node);       
	_change |= int_add_edge(t,lhs,0);
      } 
    }
  }

  void do_complex_constraint(constraint c,
			     S const &delta) {
    _stats.complex_count++;

    if(c.lhs().dereferenced()) {
      if(c.rhs().addressof()) {
	// *x=&y
	do_da_constraint(c,delta);
      } else {
	// *x=y
	do_ds_constraint(c,delta);
      }
    } else {
      // x=*y
      do_sd_constraint(c,delta);
    }
  }

  void do_da_constraint2(constraint const &c, 
			S const &delta) {
    // this function deals with constraints of the
    // form *x = &y
    unsigned int rhs(c.rhs().var());
    unsigned int off(c.lhs().offset());

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {
      if(type_safe(*j,off)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+off].node);
	S tmp(_sol[t]);
	if(tmp.insert(rhs).second) { 
	  _sol.update(t,tmp); 
	  if(!_changed[t]) {
	    _changed[t]=true;
	    _changed_count++;
	  }
	}
      }
    }
  }

  void do_ds_constraint2(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form *x = y
    unsigned int rhs(_varmap[c.rhs().var()].node);
    unsigned int loff(c.lhs().offset());
    unsigned int roff(c.rhs().offset());
    S const *sol(&_sol[rhs]);    

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {      
      if(type_safe(*j,loff)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+loff].node);
	if(int_add_edge(rhs,t,roff)) {
	  S tmp(_sol[t]);
	  if(int_set_union(tmp,*sol,roff)) {
	    _sol.update(t,tmp);
	    if(t == rhs) { sol = &_sol[rhs]; }
	    if(!_changed[t]) {
	      _changed_count++;
	      _changed[t]=true;
	    }
	  }
	}
      }
    }
  }

  void do_sd_constraint2(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form x = *y
    unsigned int lhs(_varmap[c.lhs().var()].node);
    unsigned int roff(c.rhs().offset());
    bool flag(false);
    S sol(_sol[lhs]);

    for(typename S::const_iterator j(delta.begin()); j!=delta.end();++j) {
      if(type_safe(*j,roff)) {
	// *j != NULL && *j != UNKNOWN
	unsigned int t(_varmap[*j+roff].node);       
	if(int_add_edge(t,lhs,0)) {
	  flag |= int_set_union(sol,_sol[t],0);
	}
      } 
    }
    if(flag) { 
      _sol.update(lhs,sol);
      if(!_changed[lhs]) { 
	_changed[lhs]=true;
	_changed_count++;
      }
    }
  }

  void do_complex_constraint2(constraint c,
			     S const &delta) {
    _stats.complex_count++;

    if(c.lhs().dereferenced()) {
      if(c.rhs().addressof()) {
	// *x=&y
	do_da_constraint2(c,delta);
      } else {
	// *x=y
	do_ds_constraint2(c,delta);
      }
    } else {
      // x=*y
      do_sd_constraint2(c,delta);
    }
  }

  void process_unifyqueue(void) {    
    for(std::vector<unsigned int>::const_iterator i(_unifyqueue.begin());
	i!=_unifyqueue.end();++i) {

      unsigned int np(*i);
      unsigned int n(_varmap[np].node);
      merge<set_union_func<unsigned int> >(n,np,_graph);
      if(valid_edge(diedge(n,n),_graph)) {
	// eliminate self loops
	vector_set<unsigned int> &labels(get(diedge(n,n),_graph));
	labels.erase((unsigned int) 0);
	if(labels.empty()) { erase_edge(diedge(n,n),_graph); }
      }

      if(_changed[np]) {
	_changed[np]=false;
	if(!_changed[n]) { _changed[n]=true;
	} else {
	  assert(_changed_count > 0);
	  _changed_count--;
	}
      }

      condense_nodes(n,np,_varmap);
      unifynodes(n,np);
      S tmp(_sol[n]);    
      S const &snp(_sol[np]);
      
      if(int_set_union(tmp,snp)) { 
	_sol.update(n,tmp); 
	if(!_changed[n]) {
	  _changed[n]=true;
	  _changed_count++;
	}
      }    
      _sol.clear(np); // save memory
    }
    // Finally, clear the unifyqueue.
    _unifyqueue.clear();
  }

  void unifynodes(unsigned int n, unsigned int np) {
  }

  bool redundant_constraint(unsigned int n, constraint const &c) {	    
    if(c.rhs().dereferenced()) {	     
      assert(c.rhs().var() == n);
      unsigned int lhs(c.lhs().var());
      // a = *n
      if(_varmap[lhs].node != lhs) {
	constraint tmp(expression(SCALAR,_varmap[lhs].node,0),c.rhs());
	_varmap[n].complicates.erase(c);
	if(!_varmap[n].complicates.insert(tmp).second) {
	  return true;
	}
      }
    } else {
      assert(c.lhs().var() == n);
      if(!c.rhs().addressof()) {
	unsigned int rhs(c.rhs().var());
	if(_varmap[rhs].node != rhs) {
	  // *n = a
	  constraint tmp(c.lhs(),expression(SCALAR,_varmap[rhs].node,c.rhs().offset()));
	  _varmap[n].complicates.erase(c);
	  if(!_varmap[n].complicates.insert(tmp).second) {
	    return true;
	  }
	}
      }
    }
    return false;
  }

  bool type_safe(unsigned int n, unsigned int o) {
    return n > 1 && (n+o) < _varmap[n].end;
  }

  inline bool int_set_union(S &to, S const &from, unsigned int inc = 0) {
    _stats.union_count++;
    _stats.union_size += to.size();
    _stats.union_size += from.size();

    if(inc == 0) {
      return set_union(to,from);
    } else {
      S tmp(from);
      sol_set_add<S>(tmp,inc,_varmap);
      return set_union(to,tmp);
    }
  }

  inline bool int_add_edge(unsigned int f, unsigned int t, unsigned int w) {
    if(t == f && w == 0) { 
      return false; 
    } else {
      diedge d(f,t);    
      bool r(add_edge(d,_graph));      
      r |= get(d,_graph).insert(w).second;
      if(r) { _stats.new_edge_count++; }
      return r;
    }
  }
};

template<class G, class S>
void solve_HT_FDS(G &g, 
		  std::vector<variable_info> &v,
		  store<S> &s,
		  unsigned int d, 
		  checker_stats &cs) {
  
  algorithm_HT_FDS<G,S> sc(g,v,s,d,cs);
  sc.solve();
}

#endif






















