#ifndef ALGORITHM_PW1_HPP
#define ALGORITHM_PW1_HPP

#include <set>
#include <string>
#include <boost/tuple/tuple.hpp>
#include "topological_sort.hpp"
#include "set_union.hpp"
#include "diedge.hpp"
#include "store.hpp"
#include "pcs.hpp" 

template<class G, class S, class W, class C>
class algorithm_PW1 {
private:
  G &_graph;
  std::vector<variable_info> &_varmap;
  unsigned int _domain_size;
  std::vector<unsigned int> _unifyqueue;
  store<S> &_sol;
  W _worklist;
  C _cycle_detector;
  
  // this final property is just used to obtain
  // useful information on the algorithms operation
  checker_stats &_stats;
public:
  algorithm_PW1(G &g, std::vector<variable_info> &v, 
		store<S> &s,
		unsigned int d,
		checker_stats &stats) :
    _graph(g),
    _varmap(v),
    _domain_size(d),
    _sol(s),
    _worklist(g),
    _cycle_detector(g,_varmap),
    _stats(stats) {
    //    assert(_sol.size() == size(_graph));
  }

  void solve(void) {
    for(unsigned int i=0;i!=size(_graph);++i) {
      if(_varmap[i].node == i) { _worklist.push(i); }
    }

    while(!_worklist.empty()) {     
      unsigned int n(_worklist.top());
      assert(n == _varmap[n].node);
      _worklist.pop();	
      _stats.visit_count++;
      // process nodes in topological order
      S solution(_sol[n]);	    
      // process complex constraints      
      vector_set<constraint> complicates(_varmap[n].complicates);
      for(vector_set<constraint>::const_iterator j(complicates.begin());
 	  j!=complicates.end();++j) {      
	if(!redundant_constraint(n,*j)) {
	  do_complex_constraint(*j,solution);	  
	}
      }

      _cycle_detector.finish(_unifyqueue);
      process_unifyqueue();

      // propagate solution to all successors
      typename G::const_siterator w, wend;
      for(boost::tie(w,wend) = successors(n,_graph);w!=wend;++w) {
	S tmp(_sol[*w]);
	
	if(int_set_union(tmp,solution)) { 
	  _worklist.push(*w);
	  _sol.update(*w,tmp); 
	}
      }
    }   
  }
  
private:
  void do_da_constraint(constraint const &c, 
			S const &delta) {
    // this function deals with constraints of the
    // form *x = &y
    unsigned int rhs(c.rhs().var());
    unsigned int off(c.lhs().offset());

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {
      if(type_safe(*j,off)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+off].node);
	S tmp(_sol[t]);
	
	if(tmp.insert(rhs).second) { 
	  _sol.update(t,tmp); 
	  _worklist.push(t);
	}
      }
    }
  }

  void do_ds_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form *x = y
    unsigned int rhs(_varmap[c.rhs().var()].node);
    unsigned int loff(c.lhs().offset());    
    //    S const &sol(_sol[rhs]);
    S sol(_sol[rhs]);

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {      
      if(type_safe(*j,loff)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+loff].node);
	if(t != rhs && add_edge(diedge(rhs,t),_graph)) {
	  // check for cycles 
	  _cycle_detector.new_edge(rhs,t,_unifyqueue);
	  process_unifyqueue();
	  // check for this node being collapsed
	  if(_varmap[rhs].node != rhs) {
	    rhs = _varmap[rhs].node;
	    int_set_union(sol,_sol[rhs]);
	  }

	  t = _varmap[t].node;

	  if(t != rhs) {
	    S tmp(_sol[t]);
	    
	    if(int_set_union(tmp,sol)) {
	      _worklist.push(t);
	      _sol.update(t,tmp);
	    }
	  }
	  
	  _stats.new_edge_count++;
	}
      }
    }
  }

  void do_sd_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form x = *y
    unsigned int lhs(_varmap[c.lhs().var()].node);
    unsigned int roff(c.rhs().offset());
    S sol(_sol[lhs]);
    bool flag(false);
    
    for(typename S::const_iterator j(delta.begin()); j!=delta.end();++j) {
      if(type_safe(*j,roff)) {
	// *j != NULL && *j != UNKNOWN
	unsigned int t(_varmap[*j+roff].node);       
	if(t != lhs && add_edge(diedge(t,lhs),_graph)) {
	  // check for cycles
	  _cycle_detector.new_edge(t,lhs,_unifyqueue);
	  process_unifyqueue();
	  // check for nodes being collapsed
	  if(lhs != _varmap[lhs].node) {
	    lhs = _varmap[lhs].node;
	    int_set_union(sol,_sol[lhs]);
	  }
	  t = _varmap[t].node;

	  if(t != lhs) { flag |= int_set_union(sol,_sol[t]); }
	  _stats.new_edge_count++;
	}
      }
    }
    if(flag) {
      _worklist.push(lhs);
      _sol.update(lhs,sol);
    }
  }

  void do_complex_constraint(constraint c,
			     S const &delta) {
    _stats.complex_count++;

    if(c.lhs().dereferenced()) {
      if(c.rhs().addressof()) {
	// *x=&y
	do_da_constraint(c,delta);
      } else {
	// *x=y
	do_ds_constraint(c,delta);
      }
    } else {
      // x=*y
      do_sd_constraint(c,delta);
    }
  }

  bool redundant_constraint(unsigned int n, constraint const &c) {	    
    // the point of this function is to reduce the number of
    // complex constraints floating around.
    if(c.rhs().dereferenced()) {	     
      assert(c.rhs().var() == n);
      unsigned int lhs(c.lhs().var());
      // a = *n
      if(_varmap[lhs].node != lhs) {
	constraint tmp(expression(SCALAR,_varmap[lhs].node,0),c.rhs());
	_varmap[n].complicates.erase(c);
	if(!_varmap[n].complicates.insert(tmp).second) {
	  return true;
	}
      }
    } else {
      assert(c.lhs().var() == n);
      if(!c.rhs().addressof()) {
	unsigned int rhs(c.rhs().var());
	if(_varmap[rhs].node != rhs) {
	  // *n = a
	  constraint tmp(c.lhs(),expression(SCALAR,_varmap[rhs].node,c.rhs().offset()));
	  _varmap[n].complicates.erase(c);
	  if(!_varmap[n].complicates.insert(tmp).second) {
	    return true;
	  }
	}
      }
    }
    return false;
  }

  void process_unifyqueue(void) { 
    if(_unifyqueue.empty()) { return; }

    std::vector<unsigned int>::iterator j(_unifyqueue.begin());
    std::vector<unsigned int>::iterator jend(_unifyqueue.end());    
    S tmp(_domain_size);
    while(j != jend) {
      unsigned int n(_varmap[*j].node);
      int_set_union(tmp,_sol[*j]);	
      merge_simple(n,*j,_graph);
      condense_nodes(n,*j,_varmap);
      _worklist.merge(n,*j);

      _sol.clear(*j);
      
      ++j;
      // if we've finished updating the solution of a component
      if(j == jend || _varmap[*j].node != n) {
	S const &lsol(_sol[n]);
	
	if(int_set_union(tmp,lsol) || lsol.size() < tmp.size()) { 
	  _sol.update(n,tmp); 
	  _worklist.push(n);
	}
	tmp.clear();
	// eliminate self loops
	erase_edge(diedge(n,n),_graph);
      }	  
    }
    _unifyqueue.clear(); // empty the queue now!
  } 

  bool type_safe(unsigned int n, unsigned int o) {
    return n > 1 && (n+o) < _varmap[n].end;
  }
  
  inline bool int_set_union(S &to, S const &from) {
    _stats.union_count++;
    _stats.union_size += to.size();
    _stats.union_size += from.size();
    return set_union(to,from);
  }
};
  
template<class G, class S, class W, class C>
void solve_PW1(G &g, 
	       std::vector<variable_info> &v,
	       store<S> &s,
	       unsigned int d,
	       checker_stats &cs) {

  algorithm_PW1<G,S,W,C> sc(g,v,s,d,cs);
  sc.solve();
}

#endif







