#ifndef ALGORITHM_PW3_HPP
#define ALGORITHM_PW3_HPP

#include <set>
#include <string>
#include <boost/tuple/tuple.hpp>
#include "topological_sort.hpp"
#include "set_union.hpp"
#include "diedge.hpp"
#include "store.hpp"
#include "pcs.hpp" 

template<class G, class S, class D>
class algorithm_PW3 {
private:
  G &_graph;
  std::vector<variable_info> &_varmap;
  unsigned int _domain_size;
  store<S> &_sol;
  unsigned int _changed_count;
  std::vector<bool> _changed;
  std::vector<unsigned int> _unifyqueue;
  D _cycle_detector;

  // this final property is just used to obtain
  // useful information on the algorithms operation
  checker_stats &_stats;
public:
  algorithm_PW3(G &g, std::vector<variable_info> &v, 
		store<S> &s,
		unsigned int d,
		checker_stats &stats) :
    _graph(g),
    _varmap(v),
    _domain_size(d),
    _sol(s),
    _changed_count(size(g)),
    _changed(size(g),true),
    _cycle_detector(g,v),
    _stats(stats) {

    //    assert(_sol.size() == size(_graph));
    for(unsigned int i=0;i!=_varmap.size();++i) {
      if(_varmap[i].node != i) {
	_changed[i]=false;
	_changed_count--;	    	
      }            
    }    
  }

  void solve(void) {
    while(_changed_count > 0) {
      _stats.iter_count++;

      _cycle_detector.finish(_unifyqueue);
      process_unifyqueue();

      // process nodes in topological order
      for(typename D::iterator i(_cycle_detector.begin());
	  i!=_cycle_detector.end();++i) {
	if(_varmap[*i].node == *i && _changed[*i]) {
	  _stats.visit_count++;
	  _changed[*i] = false;
	  _changed_count--;

	  // load solution for i
	  S solution(_sol[*i]);
	  
	  // process complex constraints
	  vector_set<constraint> complicates(_varmap[*i].complicates);
	  for(vector_set<constraint>::const_iterator j(complicates.begin());
	      j!=complicates.end();++j) {      
	    // attempt to reduce redundant constraints as they are expensive
	    if(!redundant_constraint(*i,*j)) {
	      do_complex_constraint(*j,solution);	  
	    }
	  }
	  
	  // propagate solution to all successors
	  typename G::const_siterator w, wend;
	  for(boost::tie(w,wend) = successors(*i,_graph);w!=wend;++w) {
	    S tmp(_sol[*w]);
	    
	    if(int_set_union(tmp,solution)) { 
	      _sol.update(*w,tmp); 
	      if(!_changed[*w]) {
		_changed[*w]=true;
		_changed_count++;
	      }
	    }
	  }
	}
      }
    }     
  }

private:
  void do_da_constraint(constraint const &c, 
			S const &delta) {
    // this function deals with constraints of the
    // form *x = &y
    unsigned int rhs(c.rhs().var());
    unsigned int off(c.lhs().offset());

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {
      if(type_safe(*j,off)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+off].node);
	S tmp(_sol[t]);
	
	if(tmp.insert(rhs).second) { 
	  _sol.update(t,tmp); 
	  if(!_changed[t]) {
	    _changed[t]=true;
	    _changed_count++;
	  }
	}
      }
    }
  }

  void do_ds_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form *x = y
    unsigned int rhs(_varmap[c.rhs().var()].node);
    unsigned int loff(c.lhs().offset());
    S const &sol(_sol[rhs]);    

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {      
      if(type_safe(*j,loff)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+loff].node);
	if(t != rhs && add_edge(diedge(rhs,t),_graph)) {
	  _cycle_detector.new_edge(rhs,t,_unifyqueue);
	  _stats.new_edge_count++;
	  S tmp(_sol[t]);
	  if(int_set_union(tmp,sol)) {
	    _sol.update(t,tmp);
	    if(!_changed[t]) {
	      _changed[t]=true;
	      _changed_count++;
	    }
	  }
	}
      }
    }
  }

  void do_sd_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form x = *y
    unsigned int lhs(_varmap[c.lhs().var()].node);
    unsigned int roff(c.rhs().offset());
    bool flag(false);
    S sol(_sol[lhs]);

    for(typename S::const_iterator j(delta.begin()); j!=delta.end();++j) {
      if(type_safe(*j,roff)) {
	// *j != NULL && *j != UNKNOWN
	unsigned int t(_varmap[*j+roff].node);       
	if(t != lhs && add_edge(diedge(t,lhs),_graph)) {
	  _cycle_detector.new_edge(t,lhs,_unifyqueue);
	  _stats.new_edge_count++;
	  flag |= int_set_union(sol,_sol[t]);
	}
      } 
    }
    if(flag) { 
      _sol.update(lhs,sol);
      if(!_changed[lhs]) {
	_changed[lhs]=true;
	_changed_count++;
      }

    }
  }

  void do_complex_constraint(constraint c,
			     S const &delta) {
    _stats.complex_count++;

    if(c.lhs().dereferenced()) {
      if(c.rhs().addressof()) {
	// *x=&y
	do_da_constraint(c,delta);
      } else {
	// *x=y
	do_ds_constraint(c,delta);
      }
    } else {
      // x=*y
      do_sd_constraint(c,delta);
    }
  }

  void process_unifyqueue(void) { 
    S tmp(_domain_size);
    std::vector<unsigned int>::iterator j(_unifyqueue.begin());
    std::vector<unsigned int>::iterator jend(_unifyqueue.end());
    while(j != jend) {
      unsigned int n(_varmap[*j].node);
      
      int_set_union(tmp,_sol[*j]);	
      merge_simple(n,*j,_graph);
      condense_nodes(n,*j,_varmap);
      if(_changed[*j]) {
	_changed[*j]=false;
	if(!_changed[n]) { _changed[n]=true;
	} else {
	  assert(_changed_count > 0);
	  _changed_count--;
	}
      }

      _sol.clear(*j);
      
      ++j;
      // if we've finished updating the solution of a component
      if(j == jend || _varmap[*j].node != n) {
	S const &lsol(_sol[n]);
	
	if(int_set_union(tmp,lsol) || lsol.size() < tmp.size()) { 
	  _sol.update(n,tmp); 
	  if(!_changed[n]) {
	    _changed[n]=true;
	    _changed_count++;
	  }
	}
	tmp.clear();
	// eliminate self loops
	erase_edge(diedge(n,n),_graph);
      }	  
    }
    _unifyqueue.clear(); // empty the queue now!
  } 

  bool redundant_constraint(unsigned int n, constraint const &c) {	    
    // the point of this function is to reduce the number of
    // complex constraints floating around.
    if(c.rhs().dereferenced()) {	     
      assert(c.rhs().var() == n);
      unsigned int lhs(c.lhs().var());
      // a = *n
      if(_varmap[lhs].node != lhs) {
	constraint tmp(expression(SCALAR,_varmap[lhs].node,0),c.rhs());
	_varmap[n].complicates.erase(c);
	if(!_varmap[n].complicates.insert(tmp).second) {
	  return true;
	}
      }
    } else {
      assert(c.lhs().var() == n);
      if(!c.rhs().addressof()) {
	unsigned int rhs(c.rhs().var());
	if(_varmap[rhs].node != rhs) {
	  // *n = a
	  constraint tmp(c.lhs(),expression(SCALAR,_varmap[rhs].node,c.rhs().offset()));
	  _varmap[n].complicates.erase(c);
	  if(!_varmap[n].complicates.insert(tmp).second) {
	    return true;
	  }
	}
      }
    }
    return false;
  }

  bool type_safe(unsigned int n, unsigned int o) {
    return n > 1 && (n+o) < _varmap[n].end;
  }

  inline bool int_set_union(S &to, S const &from) {
    _stats.union_count++;
    _stats.union_size += to.size();
    _stats.union_size += from.size();
    return set_union(to,from);
  }
};

template<class G, class S, class D>
void solve_PW3(G &g, 
		std::vector<variable_info> &v,
		store<S> &s,
	       unsigned int d,
		checker_stats &cs) {
  
  algorithm_PW3<G,S,D> sc(g,v,s,d,cs);
  sc.solve();
}

#endif







