#ifndef ALGORITHM_PW3_FDS_HPP
#define ALGORITHM_PW3_FDS_HPP

#include <set>
#include <string>
#include <boost/tuple/tuple.hpp>
#include "topological_sort.hpp"
#include "set_union.hpp"
#include "diedge.hpp"
#include "store.hpp"
#include "pcs.hpp" 

template<class G, class S>
class algorithm_PW3_FDS {
private:
  G &_graph;
  std::vector<variable_info> &_varmap;
  store<S> &_sol;
  unsigned int _changed_count;
  std::vector<bool> _changed;
  // this holds the (topological) iteration order
  // in reverse
  std::vector<unsigned int> _order;      

  // the final few vectors are used
  // by the algorithm for detecting
  // strongly connected components.
  std::vector<unsigned int> _vindex; 
  std::vector<bool> _in_component; 
  std::vector<unsigned int> _stack; 
  std::vector<unsigned int> _unifyqueue;
  std::vector<bool> _visited;
  unsigned int _index; 

  // this final property is just used to obtain
  // useful information on the algorithms operation
  checker_stats &_stats;
public:
  algorithm_PW3_FDS(G &g, std::vector<variable_info> &v, 
		    store<S> &s,
		    unsigned int d,
		    checker_stats &stats) :
    _graph(g),
    _varmap(v),
    _sol(s),
    _changed_count(size(g)),
    _changed(size(g),true),
    _vindex(size(g)), 
    _visited(size(g),false),
    _in_component(size(g),true), 
    _stats(stats) {
    //    assert(_sol.size() == size(_graph));
    for(unsigned int i=0;i!=_varmap.size();++i) {
      if(_varmap[i].node != i) {
	_changed[i]=false;
	_changed_count--;	    	
      }            
    }    
  }

  void solve(void) {
    while(_changed_count > 0) {
      _stats.iter_count++;

      // detect strongly connected components
      _index = 0;
      for(unsigned int i=0;i!=size(_graph);++i) {
	if(!_visited[i] && _varmap[i].node == i) { scc_visit(i); }
      }      
      process_unifyqueue();
      // now, topological sort graph
      _order.clear(); 
      for(unsigned int i=0;i!=size(_graph);++i) {
	if(_visited[i] && _varmap[i].node == i) { topo_visit(i); }
      }

      // process nodes in topological order
      for(std::vector<unsigned int>::reverse_iterator i(_order.rbegin());
	  i!=_order.rend();++i) {

	assert(_varmap[*i].node == *i);
	if(_changed[*i]) {
	  _stats.visit_count++;
	  _changed[*i] = false;
	  _changed_count--;
	  
	  // load solution for i
	  S solution(_sol[*i]);
	  
	  vector_set<constraint> complicates(_varmap[*i].complicates);
	  for(vector_set<constraint>::const_iterator j(complicates.begin());
	      j!=complicates.end();++j) {      
	    // attempt to reduce redundant constraints as they are expensive
	    if(!redundant_constraint(*i,*j)) {
	      do_complex_constraint(*j,solution);	  
	    }
	  }
	  
	  // propagate solution to all successors
	  typename G::const_siterator w, wend;
	  for(boost::tie(w,wend) = successors(*i,_graph);w!=wend;++w) {
	    S tmp(_sol[*w]);
	    bool flag(false);
	    // process weighted edges
	    vector_set<unsigned int> &labels(*(w.base()->second()));
	    assert(!labels.empty());
	    for(vector_set<unsigned int>::iterator j(labels.begin());
		j!=labels.end();++j) {
	      flag |= int_set_union(tmp,solution,*j);
	    }
	    if(flag) {
	      _sol.update(*w,tmp); 
	      if(!_changed[*w]) {
		_changed[*w]=true;
		_changed_count++;
	      }
	    }
	  }
	}
      }
    }
  }

private:
  void scc_visit(unsigned int n) {
    assert(_varmap[n].node == n);
    _visited[n]=true;
    _in_component[n]=false;
    _vindex[n]=_index++;
   
    // visit all successors
    typename G::const_siterator i,iend;
    for(boost::tie(i,iend) = successors(n,_graph);i!=iend;++i) {
      // only traverse zero weighted edges
      if(*(i.base()->second()->begin()) == 0) {
	unsigned int w(*i);
	if(!_visited[w]) { scc_visit(w); }
	if(!_in_component[w]) {
	  unsigned int t(_varmap[w].node);
	  if(_vindex[t] < _vindex[_varmap[n].node]) {
	    _varmap[n].node = t;
	  }
	}
      }
    }
    
    // at this point we must check to see
    // if any components have been identified
    if(_varmap[n].node == n) {
      unsigned int t(_vindex[n]);
      _in_component[n] = true;
      while(!_stack.empty() && t < _vindex[_stack.back()]) {
	unsigned int w(_stack.back());
	_varmap[w].node = n;
	_in_component[w] = true;
	_unifyqueue.push_back(w);
	_stack.pop_back(); 
      }
    } else {
      _stack.push_back(n);
    }
  }

  void topo_visit(unsigned int n) {
    assert(_varmap[n].node == n);
    _visited[n]=false;
   
    // visit all successors
    typename G::const_siterator i,iend;
    for(boost::tie(i,iend) = successors(n,_graph);i!=iend;++i) {
      // only traverse zero weighted edges
      if(_visited[*i]) { topo_visit(*i); }
    }

    _order.push_back(n);
  }

  void do_da_constraint(constraint const &c, 
			S const &delta) {
    // this function deals with constraints of the
    // form *x = &y
    unsigned int rhs(c.rhs().var());
    unsigned int off(c.lhs().offset());

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {
      if(type_safe(*j,off)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+off].node);
	S tmp(_sol[t]);
	if(tmp.insert(rhs).second) { 
	  _sol.update(t,tmp); 
	  if(!_changed[t]) {
	    _changed[t]=true;
	    _changed_count++;
	  }	  
	}
      }
    }
  }

  void do_ds_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form *x = y
    unsigned int rhs(_varmap[c.rhs().var()].node);
    unsigned int loff(c.lhs().offset());
    unsigned int roff(c.rhs().offset());
    S const *sol(&_sol[rhs]);    

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {      
      if(type_safe(*j,loff)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+loff].node);
	if(int_add_edge(rhs,t,roff)) {
	  S tmp(_sol[t]);
	  if(int_set_union(tmp,*sol,roff)) {
	    _sol.update(t,tmp);
	    if(t == rhs) { sol = &_sol[rhs]; }
	    if(!_changed[t]) {
	      _changed[t]=true;
	      _changed_count++;
	    }	  
	  }
	}
      }
    }
  }

  void do_sd_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form x = *y
    unsigned int lhs(_varmap[c.lhs().var()].node);
    unsigned int roff(c.rhs().offset());
    bool flag(false);
    S sol(_sol[lhs]);

    for(typename S::const_iterator j(delta.begin()); j!=delta.end();++j) {
      if(type_safe(*j,roff)) {
	// *j != NULL && *j != UNKNOWN
	unsigned int t(_varmap[*j+roff].node);       
	if(int_add_edge(t,lhs,0)) {
	  flag |= int_set_union(sol,_sol[t],0);
	}
      } 
    }
    if(flag) { 
      _sol.update(lhs,sol);
      if(!_changed[lhs]) {
	_changed[lhs]=true;
	_changed_count++;
      }	  
    }
  }

  void do_complex_constraint(constraint c,
			     S const &delta) {
    _stats.complex_count++;

    if(c.lhs().dereferenced()) {
      if(c.rhs().addressof()) {
	// *x=&y
	do_da_constraint(c,delta);
      } else {
	// *x=y
	do_ds_constraint(c,delta);
      }
    } else {
      // x=*y
      do_sd_constraint(c,delta);
    }
  }

  void process_unifyqueue(void) {
    S tmp(_varmap.size());
    std::vector<unsigned int>::iterator j(_unifyqueue.begin());
    std::vector<unsigned int>::iterator jend(_unifyqueue.end());
    while(j != jend) {
      unsigned int n(_varmap[*j].node);
      
      int_set_union(tmp,_sol[*j]);	
      merge<set_union_func<unsigned int> >(n,*j,_graph);
      condense_nodes(n,*j,_varmap);

      if(_changed[*j]) {
	_changed[*j]=false;
	if(!_changed[n]) { _changed[n]=true;
	} else {
	  assert(_changed_count > 0);
	  _changed_count--;
	}
      }

      _sol.clear(*j);
      
      ++j;
      // if we've finished updating the solution of a component
      if(j == jend || _varmap[*j].node != n) {
	S const &lsol(_sol[n]);
	
	if(int_set_union(tmp,lsol) || lsol.size() < tmp.size()) { 	  
	  _sol.update(n,tmp); 
	  if(!_changed[n]) {
	    _changed[n]=true;
	    _changed_count++;
	  }
	}
	tmp.clear();
	if(valid_edge(diedge(n,n),_graph)) {
	  // eliminate self loops
	  vector_set<unsigned int> &labels(get(diedge(n,n),_graph));
	  labels.erase((unsigned int) 0);
	  if(labels.empty()) { erase_edge(diedge(n,n),_graph); }
	}
      }	  
    }
    _unifyqueue.clear(); // empty the queue now!
  } // end process unify queue

  bool redundant_constraint(unsigned int n, constraint const &c) {	    
    if(c.rhs().dereferenced()) {	     
      assert(c.rhs().var() == n);
      unsigned int lhs(c.lhs().var());
      // a = *n
      if(_varmap[lhs].node != lhs) {
	constraint tmp(expression(SCALAR,_varmap[lhs].node,0),c.rhs());
	_varmap[n].complicates.erase(c);
	if(!_varmap[n].complicates.insert(tmp).second) {
	  return true;
	}
      }
    } else {
      assert(c.lhs().var() == n);
      if(!c.rhs().addressof()) {
	unsigned int rhs(c.rhs().var());
	if(_varmap[rhs].node != rhs) {
	  // *n = a
	  constraint tmp(c.lhs(),expression(SCALAR,_varmap[rhs].node,c.rhs().offset()));
	  _varmap[n].complicates.erase(c);
	  if(!_varmap[n].complicates.insert(tmp).second) {
	    return true;
	  }
	}
      }
    }
    return false;
  }

  bool type_safe(unsigned int n, unsigned int o) {
    return n > 1 && (n+o) < _varmap[n].end;
  }

  inline bool int_set_union(S &to, S const &from, unsigned int inc = 0) {
    _stats.union_count++;
    _stats.union_size += to.size();
    _stats.union_size += from.size();

    if(inc == 0) {
      return set_union(to,from);
    } else {
      S tmp(from);
      sol_set_add<S>(tmp,inc,_varmap);
      return set_union(to,tmp);
    }
  }

  inline bool int_add_edge(unsigned int f, unsigned int t, unsigned int w) {
    if(t == f && w == 0) { 
      return false; 
    } else {
      diedge d(f,t);    
      bool r(add_edge(d,_graph));
      r |= get(d,_graph).insert(w).second;
      if(r) { _stats.new_edge_count++; }
      return r;
    }
  }
};

template<class G, class S>
void solve_PW3_FDS(G &g, 
		   std::vector<variable_info> &v,
		   store<S> &s,
		   unsigned int d,
		   checker_stats &cs) {

  algorithm_PW3_FDS<G,S> sc(g,v,s,d,cs);
  sc.solve();
}

#endif







