#ifndef BIDIRECTIONAL_ADJACENCY_LIST_HPP
#define BIDIRECTIONAL_ADJACENCY_LIST_HPP

#include <map>
#include <set>
#include <vector>
#include <functional>
#include <assert.h>
#include <boost/tuple/tuple.hpp> // for tie
#include <boost/iterator_adaptors.hpp> // for transform_iterator
#include "diedge.hpp"
#include "vector_set.hpp"

#if __GNUC__ >= 3
#include <ext/functional>
#define SELECT1ST __gnu_cxx::select1st
#define PROJECT1ST __gnu_cxx::project1st
#else 
// everything else for the moment.
#include <functional>
#define SELECT1ST std::select1st
#define PROJECT1ST std::project1st
#endif

#define BDAL_USE_VECSET 1

class bdal_dummy_class {
public:
  bool operator==(bdal_dummy_class const &src) const {
    return true;
  }
};

template<class T>
class bdal_item {
public:
  typedef unsigned int first_type;
  typedef T *second_type;
private:
  unsigned int _first;
  T *_second;
public:
  bdal_item(unsigned int f, T *s) 
    : _first(f), _second(s) {
  }

  inline unsigned int first(void) const { return _first; }
  inline T *second(void) const { return _second; }

  inline bool operator<(bdal_item const &b) const {
    return _first < b._first;
  }

  inline bool operator==(bdal_item const &c) const {
    return _first == c._first;
  }
};

template<class T>
class bdal_select1st {
public:
  typedef typename T::first_type result_type;
  inline typename T::first_type operator()(T const &src) const {
    return src.first();
  }
};

template<>
class bdal_item<bdal_dummy_class> {
public:
  typedef unsigned int first_type;
  typedef bdal_dummy_class *second_type;

  unsigned int _first;
public:
  bdal_item(unsigned int f, bdal_dummy_class *s) 
    : _first(f) {
  }

  inline unsigned int first(void) const { return _first; }
  inline bdal_dummy_class *second(void) const { return static_cast<bdal_dummy_class*>(NULL); }

  inline bool operator<(bdal_item const &b) const {
    return _first < b._first;
  }

  inline bool operator==(bdal_item const &c) const {
    return _first == c._first;
  }
};

template<class T = bdal_dummy_class>
class bidirectional_adjacency_list {
  // typedefs
private:
#if BDAL_USE_VECSET == 0
  typedef std::set<bdal_item<T> > PredSetType;
  typedef std::set<bdal_item<T> > SuccSetType;
#else 
  typedef vector_set<bdal_item<T> > PredSetType;
  typedef vector_set<bdal_item<T> > SuccSetType;
#endif
public:
  // hmmm tasty typedef!
  typedef typename boost::transform_iterator_generator<bdal_select1st<typename SuccSetType::value_type>, typename SuccSetType::const_iterator>::type siterator;
  typedef typename boost::transform_iterator_generator<bdal_select1st<typename PredSetType::value_type>, typename PredSetType::const_iterator>::type piterator;

  typedef typename boost::transform_iterator_generator<bdal_select1st<typename SuccSetType::value_type>, typename SuccSetType::const_reverse_iterator>::type reverse_siterator;
  typedef typename boost::transform_iterator_generator<bdal_select1st<typename PredSetType::value_type>, typename PredSetType::const_reverse_iterator>::type reverse_piterator;

  typedef piterator const_piterator;
  typedef siterator const_siterator;
  typedef reverse_piterator const_reverse_piterator;
  typedef reverse_siterator const_reverse_siterator;

  typedef T value_type;
private:
  // friend functions
  typedef bidirectional_adjacency_list<T> self;
  friend void resize<>(unsigned int, self &);
  friend bool add_edge<>(diedge const &, self &);
  friend bool add_edge<>(diedge const &, T const &, self &);
  friend void erase_edge<>(diedge const &, self &);
  friend unsigned int size<>(self const &);
  friend std::pair<const_siterator,const_siterator> successors<>(unsigned int, self const &);
  friend std::pair<const_reverse_siterator,const_reverse_siterator> rsuccessors<>(unsigned int, self const &);

  friend unsigned int num_successors<>(unsigned int n, self const &);
  friend std::pair<const_piterator,const_piterator> predecessors<>(unsigned int, self const &);
  friend std::pair<const_reverse_piterator,const_reverse_piterator> rpredecessors<>(unsigned int, self const &);
  friend T &get<>(diedge const &e,self &);
  friend void clear<>(unsigned int, self &);
  friend bool valid_edge<>(diedge const &, self &);
private:
  typedef std::vector<SuccSetType> succ_type;
  typedef std::vector<PredSetType> pred_type;  

  succ_type _succs; // successors
  pred_type _preds; // predecessors
public:
  // constructors
  bidirectional_adjacency_list(unsigned int n = 0) 
    : _succs(n), _preds(n) {
  }
  
  ~bidirectional_adjacency_list(void) {
    for(typename std::vector<SuccSetType>::iterator i(_succs.begin());
	i!=_succs.end();++i) { 
      for(typename SuccSetType::iterator j(i->begin());
	  j!=i->end();++j) {
	delete j->second(); // free memory
      }					 
    }
  }
};

template<class T>
void resize(unsigned int n, bidirectional_adjacency_list<T> &bdl) {
  bdl._preds.resize(n);
  bdl._succs.resize(n);
}

template<class T>
bool valid_edge(diedge const &e, 
		bidirectional_adjacency_list<T> &bdl) {
  typename bidirectional_adjacency_list<T>::PredSetType::iterator i(bdl._preds[e.head].lower_bound(bdal_item<T>(e.tail,NULL)));
  if(i == bdl._preds[e.head].end() || i->first() != e.tail) {
    return false;
  } else {
    return true;
  }
}

template<class T>
bool add_edge(diedge const &e, 
	      bidirectional_adjacency_list<T> &bdl) {      
  assert(e.head < bdl._preds.size() && e.tail < bdl._succs.size());
  assert(bdl._preds.size() == bdl._succs.size());
  typename bidirectional_adjacency_list<T>::PredSetType::iterator i(bdl._preds[e.head].lower_bound(bdal_item<T>(e.tail,NULL)));

  if(i == bdl._preds[e.head].end() || i->first() != e.tail) {
    T *new_item = new T();
    bdl._preds[e.head].insert(i,bdal_item<T>(e.tail,new_item));
    bdl._succs[e.tail].insert(bdal_item<T>(e.head,new_item));
    return true;
  } else {
    return false;
  }  
}

template<class T>
bool add_edge(diedge const &e, T const &item,
	      bidirectional_adjacency_list<T> &bdl) {
  assert(e.head < bdl._preds.size() && e.tail < bdl._succs.size());
  assert(bdl._preds.size() == bdl._succs.size());
  typename bidirectional_adjacency_list<T>::PredSetType::iterator i(bdl._preds[e.head].lower_bound(bdal_item<T>(e.tail,NULL)));
  
  if(i == bdl._preds[e.head].end()) {
    T *new_item = new T(item);
    bdl._preds[e.head].insert(i,bdal_item<T>(e.tail,new_item));
    bdl._succs[e.tail].insert(bdal_item<T>(e.head,new_item));
    return true;
  } else {
    return false;
  } 
}

template<class T>
T &get(diedge const &e, bidirectional_adjacency_list<T> &bdl) {
  typename bidirectional_adjacency_list<T>::SuccSetType::iterator i(bdl._succs[e.tail].find(bdal_item<T>(e.head,NULL)));
  assert(i != bdl._succs[e.tail].end());
  assert(i->second() != NULL);
  return const_cast<T &>(*i->second()); // this is fine because it cannot affect set ordering
}

template<class T>
unsigned int size(bidirectional_adjacency_list<T> const &bdl) { 
  return bdl._preds.size();
}

template<class T>
void erase_edge(diedge const &e, 
		bidirectional_adjacency_list<T> &bdl) {
  typename bidirectional_adjacency_list<T>::PredSetType::iterator i(bdl._preds[e.head].find(bdal_item<T>(e.tail,NULL)));
  delete i->second(); // free memory ...
  bdl._succs[e.tail].erase(bdal_item<T>(e.head,NULL));
  bdl._preds[e.head].erase(i);
}

template<class T>
std::pair<typename bidirectional_adjacency_list<T>::const_siterator,
	  typename bidirectional_adjacency_list<T>::const_siterator> 
successors(unsigned int n, bidirectional_adjacency_list<T> const &bdl) {
  return std::make_pair(typename bidirectional_adjacency_list<T>::const_siterator(bdl._succs[n].begin()),
			typename bidirectional_adjacency_list<T>::const_siterator(bdl._succs[n].end()));
}

template<class T>
std::pair<typename bidirectional_adjacency_list<T>::const_reverse_siterator,
	  typename bidirectional_adjacency_list<T>::const_reverse_siterator> 
rsuccessors(unsigned int n, bidirectional_adjacency_list<T> const &bdl) {
  return std::make_pair(typename bidirectional_adjacency_list<T>::const_siterator(bdl._succs[n].rbegin()),
			typename bidirectional_adjacency_list<T>::const_siterator(bdl._succs[n].rend()));
}

template<class T>
unsigned int num_successors(unsigned int n, bidirectional_adjacency_list<T> const &bdl) {
  return bdl._succs[n].size();
}

template<class T>
unsigned int num_predecessors(unsigned int n, bidirectional_adjacency_list<T> const &bdl) {
  return bdl._preds[n].size();
}

template<class T>
std::pair<typename bidirectional_adjacency_list<T>::const_piterator,
	  typename bidirectional_adjacency_list<T>::const_piterator> 
predecessors(unsigned int n, bidirectional_adjacency_list<T> const &bdl) {
  return std::make_pair(bdl._preds[n].begin(),bdl._preds[n].end());
}

template<class T>
std::pair<typename bidirectional_adjacency_list<T>::const_reverse_piterator,
	  typename bidirectional_adjacency_list<T>::const_reverse_piterator> 
rpredecessors(unsigned int n, bidirectional_adjacency_list<T> const &bdl) {
  return std::make_pair(bdl._preds[n].rbegin(),bdl._preds[n].rend());
}


// this is really a graph algorithm because 
// it applies to all graphs which support
// bidirectional_adjacency

template<class MergeFunc /*  = PROJECT1ST <T,T> */, class T >
void merge(unsigned int n, unsigned int w, 
	   bidirectional_adjacency_list<T> &bdl) {

  MergeFunc merge_func;

  typename bidirectional_adjacency_list<T>::const_piterator i,iend;
  for(boost::tie(i,iend) = predecessors(w,bdl);
      i!=iend;++i) {
    
    unsigned int d(*i);
    if(*i == w) d = n;
    add_edge(diedge(d,n),bdl);	
    // could optimize this by using the
    // iterator to access the ?
    T &tmp(get(diedge(*i,w),bdl));
    merge_func(get(diedge(d,n),bdl),tmp);
  }
   
 typename bidirectional_adjacency_list<T>::const_siterator is,isend;
  for(boost::tie(is,isend) = successors(w,bdl);
      is!=isend;++is) {
    unsigned int d(*is);
    if(*is == w) d = n;

    add_edge(diedge(n,d),bdl);
    T &tmp(get(diedge(w,*is),bdl));
    merge_func(get(diedge(n,d),bdl),tmp);
  }
  
  // erase all edges involving ncomponent
  clear(w,bdl);
}

template<class T>
void merge_simple(unsigned int n, unsigned int w, 
		  bidirectional_adjacency_list<T> &bdl) {
  assert(n != w);
  typename bidirectional_adjacency_list<T>::const_piterator i,iend;
  for(boost::tie(i,iend) = predecessors(w,bdl);
      i!=iend;++i) {

    unsigned int d(*i);
    if(*i == w) d = n;
    add_edge(diedge(d,n),bdl);	
  }
   
 typename bidirectional_adjacency_list<T>::const_siterator is,isend;
  for(boost::tie(is,isend) = successors(w,bdl);
      is!=isend;++is) {
    unsigned int d(*is);
    if(*is == w) d = n;
    
    add_edge(diedge(n,d),bdl);
  }
  
  // erase all edges involving ncomponent
  clear(w,bdl);
}

template<class T>
void clear(unsigned int n,bidirectional_adjacency_list<T> &bdl) {
  typename bidirectional_adjacency_list<T>::
    SuccSetType::const_iterator iend(bdl._succs[n].end());
  
  for(typename bidirectional_adjacency_list<T>::
	SuccSetType::const_iterator i(bdl._succs[n].begin());
      i!=iend;++i) {
    if(i->first() != n) {
      typename bidirectional_adjacency_list<T>::
	PredSetType::iterator j(bdl._preds[i->first()].find(bdal_item<T>(n,NULL)));
      delete j->second(); // free memory ...
      bdl._preds[i->first()].erase(j);
    }
  }
  typename bidirectional_adjacency_list<T>::
    PredSetType::const_iterator ipend(bdl._preds[n].end());
  for(typename bidirectional_adjacency_list<T>::
	PredSetType::const_iterator i(bdl._preds[n].begin());
      i!=ipend;++i) {
    if(i->first() != n) {
      typename bidirectional_adjacency_list<T>::SuccSetType::iterator j
	  (bdl._succs[i->first()].find(bdal_item<T>(n,NULL)));
      delete j->second(); // free memory ...
      bdl._succs[i->first()].erase(j);
    }
  }
  bdl._preds[n].clear();
  bdl._succs[n].clear();
}

template<>
inline bool add_edge(diedge const &e, 
	      bidirectional_adjacency_list<bdal_dummy_class> &bdl) {      

  // this is for the case when we don't have any
  // edge properties
  assert(e.head < bdl._preds.size() && e.tail < bdl._succs.size());
  assert(bdl._preds.size() == bdl._succs.size());

  bidirectional_adjacency_list<bdal_dummy_class>::PredSetType::iterator i
    (bdl._preds[e.head].lower_bound(bdal_item<bdal_dummy_class>(e.tail,NULL)));

  if(i == bdl._preds[e.head].end() || i->first() != e.tail) {
    bdl._preds[e.head].insert(i,bdal_item<bdal_dummy_class>(e.tail,NULL));
    bdl._succs[e.tail].insert(bdal_item<bdal_dummy_class>(e.head,NULL));
    return true;
  } else {
    return false;
  }
}



#endif







