#ifndef BITSET_HPP
#define BITSET_HPP

#include <boost/cstdint.hpp>
#include <utility>
#include "set_union.hpp"
#include <assert.h>

// the purpose of this little dittie
// is to provide an equivalent of std::set<unsigned int>
// using a bit vector.
// note. it returns a proxy class and, hence,
//       is not really a valid container.

// Now, specialise my set_union function
// for the case when we are merging
// two bitsets
class bitset;
template<class T>
bool set_union(bitset &dest, T const &src) {
  bool r(false);
  for(typename T::const_iterator i(src.begin());
      i!=src.end();++i) {
    r |= dest.insert(*i).second;
  }
  return r;
}

template<>
bool set_union(bitset &dest, bitset const &src);

class bitset_iterator {
public:
  typedef std::forward_iterator_tag iterator_category;
  typedef unsigned int difference_type;
  typedef unsigned int value_type;
  typedef unsigned int *pointer;
  typedef unsigned int &reference;
private:
  mutable uint32_t *_ptr; // points into _array (below)
  mutable uint32_t _mask;
  mutable unsigned int _val;
public:
  inline bitset_iterator(uint32_t *p = NULL, unsigned int m = 0, unsigned int v = 0)
    : _ptr(p), _mask(m), _val(v) {
    if(p != NULL && (*p & _mask) == 0) {
      ++(*this);
    }
  }

  inline bitset_iterator operator++(int) const {
    bitset_iterator r(*this);
    ++(*this);
    return r;
  }

  inline bitset_iterator const &operator++(void) const {
    uint32_t tmp(*_ptr);
    assert(_mask != 0);
    do {
      if(_mask == 2147483648U) {
	// we've carried around
	_mask = 1U;
	++_ptr;
	tmp = *_ptr;
	while(tmp == 0) { _val += 32U; tmp=*(++_ptr); }
      } else {
	_mask <<= 1U;
      }
      ++_val;
      assert(_mask != 0);
    } while((tmp & _mask) == 0);

    return *this;
  }
  
  inline unsigned int operator*(void) const { return _val; }
  inline bool operator==(bitset_iterator const &src) const {
    return _val == src._val;
  }
  inline bool operator!=(bitset_iterator const &src) const {
    return _val != src._val;
  }
  inline bitset_iterator const &operator=(bitset_iterator const &src) const {
    _val = src._val;
    _ptr = src._ptr;
    _mask = src._mask;
    return *this;
  }

  void set(bool f) {
    if(f) {
      *_ptr |= _mask;
    } else {
      *_ptr &= ~_mask;
    }
  }
};

class bitset {
public:
  // typedefs
  typedef bitset_iterator iterator;
  typedef const bitset_iterator const_iterator;
  typedef unsigned int value_type;
  typedef unsigned int size_type;
private:
  friend bool set_union(bitset &, bitset const &);
  uint32_t *_bits;
  unsigned int _size; // number elements in set
  unsigned int _bits_size;
public:
  bitset(void);
  bitset(bitset const &);
  ~bitset(void);
  bitset const &operator=(bitset const &);
  iterator begin(void);
  const_iterator begin(void) const;
  iterator end(void);
  const_iterator end(void) const;
  
  size_type size(void) const { return _size; }
  bool empty(void) const { return _size == 0; }

  template<class InputIter>
  void insert(InputIter first, InputIter last)
  { for (; first != last; ++first) insert(*first); }
  
  std::pair<iterator, bool> insert(const value_type &);
  iterator insert(iterator start, const value_type &x);
  void erase(iterator pos);
  size_type erase(value_type const &x);
  void erase(iterator beg, iterator end);
  void clear(void);
  void swap(bitset &);
  const_iterator find(value_type const &) const;
  iterator find(value_type const &);
  iterator lower_bound(value_type const &);
  iterator upper_bound(value_type const &);

  bool operator==(bitset const &) const;
private:
  void destroy(void);

  void copy(bitset const &);
  void resize(unsigned int);
}; 

#endif







