#ifndef FACTORED_STORE_HPP
#define FACTORED_STORE_HPP

#if __GNUC__ >= 3
#include <ext/hash_set>
#define HASH_SET __gnu_cxx::hash_set
#else 
// everything else for the moment.
#include <functional>
#include <hash_set>
#define HASH_SET std::hash_set
#endif

#include <iostream>
#include <stdexcept>
#include <string>
#include "store.hpp"

template<class T>
class factored_store : public store<T> {
private:
  struct node {
    unsigned int count;
    T *ptr;
  public:
    node(unsigned int r, T *p) 
      : count(r), ptr(p) {
    }    
  };
  struct fc_comp {
    bool operator()(node const *a, node const *b) const {
      return *(a->ptr) == *(b->ptr);
    }
  };
  
  struct fc_hash {
    size_t operator()(node const *a) const {
      return a->ptr->size();
    }
  };

  typedef HASH_SET<node*, fc_hash, fc_comp> cache_type;
  cache_type _factors;
  std::vector<node *> _data;
  unsigned int _domain_size;
public:
  factored_store(unsigned int s, unsigned int d)
    : _data(s,NULL), _domain_size(d) {
  }

  factored_store(factored_store const &src) {
    copy(src);
  }
  
  ~factored_store(void) {
    for(typename cache_type::iterator i(_factors.begin());
	i!=_factors.end();) { 
      // for some reason, just using delete *i
      // causes a strange crash here ...
      node *item(*i);
      _factors.erase(i++);
      delete item->ptr;
      delete item;
    }
  }

  void clear(void) {
    // unsure if this is a good idea or not !

    for(typename cache_type::iterator i(_factors.begin());
	i!=_factors.end();) { 
      // for some reason, just using delete *i
      // causes a strange crash here ...
      node *item(*i);
      _factors.erase(i++);
      delete item->ptr;
      delete item;
    }

    std::fill(_data.begin(),_data.end(),static_cast<node*>(NULL));
  }

  void clear(unsigned int n) {
    node *p_old(_data[n]);
    if(p_old != NULL && --(p_old->count) == 0) {
      _factors.erase(p_old);
      delete p_old->ptr;
      delete p_old;
    }      
    _data[n]=NULL;
  }

  bool equal_sets(unsigned int a, unsigned int b) const {
    // this is an interesting optimisation
    return _data[a] == _data[b];
  }

  bool is_stored(unsigned int n) const { return _data[n] != NULL; }
  
  void update(unsigned int n, T const &data) {
    node *p_new(find_match(data));
    node *p_old(_data[n]);
    
    // do the reference counting stuff
    ++(p_new->count);
    if(p_old != NULL && --(p_old->count) == 0) {
      _factors.erase(p_old);
      delete p_old->ptr;
      delete p_old;
    }      
    
    _data[n] = p_new;
  }

  T const &operator[](unsigned int n) const { return *(_data[n]->ptr); }
  store<T> *clone(void) const { return new factored_store(*this); }

private:
  node *find_match(T const &data) {
    node tmp(0,const_cast<T*>(&data));
    typename cache_type::iterator i(_factors.find(&tmp));
    if(i == _factors.end()) {
      node *r(new node(0,new T(data)));
      _factors.insert(r);
      return r;
    }
    return *i;
  }

  void copy(factored_store const &src) {
    _data.resize(src._data.size(),NULL);
    _domain_size = src._domain_size;
    unsigned int index(0);
    for(typename std::vector<node *>::const_iterator i(src._data.begin());
	i!=src._data.end();++i,++index) {
      // check that the solution is non-empty
      // and then add it!
      if(*i != NULL) { update(index,*((*i)->ptr)); }
    }
  }
};

#endif











