#ifndef FIND_STRONG_COMPONENTS_HPP
#define FIND_STRONG_COMPONENTS_HPP

#include <vector>
#include "depth_first_search.hpp"

// This algorithm is based on the class algorithm by Robert Tarjan.
// The use of the STACK instead of the path relation comes
// from the paper:
//
// "On Finding the Strongly Connected Components in a Directed
//  Graph", Esko Nuutila, Eljas Soisalon-Soininen, 1994

class empty_scc_visitor {
 public:
  inline void start_vertex(unsigned int n) {}
  inline void finished_vertex(unsigned int n) {}
  inline void traversing_edge(unsigned int t, unsigned int h) {}
  inline void finished_edge(unsigned int t, unsigned int h) {}
  inline void skipping_edge(unsigned int t, unsigned int h) {}
  inline void unify(unsigned int, unsigned int) {}
};

template<class N2iMap, class SccVisitor = empty_scc_visitor>
class find_scc_visitor {
 protected:
  std::vector<bool> _in_component;
  std::vector<unsigned int> _vindex;
  std::vector<unsigned int> _stack;
  N2iMap &_root;
  unsigned int _index;
  SccVisitor _visitor;
public:
  find_scc_visitor(N2iMap &root, unsigned int s)
    : _in_component(s,false), _root(root), _vindex(s), _index(0) {
  }
  
  find_scc_visitor(N2iMap &root, unsigned int s, SccVisitor visitor)
    : _in_component(s,false), _root(root), _vindex(s), _index(0), _visitor(visitor) {
  }

  inline void start_vertex(unsigned int n) {
    _vindex[n]=_index++;
    _visitor.start_vertex(n);
    _root[n]=n;
    _in_component[n]=false;
  }

  inline void finished_vertex(unsigned int n) {
    if(_root[n] == n) {
      unsigned int t(_vindex[n]);
      _in_component[n]=true;
      while(!_stack.empty() && t < _vindex[_stack.back()]) {
	unsigned int w(_stack.back());
	_root[w]=n;
	_in_component[w]=true;
	_visitor.unify(n,w);
	_stack.pop_back(); 
      }
      _visitor.finished_vertex(n);
    } else {
      _stack.push_back(n);
    }
  }
  inline void traversing_edge(unsigned int t, unsigned int h) {
    _visitor.traversing_edge(t,h);
  }
  inline void finished_edge(unsigned int t, unsigned int h) {
    _finished_edge(t,h);
    _visitor.finished_edge(t,h);
  }
  inline void _finished_edge(unsigned int t, unsigned int h) {
    if(!_in_component[h]) {
      unsigned int tmp(_root[h]);
      if(_vindex[tmp] < _vindex[_root[t]]) {
	_root[t]=tmp;
      }
    }
  }
  inline void skipping_edge(unsigned int t, unsigned int h) {
    _finished_edge(t,h);
    _visitor.skipping_edge(t,h);
  }
};

template<class G, class N2iMap, class VisitedMap, class Visitor>
void find_strong_components(G const &g, N2iMap &root, VisitedMap &visited, Visitor visitor) {
  find_scc_visitor<N2iMap,Visitor> fsv(root,size(g),visitor);
  for(unsigned int i=0;i!=size(g);++i) {
    if(!visited[i]) {
      depth_first_search<G,std::vector<bool>,find_scc_visitor<N2iMap,Visitor>&>(i,g,visited,fsv);
    }
  }
}

template<class G, class N2iMap>
void find_strong_components(G const &g, N2iMap &root) {
  std::vector<bool> visited(size(g),false);
  find_scc_visitor<N2iMap> fsv(root,size(g));
  for(unsigned int i=0;i!=size(g);++i) {
    if(!visited[i]) {
      // this is sooo ugly because is use a state containing dfs visitor ...
      depth_first_search<G,std::vector<bool>,find_scc_visitor<N2iMap>&>(i,g,visited,fsv);
    } 
  }
}

template<class G, class N2iMap, class Visitor>
void find_strong_components(G const &g, N2iMap &root, Visitor visitor) {
  std::vector<bool> visited(size(g),false);
  find_scc_visitor<N2iMap,Visitor> fsv(root,size(g),visitor);
  for(unsigned int i=0;i!=size(g);++i) {
    if(!visited[i]) {
      // this is sooo ugly because is use a state containing dfs visitor ...
      depth_first_search<G,std::vector<bool>,find_scc_visitor<N2iMap,Visitor>&>(i,g,visited,fsv);
    } 
  }
}

#endif
