#ifndef INVSET_HPP
#define INVSET_HPP

template<class T>
class invset;

template<class T>
bool set_union(invset<T> &to, invset<T> const &from) {
  // there are basically 4 cases to consider:
  // S U S, S^-1 U S, S U S^-1 and S^-1 U S^-1
  if(static_cast<void const *>(&to) != static_cast<void const *>(&from)) {
    if(!to._imode && !from._imode) {      
      if(set_union(to._set,from._set)) {	
	if(to._set.size() > to._domain) { to.invert(); }
	return true;
      }   
    } else if(to._imode && from._imode) {
      return set_intersection(to._set,from._set);
    } else if(to._imode && !from._imode) {
      return set_difference(to._set,from._set);
    } else {
      invset<T> tmp(from);
      if(set_difference(tmp._set,to._set)) {
	to.swap(tmp);
	return true;
      }
    }
  }
  
  return false;
}

template<class T>
class invset_iterator {
public:
  typedef std::forward_iterator_tag iterator_category;
  typedef unsigned int difference_type;
  typedef unsigned int value_type;
  typedef unsigned int *pointer;
  typedef unsigned int &reference;
private:
  mutable unsigned int _counter;
  mutable bool _imode;
  mutable typename T::const_iterator _iter;
  // note the following could be removed by always insuring
  // that the domain size was in the inverse array.  
  mutable typename T::const_iterator _end;
public:
  inline invset_iterator(unsigned int c, bool ms, typename T::const_iterator i,
			 typename T::const_iterator e)
    : _counter(c), _imode(ms), _iter(i), _end(e) {
  }

  inline invset_iterator operator++(int) const {
    invset_iterator r(*this);
    ++(*this);
    return r;
  }

  inline invset_iterator const &operator++(void) const {
    if(_imode) {
      // ok, we're in inverse mode.            
      ++_counter;
      while(_iter != _end && _counter == *_iter) {
	++_counter;
	++_iter;
      }
    } else {
      ++_iter;
    }

    return *this;
  }
  
  inline unsigned int operator*(void) const { 
    if(_imode) { 
      return _counter;
    } else {
      return *_iter;
    }
  }

  inline bool operator==(invset_iterator const &src) const {    
    // again possible optimisation.
    return _counter == src._counter && _iter == src._iter;
  }
  inline bool operator!=(invset_iterator const &src) const {
    return _counter != src._counter || _iter != src._iter;
  }

  inline invset_iterator const &operator=(invset_iterator const &src) const {
    _counter = src._counter;
    _imode = src._imode;
    _iter = src._iter;
    _end = src._end;
    return *this;
  }
};

template<class T>
class invset {
public:
  typedef unsigned int value_type;
  typedef typename T::key_compare key_compare;
  typedef typename T::key_compare value_compare;
  typedef typename T::pointer pointer;
  typedef typename T::reference reference;
  typedef typename T::const_reference const_reference;
  typedef typename T::size_type size_type;
  typedef typename T::difference_type difference_type;
  typedef invset_iterator<T> iterator;
  typedef const invset_iterator<T> const_iterator;

  friend bool set_union<>(invset &, invset const &);
private:
  unsigned int _domain;
  bool _imode; // if true then we're in inverse mode
  T _set;
public:
  invset(unsigned int d) : _imode(false), _domain(d) {
  }

  template<class InputIterator>
  invset(InputIterator beg, InputIterator end) {
    for(;beg!=end;++beg) {
      insert(*beg);      
    }
  }

  invset(invset const &src) 
    : _domain(src._domain),_imode(src._imode), _set(src._set) {
  }

  iterator begin(void) {
    if(_imode) {
      unsigned int start(0);
      typename T::iterator j(_set.begin()),jb(_set.end());
      while(j != jb && start == *j) { 
	++start;
	++j;
      }
      return iterator(start,_imode,j,jb);
    } else {
      return iterator(0,_imode,_set.begin(),_set.end());
    }
  }

  iterator end(void) {
    return iterator(_imode ? _domain : 0,_imode,_set.end(),_set.end());
  }

  const_iterator begin(void) const {
    if(_imode) {
      unsigned int start(0);
      typename T::const_iterator j(_set.begin()),jb(_set.end());
      while(j != jb && start == *j) { 
	++start;
	++j;
      }
      return const_iterator(start,_imode,j,jb);
    } else {
      return const_iterator(0,_imode,_set.begin(),_set.end());
    }
  }

  const_iterator end(void) const {
    return const_iterator(_imode ? _domain : 0,_imode,_set.end(),_set.end());
  }

  size_type size(void) const { return _imode ? _domain - _set.size() : _set.size(); }
  bool empty(void) const { return !_imode && _set.size() == 0; }

  std::pair<iterator, bool> insert(unsigned int v) {
    // note, there is a problem if we insert an element
    // which causes the invert to take place as this
    // will invalidate the end() value?

    if(_imode) {
      bool r = _set.erase(v) == 1;
      // item was erased
      return std::make_pair(iterator(v,true,_set.lower_bound(v),_set.end()),r);
    } else {
      if(_set.size () >= (_domain >> 1)) {
	invert();
	return insert(v);
      } else {
	std::pair<typename T::iterator, bool> r(_set.insert(v));
	return std::make_pair(iterator(0,false,r.first,_set.end()),r.second);
      }
    }
  }

  template<class InputIter>
  void insert(InputIter first, InputIter last)
  { for (; first != last; ++first) insert(*first); }

  iterator insert(iterator start, unsigned int x) {
    return insert(x).first;
  }

  void swap(invset &src) {
    bool imode(src._imode);
    unsigned int domain(src._domain);
    _set.swap(src._set);
    src._imode = _imode;
    src._domain = _domain;
    _imode = imode;
    _domain = domain;
  }
  
  void clear(void) {
    _set.clear();
    _imode = false;
  }

  bool operator==(invset const &src) const {
    return _imode == src._imode && _domain == src._domain && _set == src._set;
  }

  unsigned int domain(void) const { return _domain; }
private:
  void invert(void) {
    assert(!_imode);
    T nset;
    typename T::iterator j(_set.begin());
    typename T::iterator jb(_set.end());
    for(unsigned int i=0;i!=_domain;++i) {
      if(j != jb && *j == i) {	
	++j;
      } else {
	nset.insert(nset.end(),i);
      }      
    }
    _set.swap(nset);
    _imode=true;
  }
};


#endif
