#ifndef DIFFERENCE_PROPAGATION_BASE_HPP
#define DIFFERENCE_PROPAGATION_BASE_HPP

#include "solver_common.hpp"

#define USE_SET_UNION_DIFF 0

template<class G, class S>
class difference_propagation_base : public solver_common<G,S> {
protected:
  S _delta;
private:
  store<S> *_deltas;     // contains the delta information
public:
  difference_propagation_base(G &g, std::vector<variable_info> &v, store<S> &s, 
		       checker_stats &stats, bool f) 
    : solver_common<G,S>(g,v,s,stats,f), _deltas(s.clone()) {
    // here we empty the current solutions.
    // if we don't do this then the _delta computation in
    // enter_vertex would give zero back for the first time ...
    S tmp;
    for(unsigned int i=0;i!=size(g);++i) {
      if(v[i].node == i) {
	_sol.update(i,tmp);
      }
    }
  }

  void enter_vertex(unsigned int n) {
    if(_deltas->is_stored(n)) { 
      S ns(_sol[n]);
      S const &nd((*_deltas)[n]);
      _delta.clear();     
      
      _stats.union_count++;
      _stats.union_size += std::min(ns.size(),nd.size());
      
      if(union_diff(ns,nd,_delta)) {
	_sol.update(n,ns);
      }    
      // clear delta information on node n
      _deltas->clear(n);
    }
  }

  void enter_vertex3(unsigned int n) {
    if(_deltas->is_stored(n)) { 
      _delta = (*_deltas)[n]; 
    } else {
      _delta.clear();
    }
  }

  void enter_vertex2(unsigned int n) {
    S const &ns(_sol[n]); 
    S nd;

    if(_deltas->is_stored(n)) { nd = (*_deltas)[n]; }
    _delta.clear();

    for(typename S::iterator i(nd.begin());
	i!=nd.end();++i) {
      if(ns.find(*i) == ns.end()) {
	_delta.insert(_delta.end(),*i);
      }
    }
    
    if(nd.size() > _delta.size()) {
      _deltas->update(n,_delta);
    }
  }

  void finish_vertex(unsigned int n) {
    if(_deltas->is_stored(n)) { 
      S ns(_sol[n]);
      if(set_union(ns,(*_deltas)[n])) {
	_sol.update(n,ns);
      }      
      _deltas->clear(n);
    }    
  }

  
  bool do_propagate(unsigned int dest, vector_set<unsigned int> &labels) {
    bool flag(false);
    S dest_delta;
    
    if(_deltas->is_stored(dest)) {
      dest_delta = (*_deltas)[dest];
    } 
    
    for(vector_set<unsigned int>::const_iterator j(labels.begin());
	j!=labels.end();++j) {
      // propagate along edge c->*i, with label *j.      
      flag |= propagate(dest_delta,_delta,*j);
    }
    
    if(flag) { _deltas->update(dest,dest_delta); }
    return flag;
  }

  bool internal_unify(unsigned int n, unsigned int np) {
#if 0
    bool flag(false);
    S dns;
    if(_deltas->is_stored(n)) { dns = (*_deltas)[n]; }
    
    assert(_sol.is_stored(np));
    
    if(!_sol.equal_sets(n,np)) {
      S const &ns(_sol[n]);
      S const &nps(_sol[np]);
      S intersect;
      
      set_intersection(ns.begin(),ns.end(),
		       nps.begin(),nps.end(),
		       inserter(intersect,intersect.begin()));
            
      flag = true;
      // open question regarding whether this can
      // be done better ...
      set_difference(ns.begin(),ns.end(),
		     intersect.begin(),intersect.end(),
		     inserter(dns,dns.begin()));
      
      set_difference(nps.begin(),nps.end(),
		     intersect.begin(),intersect.end(),
		     inserter(dns,dns.begin()));

      _sol.update(n,intersect);
    }
    
    if(_deltas->is_stored(np)) { 
      S const &dnps((*_deltas)[np]);
    
      _stats.union_count++;
      _stats.union_size += std::min(dns.size(),dnps.size());    

      if(set_union(dns,dnps)) {
	flag = true;
      }
    }

    if(flag) { _deltas->update(n,dns); }
    
    return flag;
#else
    S ns(_sol[n]);
    set_union(ns,_sol[np]);
    
    if(_deltas->is_stored(n))  { set_union(ns,(*_deltas)[n]); }
    if(_deltas->is_stored(np)) { set_union(ns,(*_deltas)[np]); }

    _sol.update(n,S());
    _deltas->update(n,ns);
    // finally, free up some memory ...
    _sol.clear(np);
    _deltas->clear(np);

    return true;   
#endif
  }
  
  bool do_new_edge(unsigned int tail, unsigned int head, unsigned int off) {
    if(_deltas->is_stored(head)) { 
      S hs((*_deltas)[head]);
      if(propagate(hs,_sol[tail],off)) { 
	return false;
      }
      _deltas->update(head,hs);
    } else {
      _deltas->update(head,_sol[tail]);
    }
    return true;
  }
  
  bool do_process_trivial(unsigned int node, unsigned int val, bool mode) {
    if(mode) {
      S tmp(_sol[node]);
      if(tmp.find(val) != tmp.end()) { return false; }
    } 

    S tmp;
    if(_deltas->is_stored(node)) { tmp = (*_deltas)[node]; }
    
    if(tmp.insert(val).second) { 
      _deltas->update(node,tmp); 
      return true;
    } else {
      return false;
    }
  }

  inline bool union_diff(S &to, S const &from, S &diff) {
#if USE_SET_UNION_DIFF == 1
    return set_union_diff(to,from,diff);
#else
    // the point of this version is that it's
    // not linear in the size of _sol[*i]
    
    bool flag(false);
    for(typename S::const_iterator j(from.begin());
	j!=from.end();++j) {
      if(to.insert(*j).second) { diff.insert(*j); flag = true; }
    }
    return flag;
#endif
  }
};

#endif




