#include <set>
#include <stdexcept>

#include "bidirectional_adjacency_list.hpp"
#include "factored_store.hpp"
#include "simple_store.hpp"
#include "vector_set.hpp"
#include "bitset.hpp"
#include "pcs.hpp"

#include "solver/heintze_tardieu/heintze_tardieu_solver.hpp"

using namespace std;

typedef bidirectional_adjacency_list<vector_set<unsigned int> >  graph_t;

template<class T>
store<vector_set<unsigned int> > *heintze_tardieu_solver_main_template(vector<unsigned int> &params,
								   graph_t &g, 
								   vector<variable_info> &varmap,
								   checker_stats &stats,
								   bool ftc) {  
  store<T> *sol = NULL;

  switch(params[1]) {
  case SIMPLE_STORE:
    sol = new simple_store<T>(size(g));
    break;    
  case FACTORED_STORE:
    sol = new factored_store<T>(size(g));
    break;    
  default:
    throw runtime_error("invalid store selected");
  }

  bool ocd,tred;

  switch(params[3]) {
  case DFS_ONLINE_CYCLE_DETECTION:
    ocd=true;
    break;
  case STATIC_CYCLE_DETECTION:
  case NO_CYCLE_DETECTION:
  case FULL_VARIABLE_SUBST:
    ocd=false;
    break;
  default:
    throw runtime_error("invalid cycle detector specified");
  }

  switch(params[4]) {
  case TRANSITIVE_REDUCTION:
    tred=true;
    break;
  case NO_TRANSITIVE_REDUCTION:
    tred=false;
    break;
  default:
    throw runtime_error("invalid transitive reduction specifier");
  }

  for(unsigned int i=0;i!=varmap.size();++i) {
    T tmp;
    tmp.insert(varmap[i].solution.begin(),varmap[i].solution.end());
    sol->update(i,tmp);
  }

  heintze_tardieu_solve<graph_t,T>(g, varmap, *sol, stats); 

  store<vector_set<unsigned int> > *r = new factored_store<vector_set<unsigned int> >(size(g));
  
  for(unsigned int i=0;i!=varmap.size();++i) {
    if(sol->is_stored(i)) {
      vector_set<unsigned int> tmp;
      tmp.insert((*sol)[i].begin(),(*sol)[i].end());
      r->update(i,tmp);
    }
  }

  delete sol;

  return r;
}

store<vector_set<unsigned int> > *heintze_tardieu_solver_main(vector<unsigned int> &params,
							      graph_t &g, 				     
							      vector<variable_info> &varmap,
							      checker_stats &stats,
							      bool ftc) {  

  if(params.size() != 5) {
    throw runtime_error("Insufficient arguments to Heintze-Tardieu Algorithm");
  }


  switch(params[2]) {
  case BITSET:
    return heintze_tardieu_solver_main_template<bitset>(params,g,varmap,stats,ftc);
    break;
  case VECTOR_SET:
    return heintze_tardieu_solver_main_template<vector_set<unsigned int> >(params,g,varmap,stats,ftc);
    break;
  case STLSET:
    return heintze_tardieu_solver_main_template<std::set<unsigned int> >(params,g,varmap,stats,ftc);
    break;
  default:
    throw runtime_error("invalid set implementation");
  }
}

