#ifndef PEARCE_SOLVER_HPP
#define PEARCE_SOLVER_HPP

#include <boost/tuple/tuple.hpp>
#include "pcs.hpp"
#include "solver_common.hpp"
#include "set_union.hpp"
#include "store.hpp"
#include "find_strong_components.hpp"
#include "filter_graph.hpp"
#include "topological_sort.hpp"

#define USE_SET_UNION_DIFF 0
#define HACK_TEST 1

class ps_scc_visitor {
private:
  std::vector<unsigned int> &_order;
  std::vector<unsigned int> &_queue;
public:
  ps_scc_visitor(std::vector<unsigned int> &o,
		 std::vector<unsigned int> &q) 
    : _order(o), _queue(q) {
  }
  inline void start_vertex(unsigned int n) {}
  inline void finished_vertex(unsigned int n) {
    _order.push_back(n);
  }
  inline void traversing_edge(unsigned int t, unsigned int h) {}
  inline void finished_edge(unsigned int t, unsigned int h) {}
  inline void skipping_edge(unsigned int t, unsigned int h) {}
  inline void unify(unsigned int n, unsigned int w) {
    _queue.push_back(w);
  }
};

template<class G, class S>
class pearce_solver_common : public solver_common<G,S> {
private:
  store<S> *_deltas;
  std::vector<unsigned int> _rorder; // topological order in reverse
  std::vector<unsigned int> _unifyqueue;
  std::vector<bool> _visited;
  bool _change;
  checker_stats &_stats;
public:
  pearce_solver_common(G &g, std::vector<variable_info> &v, 
		       store<S> &s, 
		       checker_stats &stats,
		       bool ftc = false) 
    : solver_common<G,S>(g,v,s,ftc), _deltas(s.clone()),
    _rorder(size(g)), _visited(size(g),false), _stats(stats) {
    // we empty sol here ...
    for(unsigned int i=0;i!=_varmap.size();++i) {
      _sol.update(i,S());
    }
    topological_sort(g,_rorder);
  }

  ~pearce_solver_common(void) {
    delete _deltas;
  }

  void solve(void) {
    unsigned int iter(0);
    unsigned int dsize(0);
    unsigned int dcount(0);
    do { 
      _change = false;
      for(std::vector<unsigned int>::reverse_iterator i(_rorder.rbegin());
	  i!=_rorder.rend();++i) {
	if(_varmap[*i].node == *i) {
	  bool flag(false);
	  S delta;
	  S tmp;
	  tmp = _sol[*i];
#if USE_SET_UNION_DIFF == 1
	  flag = set_union_diff(tmp,(*_deltas)[*i],delta);
#else
	  // the point of this version is that it's
	  // not linear in the size of _sol[*i]
	  S const &t2((*_deltas)[*i]);
	  for(typename S::const_iterator j(t2.begin());
	      j!=t2.end();++j) {
	    if(tmp.insert(*j).second) { delta.insert(*j); flag = true; }
	  }
#endif
	  if(flag) { _sol.update(*i,tmp); }
	  _deltas->update(*i,S());

	  dsize += delta.size();
	  dcount ++;

	  if(!delta.empty()) {	    
	    _stats.visit_count++;
	    vector_set<constraint>::const_iterator jend(_varmap[*i].complicates.end());
	    for(vector_set<constraint>::const_iterator j(_varmap[*i].complicates.begin());
		j!=jend;++j) {	
	      _stats.complex_count++;
	      do_complex_constraint(*j,delta);
	    }
	    
	    // now propagate solution from *i to all successors
	    typename G::const_siterator is, isend;
	    for(boost::tie(is,isend) = successors(*i,_graph);is!=isend;++is) {
	      // there is a major optimization opportunity here
	      // because it's unnecessary to perform the 
	      // full lookup to determine the labels associated
	      // with this edge.
	      bool flag(false);
	      unsigned int dest(_varmap[*is].node);
	      S dest_set((*_deltas)[dest]);
	      
	      std::set<unsigned int> &labels(*(is.base()->second()));
	      for(std::set<unsigned int>::const_iterator j(labels.begin());
		  j!=labels.end();++j) {
		// propagate along edge c->*i, with label *j.
		
		flag |= propagate(dest_set,delta,*j);
	      }
	      
	      if(flag) {
		_deltas->update(dest,dest_set);	  
		_change=true;
	      }
	    }
	  }
	}
      }
      _rorder.clear();
      filter_graph<G,zero_edge_select> fg(_graph);
      n2c_wrapper n2c(_varmap); // uggly
      ps_scc_visitor visitor(_rorder,_unifyqueue);
      find_scc_visitor<n2c_wrapper,ps_scc_visitor> fsv(n2c,size(_graph),visitor);
      for(unsigned int i=0;i!=_varmap.size();++i) {
	if(!_visited[i] && _varmap[i].node == i) {
	  depth_first_search<filter_graph<G,zero_edge_select>,std::vector<bool>,
	    find_scc_visitor<n2c_wrapper,ps_scc_visitor>&>(i,fg,_visited,fsv);
	} else {
	  _visited[i]=true;
	}
      }
      // process the unify queue.
      process_unifyqueue();
#if HACK_TEST == 1
      // quick hackery      
      fill(_visited.begin(),_visited.end(),false);
      topological_sort(_graph,_rorder,_visited);
#endif
      // could get rid of this ...
      fill(_visited.begin(),_visited.end(),false);
      double tmp(dsize);
      tmp /= dcount;
      dsize=0;
      dcount=0;

    } while(_change);
  }

private:
  void new_edge(unsigned int tail, unsigned int head, unsigned int lab) {
    bool flag(false);
    S head_set((*_deltas)[head]);
    if(set_union(head_set,_sol[tail])) {
      _deltas->update(head,head_set);
      _change=true;
    }
  }

  void process_trivial(unsigned int n, unsigned int v) {
    S tmp((*_deltas)[n]);
    if(tmp.insert(v).second) {
      _change = true;
      _deltas->update(n,tmp);
    }
  }

  void process_unifyqueue(void) {
    std::vector<unsigned int>::const_iterator iend(_unifyqueue.end());
    for(std::vector<unsigned int>::const_iterator i(_unifyqueue.begin());
	i!=iend;++i) {
      unsigned int np(*i);
      unsigned int n(_varmap[np].node);      
      S nd((*_deltas)[n]);
      S intersect;
      
      // first we compute the intersection of their
      // current solutions
      std::set_intersection(_sol[n].begin(),_sol[n].end(),
			    _sol[np].begin(),_sol[np].end(),
			    inserter(intersect,intersect.begin()));

      // if this doesn't cover both sets then
      // we must do more work.
      
      if(_sol[n].size() > intersect.size() ||
	 _sol[np].size() > intersect.size()) {
	_change = true;
	// open question regarding whether this can
	// be done better ...
	std::set_difference(_sol[n].begin(),_sol[n].end(),
			    intersect.begin(),intersect.end(),
			    inserter(nd,nd.begin()));
	
	std::set_difference(_sol[np].begin(),_sol[np].end(),
			    intersect.begin(),intersect.end(),
			    inserter(nd,nd.begin()));
	
	_sol.update(n,intersect);
      }
      
      _change |= set_union(nd,(*_deltas)[np]);
      _deltas->update(n,nd);

      merge<set_union_func<unsigned int> >(n,np,_graph);
      condense_nodes(n,np,_varmap);

      //      _sol.update(n,S());
      _sol.clear(np); // save memory space
      _deltas->clear(np); // save memory space
    }
    _unifyqueue.clear();
  }
};

template<class G, class S>
void pearce_solve(G &g, 
		  std::vector<variable_info> &v,
		  store<S> &s,
		  checker_stats &stats,
		  bool f) {
  pearce_solver_common<G,S> sc(g,v,s,stats,f);
  sc.solve();
}

#endif
