#ifndef SOLVER_COMMON_HPP
#define SOLVER_COMMON_HPP

#include "weighted_diedge.hpp"
#include "bidirectional_adjacency_list.hpp"
#include "diedge.hpp"
#include "bitset.hpp"
#include "store.hpp"
#include "pcs.hpp"

// basicall this provides a number of functions
// which are in fact common to several 
// solvers ...

template<class G, class S>
class solver_common {
protected:
  G &_graph;
  std::vector<variable_info> &_varmap;
  std::vector<unsigned int> _unifyqueue;
  std::vector<diedge> _tqueue;
  store<S> &_sol;
  bool _full_type_checking;
  checker_stats &_stats;
  bidirectional_adjacency_list<> _transitive_edges;
public:
  solver_common(G &g, std::vector<variable_info> &v, store<S> &s, checker_stats &stats, bool f)
    : _graph(g), _varmap(v), _sol(s), _stats(stats), _full_type_checking(f),
      _transitive_edges(size(g)) {
  }
protected:
  virtual void process_trivial(unsigned int, unsigned int) = 0;
  virtual void new_edge(unsigned int, unsigned int, unsigned int) = 0;
  virtual bool unifynodes(unsigned int, unsigned int) = 0;

  void do_da_constraint(constraint const &c, 
			S const &delta) {
    // this function deals with constraints of the
    // form *x = &y
    
    unsigned int rhs(c.rhs().var());
    unsigned int lhs(c.lhs().var());
    unsigned int off(c.lhs().offset());

    typename S::const_iterator jend(delta.end());
    for(typename S::const_iterator j(delta.begin());
	j!=jend;++j) {
      unsigned int sj(*j);

      if(sj > 1) {
	// *x != NULL && *x != UNKNOWN
	if(_varmap[sj].end <= (sj+off)) {
	  // type error.
	  if(_full_type_checking) {
	    std::vector<constraint> cs;
	    cs.push_back(c);
	    cs.push_back(constraint(expression(SCALAR,lhs),
				    expression(ADDRESSOF,sj)));
	    throw internal_contradiction(cs);
	  }
	  // otherwise ignore
	} else {
	  unsigned int t(_varmap[sj+off].node);
	  process_trivial(t,rhs);
	}
      }
    }
  }

  void do_ds_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form *x = y
    unsigned int rhs(_varmap[c.rhs().var()].node);
    unsigned int lhs(c.lhs().var());
    unsigned int loff(c.lhs().offset());
    unsigned int roff(c.rhs().offset());

    typename S::const_iterator jend(delta.end());    
    for(typename S::const_iterator j(delta.begin());
	j!=jend;++j) {      
      unsigned int sj(*j);

      if(sj > 1) {
	// *x != NULL && *x != UNKNOWN
	if(_varmap[sj].end <= (sj+loff)) {
	  // type error
	  if(_full_type_checking) {
	    std::vector<constraint> cs;
	    cs.push_back(c);
	    cs.push_back(constraint(expression(SCALAR,lhs),
				    expression(ADDRESSOF,sj)));
	    throw internal_contradiction(cs);
	  }
	  // otherwise ignore
	} else {
	  unsigned int t(_varmap[sj+loff].node);
	  if(add_edge_helper(rhs,t,roff)) {
	    new_edge(rhs,t,roff);
	    // this is needed because the new_edge function
	    // might cause a cycle to be detected and
	    // collapsed
	    rhs = _varmap[c.rhs().var()].node;
	  }
	}
      }
    }
  }

  void do_sd_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form x = *y
    bool flag(false);
    unsigned int rhs(c.rhs().var());
    unsigned int lhs(_varmap[c.lhs().var()].node);
    unsigned int loff(c.lhs().offset());
    unsigned int roff(c.rhs().offset());

    typename S::const_iterator jend(delta.end());
    for(typename S::const_iterator j(delta.begin());
	j!=jend;++j) {
      unsigned int sj(*j);

      if(sj > 1) {
	// *x != NULL && *x != UNKNOWN
	if(_varmap[sj].end <= (sj+roff)) {
	  // type error
	  if(_full_type_checking) {
	    std::vector<constraint> cs;
	    cs.push_back(c);
	    cs.push_back(constraint(expression(SCALAR,rhs),
				    expression(ADDRESSOF,sj)));
	    throw internal_contradiction(cs);
	  }
	  // otherwise ignore
	} else {
	  unsigned int t(_varmap[sj+roff].node);
	  
	  if(add_edge_helper(t,lhs,0)) {
	    new_edge(t,lhs,0);
	    lhs = _varmap[c.lhs().var()].node;
	  }
	}
      } 
    }
  }

  // Note, I copy the constraint because, when using
  // the PKH cycle detector nodes can be merged and,
  // hence complicates merged, at any time ...
  void do_complex_constraint(constraint c,
			     S const &delta) {
    _stats.complex_count++;

    if(c.lhs().dereferenced()) {
      if(c.rhs().addressof()) {
	// *x=&y
	do_da_constraint(c,delta);
      } else {
	// *x=y
	do_ds_constraint(c,delta);
      }
    } else {
      // x=*y
      do_sd_constraint(c,delta);
    }
  }

  bool propagate(S &to, S const &from, unsigned int label) {
    _stats.union_count++;
    _stats.union_size += std::min(to.size(),from.size());

    if(label == 0) {
      return set_union(to,from);
    } else {
      S tmp(from);
      sol_set_add<S>(tmp,label,_varmap,_full_type_checking);
      return set_union(to,tmp);
    }
  }

  bool add_edge_helper(unsigned int t, unsigned int h, unsigned int s) {
    // helper function
    assert(_varmap[t].node == t && _varmap[h].node == h);
    diedge d(t,h);
    if(s != 0 || !valid_edge(d,_transitive_edges)) {
      // do not add transitive edges to graph
      bool r(add_edge(d,_graph));
      r |= get(d,_graph).insert(s).second;
      return r;
    } else {
      return false;
    }
  }
  
  bool process_unifyqueue(void) {
    bool r(false);
    // first, we must delete all transitive edges from the graph

    std::vector<diedge>::const_iterator jend(_tqueue.end());
    for(std::vector<diedge>::const_iterator j(_tqueue.begin());
	j!=jend;++j) {      

      if(add_edge(*j,_transitive_edges)) {
	_stats.red_count++;
	vector_set<unsigned int> &labels(get(*j,_graph));
	if(labels.size() == 1) {
	  assert(*labels.begin() == 0);
	  erase_edge(*j,_graph);
	} else {
	  assert(!labels.empty());
	  assert(*labels.begin() == 0);
	  labels.erase((unsigned int) 0);
	}
      }
    }

    _tqueue.clear();
    // all done!

    std::vector<unsigned int>::const_iterator iend(_unifyqueue.end());
    for(std::vector<unsigned int>::const_iterator i(_unifyqueue.begin());
	i!=iend;++i) {
      unsigned int np(*i);
      unsigned int n(_varmap[np].node);
      merge<set_union_func<unsigned int> >(n,np,_graph);
      merge_simple(n,np,_transitive_edges);
      condense_nodes(n,np,_varmap);
      // hook for the various implementations to do things
      r |= unifynodes(n,np);
    }
    // Finally, clear the unifyqueue.
    _unifyqueue.clear();
    return r;
  }
};

#endif








