#ifndef WORKLIST_SOLVER_HPP
#define WORKLIST_SOLVER_HPP

#include <set>
#include <string>
#include <boost/tuple/tuple.hpp>
#include "diedge.hpp"
#include "pcs.hpp" 
#include "set_union.hpp"
#include "store.hpp"
#include "worklist_strategy.hpp"
#include "solver_common.hpp"
#include "cycle_detector.hpp"

template<class G, class S, class B>
class worklist_solver_common : public B {
private:
  worklist_strategy &_worklist;
  cycle_detector &_cycle_detector; 
public:
  worklist_solver_common(G &g, std::vector<variable_info> &v, 
			 store<S> &s, worklist_strategy &strategy,
			 cycle_detector &cd, 
			 checker_stats &stats,
			 bool ftc = false) 
    : B(g,v,s,stats,ftc), 
      _worklist(strategy), 
      _cycle_detector(cd)
  {
    //    assert(_sol.size() == size(_graph));
  }

  void solve(void) {
    // push all nodes onto the worklist
    for(unsigned int i=0;i!=size(_graph);++i) {
      typename G::const_siterator sb, se;
      //      boost::tie(sb,se) = successors(i,_graph);
      if(_varmap[i].node == i) {
	_worklist.push(i);
      }
    }

    while(!_worklist.empty()) {
      // now iterate until the worklist is empty
      while(!_worklist.empty()) {
	unsigned int c(_worklist.top()); // next
	assert(_varmap[c].node == c);
	_worklist.pop();
	//	enter_vertex2(c); 
	enter_vertex(c);
	_stats.visit_count++;
	
	// now propagate solution from c to all its successors
	typename G::const_siterator is, isend;
	for(boost::tie(is,isend) = successors(c,_graph);is!=isend;++is) {
	  unsigned int dest(_varmap[*is].node);
	  vector_set<unsigned int> &labels(*(is.base()->second()));
	  if(do_propagate(dest,labels)) { _worklist.push(dest);  }
	}
      }

#if 0
      for(unsigned int c=0;c!=_varmap.size();++c) {
	if(_varmap[c].node == c) {
	  finish_vertex(c);
	}
      }
#endif
      // process all complex constraints, generating a set
      // of new edges to add to the graph.
      for(unsigned int c=0;c!=_varmap.size();++c) {
	if(_varmap[c].node == c && !_varmap[c].complicates.empty()) {	  
	  S sol(_sol[c]);
	  vector_set<constraint> comps(_varmap[c].complicates);
	  vector_set<constraint>::const_iterator iend(comps.end());
	  for(vector_set<constraint>::const_iterator i(comps.begin());
	      i!=iend;++i) {	
	    do_complex_constraint(*i,sol);
	  }
	}
      }

      // check for any new cycles
      do_detect_cycles();	  	  
      _worklist.graph_change(_graph);
    }
  }       

private:
  void do_detect_cycles(void) {
    // do nout for time being
    _cycle_detector.finish(_unifyqueue);
    process_unifyqueue();
  }

  void new_edge(unsigned int tail, unsigned int head, unsigned int off) {   
    if(off == 0) {
      _cycle_detector.new_edge(tail,head,_unifyqueue);
      // this next statement is important if we're 
      // using a unit change cycle detector, such as PKH.
      process_unifyqueue();
      head = _varmap[head].node;
      tail = _varmap[tail].node;
      if(head == tail) { return; }
    }

    if(do_new_edge(tail,head,off)) {
      _worklist.push(head);
    }
  }

  bool unifynodes(unsigned int n, unsigned int np) {    
    assert(np < size(_graph));
    assert(n != np);

    bool r = internal_unify(n,np);
    _worklist.merge(n,np);
    if(r) { _worklist.push(n); }
    return r;
  }

  void process_trivial(unsigned int node, unsigned int val) {    
    if(do_process_trivial(node,val,true)) {
      _worklist.push(node);
    }	  
  }
};

template<class G, class S, class B>
void worklist_solve(G &g, 
		    std::vector<variable_info> &v,
		    store<S> &s,
		    worklist_strategy &w,
		    cycle_detector &cd, 
		    checker_stats &stats,
		    bool f) {
  worklist_solver_common<G,S,B> sc(g,v,s,w,cd,stats,f);
  sc.solve();
}

#endif







