#include <set>
#include <stdexcept>
#include "vector_set.hpp"
#include "bidirectional_adjacency_list.hpp"
#include "pcs.hpp"
#include "worklist_solver.hpp"
#include "old_worklist_solver.hpp"
#include "simple_worklist_base.hpp"
#include "difference_propagation_base.hpp"
#include "bitset.hpp"
#include "lrf_strategy.hpp"
#include "lifo_strategy.hpp"
#include "fifo_strategy.hpp"
#include "oto1_strategy.hpp"
#include "factored_store.hpp"
#include "dummy_cycle_detector.hpp"
#include "PKH_online_cycle_detector.hpp"
#include "MNR_online_cycle_detector.hpp"
#include "BMNR_online_cycle_detector.hpp"
#include "simple_cycle_detector.hpp"
#include "simple_store.hpp"
#include "factored_store.hpp"

using namespace std;

typedef bidirectional_adjacency_list<vector_set<unsigned int> >  graph_t;

template<class T>
store<vector_set<unsigned int> > *worklist_solver_main_template(vector<unsigned int> params,
								graph_t &g, 				     
								vector<variable_info> &varmap,
								checker_stats &stats,
								bool ftc) {  

  worklist_strategy *strategy = NULL;
  cycle_detector *detector = NULL;
  store<T> *sol = NULL;

  switch(params[1]) {
  case SIMPLE_STORE:
    sol = new simple_store<T>(size(g));
    break;    
  case FACTORED_STORE:
    sol = new factored_store<T>(size(g));
    break;    
  default:
    throw runtime_error("invalid store selected");
  }

  switch(params[3]) {
  case DFS_ONLINE_CYCLE_DETECTION:
    detector = new simple_cycle_detector(g,varmap);
    break;
  case PKH_CYCLE_DETECTION:
    detector = new PKH_online_cycle_detector(g,varmap);
    break;
  case MNR_CYCLE_DETECTION:
    detector = new MNR_online_cycle_detector(g,varmap);
    break;
  case BATCH_MNR_CYCLE_DETECTION:
    detector = new BMNR_online_cycle_detector(g,varmap);
    break;
  case NO_CYCLE_DETECTION:
  case STATIC_CYCLE_DETECTION:
  case FULL_VARIABLE_SUBST:
    detector = new dummy_cycle_detector();
    break;    
  default:
    throw runtime_error("invalid cycle detector specified!");
  }  

  switch(params[4]) {
  case LIFO:
    strategy = new lifo_strategy(g);
    break;    
  case FIFO:
    strategy = new fifo_strategy(g);
    break;    
  case LRF:
    strategy = new lrf_strategy(g);
    break;
  case OTO1:
    strategy = new oto1_strategy(g);
    break;
  default:
    throw runtime_error("invalid iteration order");
  }
  
  for(unsigned int i=0;i!=varmap.size();++i) {
    T tmp;
    tmp.insert(varmap[i].solution.begin(),varmap[i].solution.end());
    sol->update(i,tmp);
  }
  
  switch(params[5]) {
  case NEW_FORM_FACTOR:
    switch(params[6]) {
    case NO_DIFFERENCE_PROPAGATION:
      worklist_solve<graph_t,T,
	simple_worklist_base<graph_t,T> >(g,varmap,*sol,*strategy,*detector,
					  stats,ftc);
      break;
    case DIFFERENCE_PROPAGATION:
      worklist_solve<graph_t,T,
	difference_propagation_base<graph_t,T> >(g,varmap,*sol,*strategy,*detector,
						 stats,ftc);
      break;
    default:
      throw runtime_error("Invalid difference propagation parameter");
      break;
    }
    break;
  case OLD_FORM_FACTOR:
    switch(params[6]) {
    case NO_DIFFERENCE_PROPAGATION:
      old_worklist_solve<graph_t,T,
	simple_worklist_base<graph_t,T> >(g,varmap,*sol,*strategy,
					  *detector,stats,ftc);
      break;
    case DIFFERENCE_PROPAGATION:
      old_worklist_solve<graph_t,T,
	difference_propagation_base<graph_t,T> >(g,varmap,*sol,*strategy,
						 *detector,stats,ftc);
      break;
    default:
      throw runtime_error("Invalid difference propagation parameter");
      break;
    }
    break;
  default:
    throw runtime_error("invalid form factor");
  }

  // now, create the final solution and return it ...
  store<vector_set<unsigned int> > *r = new factored_store<vector_set<unsigned int> >(size(g));

  for(unsigned int i=0;i!=varmap.size();++i) {
    if(sol->is_stored(i)) {
      vector_set<unsigned int> tmp;
      tmp.insert((*sol)[i].begin(),(*sol)[i].end());
      r->update(i,tmp);
    }
  }

  delete sol;
  delete detector;
  delete strategy;

  return r;
}

store<vector_set<unsigned int> > *worklist_solver_main(vector<unsigned int> &params,
						       graph_t &g, 				     
						       vector<variable_info> &varmap,
						       checker_stats &stats,
						       bool ftc) {  

  if(params.size() != 7) {
    throw runtime_error("insufficient arguments to algorithm Worklist");
  }


  switch(params[2]) {
  case BITSET:
    return worklist_solver_main_template<bitset>(params,g,varmap,stats,ftc);
    break;
  case VECTOR_SET:
    return worklist_solver_main_template<vector_set<unsigned int> >(params,g,varmap,stats,ftc);
    break;
  case STLSET:
    return worklist_solver_main_template<std::set<unsigned int> >(params,g,varmap,stats,ftc);
    break;
  default:
    throw runtime_error("invalid set implementation");
  }
}




