#ifndef PKH_ONLINE_STRONGLY_CONNECTED_COMPONENTS
#define PKH_ONLINE_STRONGLY_CONNECTED_COMPONENTS

#include <vector>
#include <algorithm>  // for sort
#include <boost/tuple/tuple.hpp> // for tie
#include "topological_sort.hpp"
#include "diedge.hpp"

class PKH_online_strongly_connected_components {
private:
  typedef PKH_online_strongly_connected_components self;
  template<class G> friend void new_edge(diedge const &, 
					 G const &,
					 std::vector<unsigned int> &,
					 self &);
  friend void reassign(unsigned int, 
		       std::vector<unsigned int> &, 
		       std::vector<unsigned int> &, 
		       std::vector<unsigned int> &, 
		       self &);
  template<class G> friend void dfs_f(unsigned int, 
				      unsigned int,
				      std::vector<unsigned int> &, 
				      G const &,
				      self &);
  template<class G> friend void dfs_b(unsigned int, 
				      unsigned int,
				      std::vector<unsigned int> &, 
				      G const &,
				      self &);
private:
  std::vector<unsigned int> _c2i; 
  std::vector<unsigned int> _i2c; // actually not needed!!!
  std::vector<unsigned int> _data;
public:
  PKH_online_strongly_connected_components(unsigned int n) {
    _c2i.resize(n);
    _i2c.resize(n);
    _data.resize(2*n,false);
    
    for(unsigned int i=0;i!=n;++i) {
      _c2i[i] = i;
      _i2c[i] = i;
    }
  }

  template<class G>
  PKH_online_strongly_connected_components(G const &g)
    : _c2i(size(g)), _i2c(size(g)), _data(2*size(g),false) {
    // pre: graph is a DAG
    std::vector<unsigned int> order(size(g));
    // topologically sort graph.
    topological_sort(g,order);
    unsigned int index(0);
    // remember topological sort is produced
    // in reverse order.
    std::vector<unsigned int>::reverse_iterator iend(order.rend());
    for(std::vector<unsigned int>::reverse_iterator i(order.rbegin());
	i!=iend;++i,++index) {
      _c2i[*i]=index;
      _i2c[index]=*i;
    }
  }
};

#define VISITED(x) self._data[(x<<1)]
#define IN_COMPONENT(x) self._data[(x<<1)+1]
#define C2I(x) self._c2i[x]
#define I2C(x) self._i2c[x]

template<class G>
void new_edge(diedge const &e, 
	      G const &graph,
	      std::vector<unsigned int> &scc,
	      PKH_online_strongly_connected_components &self) {

  unsigned int lb(self._c2i[e.head]);
  unsigned int ub(self._c2i[e.tail]);	

  if(lb < ub) {
    // in this case we are forced
    // to update the topological
    // order
    static std::vector<unsigned int> reachable,reaching;
    VISITED(e.tail)=true;
    IN_COMPONENT(e.tail) = true;
    // should be able to replace these with
    // generic call to depth_first_visit
    dfs_f(e.head,ub,reachable,graph,self);

    // this is annoying!
    std::vector<unsigned int>::const_iterator iend(reachable.end());
    for(std::vector<unsigned int>::const_iterator i(reachable.begin());
	i!=iend;++i) {
      VISITED(I2C(*i)) = false;
    }

    dfs_b(e.tail,lb,reaching,graph,self);

    std::sort(reachable.begin(),reachable.end());
    std::sort(reaching.begin(),reaching.end());	  

    reassign(e.tail, reachable, reaching, scc, self);

    reachable.clear();
    reaching.clear();
  }  
}

template<class G>
void dfs_f(unsigned int n, unsigned int ub, std::vector<unsigned int> &reachable, 
	   G const &graph,
	   PKH_online_strongly_connected_components &self) {	   
  assert(n < size(graph));
  VISITED(n)=true;    
  reachable.push_back(C2I(n));
  
  typename G::const_siterator i,iend;
  for(boost::tie(i,iend) = successors(n,graph);i!=iend;++i) {
    assert(C2I(*i) >= C2I(n) || C2I(n) == ub);
    if(C2I(*i) <= ub) {
      if(!VISITED(*i)) { dfs_f(*i,ub,reachable,graph,self); }
      IN_COMPONENT(n) = IN_COMPONENT(n) | IN_COMPONENT(*i);
    }
  }
}

template<class G>
void dfs_b(unsigned int root, unsigned int lb, std::vector<unsigned int> &reaching,
	   G const &graph,
	   PKH_online_strongly_connected_components &self) {

  std::vector<unsigned int> worklist;    
  reaching.push_back(C2I(root));
  worklist.push_back(root);
  VISITED(root)=true;
  
  while(worklist.size() > 0) { 
    unsigned int n(worklist.back());
    worklist.pop_back();
    
    if(!IN_COMPONENT(n)) {
      reaching.push_back(C2I(n));
    }

    typename G::const_piterator i,iend;
    for(boost::tie(i,iend) = predecessors(n,graph);i!=iend;++i) {
      assert(C2I(*i) <= C2I(n) || C2I(n) == lb);
      if(C2I(*i) > lb && !VISITED(*i)) {
	VISITED(*i)=true;
	worklist.push_back(*i);
      }
    }
  }
}

#undef IN_COMPONENT
#undef VISITED
#undef C2I
#undef I2C

#endif




















