// A fast solver of constraints containing indirections
// ====================================================
// Author David J. Pearce, Feb 2003
// 
// Notes:
//
// This algorithm is a reformulation of the algorithm
// by Nevin Heintze and Olivier Tardieu which was presented 
// in their paper:
//
// "Ultra-fast Aliasing Analysis using CLA: A Million Lines
//  of C Code in a Second"
//
// This paper can be found online at www.cs.bell-labs.com/who/nch/
//
// I have reformulated the algorithm so that it now uses a
// cycle detection scheme which is based on Tarjan's algorithm
// for detecting scc's. 
//
// The use of the STACK instead of the path relation comes
// from the paper:
//
// "On Finding the Strongly Connected Components in a Directed
//  Graph", Esko Nuutila, Eljas Soisalon-Soininen, 1994

#ifndef HEINTZE_TARDIEU_SOLVER_HPP
#define HEINTZE_TARDIEU_SOLVER_HPP

#include <boost/tuple/tuple.hpp>
#include "set_union.hpp"
#include "filter_graph.hpp"
#include "find_strong_components.hpp"
#include "pcs.hpp"

#define HT_VISIT_ALL 0

class ht_scc_visitor {
private:
  std::vector<unsigned int> &_order;
  std::vector<unsigned int> &_queue;
public:
  ht_scc_visitor(std::vector<unsigned int> &o,
		 std::vector<unsigned int> &q) 
    : _order(o), _queue(q) {
  }
  inline void start_vertex(unsigned int n) {}
  inline void finished_vertex(unsigned int n) {
    _order.push_back(n);
  }
  inline void traversing_edge(unsigned int t, unsigned int h) {}
  inline void finished_edge(unsigned int t, unsigned int h) {}
  inline void skipping_edge(unsigned int t, unsigned int h) {}
  inline void unify(unsigned int n, unsigned int w) {
    _queue.push_back(w);
  }
};

template<class G, class S>
class heintze_tardieu_common {
private:
  // the visited vector corresponds to the cache in the
  // Hetinze paper. We can flush the cache by marking all
  // nodes as unvisited.
  std::vector<bool> _visited;
  std::vector<bool> _cached;
  // the following three vectors are used by the 
  // update_solution function to identify strongly
  // connected components.
  std::vector<unsigned int> _vindex;  // visitation index
  std::vector<bool> _in_component; 
  std::vector<unsigned int> _stack; 
  unsigned int _index;
  // the remainder contain the solution and other bits n bobs
  G &_graph;                           // the constraint graph!
  std::vector<variable_info> &_varmap;
  std::vector<S> &_sol;
  bool _full_type_checking;                           // full type checking ?
  bool _cycle_detection;
  checker_stats &_stats;
public:
  heintze_tardieu_common(G &g, std::vector<variable_info> &v,
			 std::vector<S> &s, checker_stats &stats,
			 bool f, bool c) 
    : _visited(size(g),false), _cached(size(g),false), 
			   _vindex(size(g)), _in_component(size(g),false), 
			   _graph(g), _varmap(v), 
			   _sol(s), _stats(stats), 
			   _full_type_checking(f),
			   _cycle_detection(c) {    
  }
  
  void solve(void) {
    bool change;
    std::vector<unsigned int> unifyqueue;

    do {
      change = false;
      
      // this operation corresponds to flushing the cache
      // in the paper.
      
      if(_cycle_detection) { fill(_visited.begin(),_visited.end(),false); }
      fill(_cached.begin(),_cached.end(),false);
      
      // this loop goes through each variable p which 
      // has a constraint of the form *p = y, y=*p
      // *p=&y
    
      for(unsigned int i=0;i!=_varmap.size();++i) {
	
	// this call to update_solution is equivalent
	// to the getLvals call in the paper. essentially,
	// in this formulation of their algorithm no cache is
	// used. instead we maintain a bit vector which indicates
	// which solutions have been updated on this iteration.	

#if HT_VISIT_ALL == 0 
	if(!_varmap[i].complicates.empty()) {
#else 
	  if(_varmap[i].node == i) {
#endif
	  if(!_cached[i]) {
	    _index=0;	  
	    // you'll notice that in the paper, cycle_detection
	    // and update_solution are actually just one function.
	    // however, with the presence of labelled edges which 
	    // means we can only collapse cycles with a zero shortest
	    // path between any two nodes, I have split it into
	    // two seperate functions ...
	    if(_cycle_detection) {
	      _index=0;
	      cycle_detect(i,unifyqueue);
	      change |= process_unifyqueue(unifyqueue);
	      unifyqueue.clear();
	    }	      
	    change |= update_solution(i);
	  }

	  assert(_varmap[i].node == i);
	  
	  // ASSERT: i is always a component here. this is because
	  //         the search begins from i and, hence, any 
	  //         new compomnent involving i must be named i

#if HT_VISIT_ALL == 1
	if(!_varmap[i].complicates.empty()) {
#endif	  
	  change |= update_graph(i,_varmap[i].complicates, _sol[i].delta());	  
	}
#if HT_VISIT_ALL == 1
	  }
#endif
      }
    } while(change);    
    // at this point we can compute the entire solution using a 
    // topological traversal of the strong components

    // must prevent nodes which  from
    // being visited. Otherwise they will be
    // regarded as components and remarked
    // such that varmap[x].node = x.

#if HT_VISIT_ALL == 0
#if 0 // added to test pearce_ht_solver
    for(unsigned int i=0;i!=_varmap.size();++i) {
      if(_varmap[i].node != i) {
	_visited[i]=true;
      } else {
	_visited[i]=false;
      }
    }
    // first step compute the strongly connected components
    // of the filtered graph.
    std::vector<unsigned int> order;
    filter_graph<G,zero_edge_select> fg(_graph);
    n2c_wrapper n2c(_varmap); // uggly
    find_strong_components(fg,n2c,_visited,
                           ht_scc_visitor(order,unifyqueue));
    // process the unify queue.
    process_unifyqueue(unifyqueue);
    // now, process components in topological order
    // note there is possible problem with positive weight cycles.
    do {
      change = false;
      for(std::vector<unsigned int>::reverse_iterator i(order.rbegin());
	  i!=order.rend();++i) {
	assert(_varmap[*i].node == *i);
	typename G::const_siterator j,jend;
	for(boost::tie(j,jend) = successors(*i,_graph);
	    j!=jend;++j) {	
	  std::set<unsigned int> &labels(*(j.base()->second()));
	  for(std::set<unsigned int>::const_iterator k(labels.begin());
	      k!=labels.end();++k) {
	    change |= propagate(*i,*j,*k);
	  }
	}
      }
    } while(change);
#endif
#endif
  }
protected:
  
  bool update_graph(unsigned int var, std::vector<constraint> const &constraints,
		    typename S::delta_type const &delta) {
    // the point of this function is to go through each complex constraint that
    // var is involved in and update the graph where necessary.
    bool r(false);
    
    std::vector<constraint>::const_iterator iend(constraints.end());
    for(std::vector<constraint>::const_iterator i(constraints.begin());
	i!=iend;++i) {      
      _stats.complex_count++;
      constraint const &c(*i);

      if(c.lhs().dereferenced()) {
	// *p = y, *p = &y
	if(c.rhs().addressof()) {
	  // *p = &y
	  r |= do_da_constraint(c,delta);
	} else {
	  // *p = y
	  r |= do_ds_constraint(c,delta);
	}
      } else {
	// y = *p
	r |= do_sd_constraint(c,delta);
      }
    }
    
    return r;
  }
  
  bool do_da_constraint(constraint const &c,
			typename S::delta_type const &delta) {
    bool r(false);
    unsigned int rhs(c.rhs().var());
    unsigned int lhs(c.lhs().var());
    unsigned int off(c.lhs().offset());
    
    typename S::delta_type::const_iterator iend(delta.end());
    for(typename S::delta_type::const_iterator i(delta.begin());
	i!=iend;++i) {
      unsigned int si(*i);
      if(si > 1) {
	// *x != NULL && *x != UNKNOWN
	if(_varmap[si].end <= (si+off)) {
	  // type error.
	  if(_full_type_checking) {
	    std::vector<constraint> cs;
	    cs.push_back(c);
	    cs.push_back(constraint(expression(SCALAR,lhs),
				    expression(ADDRESSOF,si)));
	    throw internal_contradiction(cs);
	  }
	  // otherwise ignore
	} else {
	  unsigned int t(_varmap[si+off].node);
	  r |= _sol[t].insert(rhs);
	}
      }
    }    

    return r;
  }
  
  bool do_ds_constraint(constraint const &c,
			typename S::delta_type const &delta) {
    bool r(false);
    unsigned int rhs(_varmap[c.rhs().var()].node);
    unsigned int lhs(c.lhs().var());
    unsigned int loff(c.lhs().offset());
    unsigned int roff(c.rhs().offset());

    typename S::delta_type::const_iterator iend(delta.end());
    for(typename S::delta_type::const_iterator i(delta.begin());
	i!=iend;++i) {
      unsigned int si(*i);

      if(si > 1) {
	// *x != NULL && *x != UNKNOWN
	if(_varmap[si].end <= (si+loff)) {
	  // type error
	  if(_full_type_checking) {
	    std::vector<constraint> cs;
	    cs.push_back(c);
	    cs.push_back(constraint(expression(SCALAR,lhs),
				    expression(ADDRESSOF,si)));
	    throw internal_contradiction(cs);
	  }
	  // otherwise ignore
	} else {
	  unsigned int t(_varmap[si+loff].node);
	  r |= add_edge_helper(rhs,t,roff);
	}
      }
    }
    
    return r;
  }
  
  bool do_sd_constraint(constraint const &c,
			typename S::delta_type const &delta) {
    bool r(false);
    unsigned int rhs(c.rhs().var());
    unsigned int lhs(_varmap[c.lhs().var()].node);
    unsigned int loff(c.lhs().offset());
    unsigned int roff(c.rhs().offset());
    
    typename S::delta_type::const_iterator iend(delta.end());
    for(typename S::delta_type::const_iterator i(delta.begin());
	i!=iend;++i) {
      unsigned int si(*i);
      if(si > 1) {
	// *x != NULL && *x != UNKNOWN
	if(_varmap[si].end <= (si+roff)) {
	  // type error
	  if(_full_type_checking) {
	    std::vector<constraint> cs;
	    cs.push_back(c);
	    cs.push_back(constraint(expression(SCALAR,rhs),
				    expression(ADDRESSOF,si)));
	    throw internal_contradiction(cs);
	  }
	  // otherwise ignore
	} else {
	  unsigned int t(_varmap[si+roff].node);
	  r |= add_edge_helper(t,lhs,0);
	}
      }
    }

    return r;
  }
  
  bool update_solution(unsigned int n) {
    bool r(false);
    _cached[n]=true;
    _stats.visit_count++;
    // now process all predecessors
    typename G::const_piterator i,iend;
    for(boost::tie(i,iend) = predecessors(n,_graph);i!=iend;++i) {
      unsigned int w(*i);
      if(!_cached[w]) { r |= update_solution(w); }

      std::set<unsigned int> &labels(*(i.base()->second()));    
      for(std::set<unsigned int>::const_iterator j(labels.begin());
	  j!=labels.end();++j) {
	// propagate along edge w->n, with label *j.
	r |= propagate(w,n,*j);
      }
    }

    return r;
  }

  void cycle_detect(unsigned int n, 
		    std::vector<unsigned int> &unifyqueue) {
    assert(_varmap[n].node == n);
    _visited[n]=true;
    _in_component[n]=false;
    _vindex[n]=_index++;
    // now process all successors
    typename G::const_piterator i,iend;
    for(boost::tie(i,iend) = predecessors(n,_graph);i!=iend;++i) {
      unsigned int w(*i);
      std::set<unsigned int> &labels(*(i.base()->second()));
      if(!labels.empty() && *labels.begin() == 0) {
	// only traverse edges with zero label
	if(!_visited[w]) { 
	  cycle_detect(w,unifyqueue); 
	}
	if(!_in_component[w]) {
	  unsigned int t(_varmap[w].node);
	  if(_vindex[t] < _vindex[_varmap[n].node]) {
	    _varmap[n].node = t;
	  }
	}
      }
    }
    
    // at this point we must check to see
    // if any components have been identified
    if(_varmap[n].node == n) {
      unsigned int t(_vindex[n]);
      _in_component[n] = true;
      while(!_stack.empty() && t < _vindex[_stack.back()]) {
       unsigned int w(_stack.back());
       _varmap[w].node = n;
       _in_component[w] = true;
       unifyqueue.push_back(w);
       _stack.pop_back(); 
     }
   } else {
     _stack.push_back(n);
   }
  }

  bool propagate(unsigned int from, unsigned int to, unsigned int label) {
    assert(from < _varmap.size() && to < _varmap.size());
    if(label == 0) {
      return _sol[to].merge(_sol[from]);
    } else {
      bool flag(false);      
      typename S::delta_type delta(_sol[from].delta());
      typename S::delta_type::const_iterator iend(delta.end());
      for(typename S::delta_type::const_iterator i(delta.begin());
	  i!=iend;++i) {
	unsigned int r(*i);
	// segfaults at next line when using vector_set.
	if(_varmap[r].end <= (r+label)) {
	  if(_full_type_checking) {
	    std::vector<constraint> cs;
	    throw internal_contradiction(cs);
	  }
	  // otherwise ignore 
	} else {
	  flag |= _sol[to].insert(r + label);
	}
      }
      return flag;
    }
  }

  bool process_unifyqueue(std::vector<unsigned int> const &unifyqueue) {
    bool r(false);
    std::vector<unsigned int>::const_iterator iend(unifyqueue.end());
    for(std::vector<unsigned int>::const_iterator i(unifyqueue.begin());
	i!=iend;++i) {
      unsigned int np(*i);
      unsigned int n(_varmap[np].node);
      r |= _sol[n].merge(_sol[np]);
      merge<set_union_func<unsigned int> >(n,np,_graph);
      condense_nodes(n,np,_varmap);
      // I use the "swap trick" here to make
      // sure that memory is freed.
      S empty;
      _sol[np].swap(empty); // save memory space     
    }
    return r;
  }
  
  bool add_edge_helper(unsigned int t, unsigned int h, unsigned int s) {
    // helper function
    diedge d(t,h);
    bool r(add_edge(d,_graph));
    r |= get(d,_graph).insert(s).second;
    return r;
  }
};

template<class G, class S>
void heintze_tardieu_solve(G &g, 
			   std::vector<variable_info> &v,
			   std::vector<S> &s,
			   checker_stats &cs, 
			   bool f, bool c) {

  heintze_tardieu_common<G,S> sc(g,v,s,cs,f,c);
  sc.solve();
}

#endif





