#ifndef PKH03D_WORKLIST_SOLVER_HPP
#define PKH03D_WORKLIST_SOLVER_HPP

#include <set>
#include <string>
#include <boost/tuple/tuple.hpp>
#include "../../worklist_strategy.hpp"
#include "../../topological_sort.hpp"

#include "../../solver_common.hpp"
#include "../../lrf_strategy.hpp"
#include "../../set_union.hpp"
#include "../../diedge.hpp"
#include "../../store.hpp"
#include "../../pcs.hpp" 

template<class G, class S>
class PKH03d_worklist_algorithm {
private:
  G &_graph;
  std::vector<variable_info> &_varmap;
  store<S> &_sol;
  lrf_strategy _worklist;

  // this final property is just used to obtain
  // useful information on the algorithms operation
  checker_stats &_stats;
public:
  PKH03d_worklist_algorithm(G &g, std::vector<variable_info> &v, 
			    store<S> &s,
			    checker_stats &stats) :
    _graph(g),
    _varmap(v),
    _sol(s),
    _worklist(g),
    _stats(stats) {
    //    assert(_sol.size() == size(_graph));
  }

  void solve(void) {
    for(unsigned int i=0;i!=size(_graph);++i) {
      if(_varmap[i].node == i) { _worklist.push(i); }
    }

    while(!_worklist.empty()) {     
      unsigned int n(_worklist.top());
      _worklist.pop();	
      _stats.visit_count++;
      // process nodes in topological order
      S solution(_sol[n]);	    
      // process complex constraints      
      vector_set<constraint> complicates(_varmap[n].complicates);
      for(vector_set<constraint>::const_iterator j(complicates.begin());
	  j!=complicates.end();++j) {      
	do_complex_constraint(*j,solution);	  
      }

      // propagate solution to all successors
      typename G::const_siterator w, wend;
      for(boost::tie(w,wend) = successors(n,_graph);w!=wend;++w) {
	S tmp(_sol[*w]);
	
	if(int_set_union(tmp,solution)) { 
	  _worklist.push(*w);
	  _sol.update(*w,tmp); 
	}
      }
    }   
  }
  
private:
  void do_da_constraint(constraint const &c, 
			S const &delta) {
    // this function deals with constraints of the
    // form *x = &y
    unsigned int rhs(c.rhs().var());
    unsigned int off(c.lhs().offset());

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {
      if(type_safe(*j,off)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+off].node);
	S tmp(_sol[t]);
	
	if(tmp.insert(rhs).second) { 
	  _sol.update(t,tmp); 
	  _worklist.push(t);
	}
      }
    }
  }

  void do_ds_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form *x = y
    unsigned int rhs(_varmap[c.rhs().var()].node);
    unsigned int loff(c.lhs().offset());    

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {      
      if(type_safe(*j,loff)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+loff].node);
	if(t != rhs && add_edge(diedge(rhs,t),_graph)) {
	  _stats.new_edge_count++;
	  _worklist.push(rhs);
	}
      }
    }
  }

  void do_sd_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form x = *y
    unsigned int lhs(_varmap[c.lhs().var()].node);
    unsigned int roff(c.rhs().offset());
    bool flag(false);
    
    for(typename S::const_iterator j(delta.begin()); j!=delta.end();++j) {
      if(type_safe(*j,roff)) {
	// *j != NULL && *j != UNKNOWN
	unsigned int t(_varmap[*j+roff].node);       
	if(t != lhs && add_edge(diedge(t,lhs),_graph)) {
	  _stats.new_edge_count++;
	  _worklist.push(t);
	}
      }
    }
  }

  void do_complex_constraint(constraint c,
			     S const &delta) {
    _stats.complex_count++;

    if(c.lhs().dereferenced()) {
      if(c.rhs().addressof()) {
	// *x=&y
	do_da_constraint(c,delta);
      } else {
	// *x=y
	do_ds_constraint(c,delta);
      }
    } else {
      // x=*y
      do_sd_constraint(c,delta);
    }
  }

  bool type_safe(unsigned int n, unsigned int o) {
    return n > 1 && (n+o) < _varmap[n].end;
  }
  
  inline bool int_set_union(S &to, S const &from) {
    _stats.union_count++;
    _stats.union_size += to.size();
    _stats.union_size += from.size();
    return set_union(to,from);
  }
  };
  
template<class G, class S>
void PKH03d_worklist_solve(G &g, 
			   std::vector<variable_info> &v,
			   store<S> &s,
			   checker_stats &cs) {

  PKH03d_worklist_algorithm<G,S> sc(g,v,s,cs);
  sc.solve();
}

#endif







