#ifndef PKH03E_WORKLIST_SOLVER_HPP
#define PKH03E_WORKLIST_SOLVER_HPP

#include <set>
#include <string>
#include <stdexcept>
#include <boost/tuple/tuple.hpp>
#include "../../worklist_strategy.hpp"
#include "../../topological_sort.hpp"

#include "../../solver_common.hpp"
#include "../../set_union.hpp"
#include "../../diedge.hpp"
#include "../../store.hpp"
#include "../../pcs.hpp" 

#if __GNUC__ >= 3
#include <ext/hash_set>
#define HASH_SET __gnu_cxx::hash_set
#else 
// everything else for the moment.
#include <functional>
#include <hash_set>
#define HASH_SET std::hash_set
#endif

template<class G, class S>
class PKH03e_worklist_algorithm {
private:
  struct diedge_hash {
  public:
    size_t operator()(diedge const &d) const {
      return d.head+d.tail;
    }
  };

  G &_graph;
  std::vector<variable_info> &_varmap;
  store<S> &_sol;  
  bool _change;
  // this holds the (topological) iteration order
  // in reverse
  std::vector<unsigned int> _order;      
  std::vector<unsigned int> _n2i;      
  //  HASH_SET<diedge,diedge_hash> _transitive_edges;
  G _transitive_edges;

  // the final few vectors are used
  // by the algorithm for detecting
  // strongly connected components.
  std::vector<unsigned int> _vindex; 
  std::vector<bool> _in_component; 
  std::vector<unsigned int> _stack; 
  std::vector<unsigned int> _unifyqueue;
  std::vector<bool> _visited;
  std::vector<unsigned int> _dfs_visited;
  std::vector<unsigned int> _end;
  unsigned int _index; 

  // this final property is just used to obtain
  // useful information on the algorithms operation
  checker_stats &_stats;
public:
  PKH03e_worklist_algorithm(G &g, std::vector<variable_info> &v, 
			    store<S> &s,
			    checker_stats &stats) :
    _graph(g),
    _varmap(v),
    _sol(s),
    _order(size(g)),
    _n2i(size(g)),
    _transitive_edges(size(g)),
    _vindex(size(g)), 
    _visited(size(g),false),
    _dfs_visited(size(g),size(g)),
    _in_component(size(g),true), 
    _end(size(g)),    
    _stats(stats) {
    //    assert(_sol.size() == size(_graph));
  }

  void solve(void) {
    _change=true;
    while(_change) {
      _change=false;
      _stats.iter_count++;

      // detect strongly connected components
      _index = 0;
      _order.clear(); 
      for(unsigned int i=0;i!=size(_graph);++i) {
	if(!_visited[i] && _varmap[i].node == i) { visit(i); }
      }
      fill(_visited.begin(),_visited.end(),false);
      process_unifyqueue(); // collapse cycles

      /*      _index = 0;
      for(unsigned int i=0;i!=size(_graph);++i) {
	if(!_visited[i] && _varmap[i].node == i) { reduct_visit(i); }
      }
      */
      fill(_visited.begin(),_visited.end(),false);
      fill(_dfs_visited.begin(),_dfs_visited.end(),size(_graph));
      
      for(std::vector<unsigned int>::reverse_iterator i(_order.rbegin());
	i!=_order.rend();++i) {
	if(_varmap[*i].node == *i && !_varmap[*i].complicates.empty()) {
	  S solution;
	  compute_solution(*i,*i,solution);
	  // process complex constraints
	  vector_set<constraint> complicates(_varmap[*i].complicates);
	  for(vector_set<constraint>::const_iterator j(complicates.begin());
	      j!=complicates.end();++j) {      
	    do_complex_constraint(*j,solution);	  
	  }
	}
      }
      fill(_dfs_visited.begin(),_dfs_visited.end(),size(_graph));
    }     

    // at this point we compute the entire solution using a 
    // topological traversal of the (acyclic)     

    for(std::vector<unsigned int>::reverse_iterator i(_order.rbegin());
	i!=_order.rend();++i) {
      if(_varmap[*i].node == *i) {
	_stats.visit_count++;
	// proceed through predecessors of i
	S sol(_sol[*i]);
	bool flag(false);
	typename G::const_piterator j,jend;
	for(boost::tie(j,jend) = predecessors(*i,_graph);j!=jend;++j) {
	  flag |= int_set_union(sol,_sol[*j]);
	}
	// write the final solution for *i
	if(flag) { _sol.update(*i,sol); }
      }
    }
  }

private:
  void visit(unsigned int n) {
    assert(_varmap[n].node == n);
    _visited[n]=true;
    _in_component[n]=false;
    _vindex[n]=_index++;
   
    // visit all successors
    typename G::const_siterator i,iend;
    for(boost::tie(i,iend) = successors(n,_graph);i!=iend;++i) {
      unsigned int w(*i);
      if(!_visited[w]) { visit(w); }
      if(!_in_component[w]) {
	unsigned int t(_varmap[w].node);
	if(_vindex[t] < _vindex[_varmap[n].node]) {
	  _varmap[n].node = t;
	}
      }
    }
    
    // at this point we must check to see
    // if any components have been identified
    if(_varmap[n].node == n) {
      _n2i[n]=_order.size();
      _order.push_back(n);
      unsigned int t(_vindex[n]);
      _in_component[n] = true;
      while(!_stack.empty() && t < _vindex[_stack.back()]) {
	unsigned int w(_stack.back());
	_varmap[w].node = n;
	_in_component[w] = true;
	_unifyqueue.push_back(w);
	_stack.pop_back(); 
      }
    } else {
      _stack.push_back(n);
    }
  }
		
  void dfs_visit(unsigned int n,			  
		 unsigned int search_no) {
    // this is bog standard DFS algorithm.
    _dfs_visited[n] = search_no;
    _vindex[n] = _index++;
    if(_index == 0) { throw std::runtime_error("integer overflow"); }

    typename G::const_siterator i,iend;
    for(boost::tie(i,iend) = successors(n,_graph);i!=iend;++i) {
      unsigned int w(*i);
      if(_dfs_visited[w] != search_no) {
	dfs_visit(w,search_no);      
      }
    }
  }

  void compute_solution(unsigned int n,			  
			unsigned int search_no,
			S &solution) {
    _stats.visit_count++;
    _dfs_visited[n] = search_no;
    set_union(solution,_sol[n]);    

    typename G::const_piterator i,iend;
    for(boost::tie(i,iend) = predecessors(n,_graph);i!=iend;++i) {
      unsigned int w(*i);
      if(_dfs_visited[w] != search_no) {
	compute_solution(w,search_no,solution);      
      }
    }
  }

  void reduct_visit(unsigned int n) {    
    _visited[n] = true;
    _vindex[n] = _index++;
    // just a saftey check ...
    if(_index == 0) { throw std::runtime_error("integer overflow"); }
    
    typedef std::vector<unsigned int> se_t;
    std::vector<unsigned int> sorted_out_edges; 
    sorted_out_edges.reserve(num_successors(n,_graph));
    
    typename G::const_siterator i,iend;
    for(boost::tie(i,iend) = successors(n,_graph);i!=iend;++i) {
      sorted_out_edges.push_back(*i);
    }
    
    sort(sorted_out_edges.begin(),sorted_out_edges.end(),priority_comp<std::vector<unsigned int> >(_n2i));
    _end[n]=_index - 1;
    
    for(se_t::reverse_iterator j(sorted_out_edges.rbegin());
	j!=sorted_out_edges.rend();++j) {
      unsigned int w(*j);
      if(!_visited[w]) { reduct_visit(w);      
      } else {
	if(_vindex[n] < _vindex[w] &&
	   _vindex[w] <= _end[n]) {
	  // bingo, transitive edge discovered
	  erase_edge(diedge(n,w),_graph);
	  add_edge(diedge(n,w),_transitive_edges);
	} else {
	  dfs_visit(w,n);
	}
      }
      _end[n]=_index-1;
    }  
  }

  void do_da_constraint(constraint const &c, 
			S const &delta) {
    // this function deals with constraints of the
    // form *x = &y
    unsigned int rhs(c.rhs().var());
    unsigned int off(c.lhs().offset());

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {
      if(type_safe(*j,off)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+off].node);
	S tmp(_sol[t]);
	
	if(tmp.insert(rhs).second) { 
	  _sol.update(t,tmp);
	  _change=true;
	}
      }
    }
  }

  void do_ds_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form *x = y
    unsigned int rhs(_varmap[c.rhs().var()].node);
    unsigned int loff(c.lhs().offset());

    for(typename S::const_iterator j(delta.begin());j!=delta.end();++j) {      
      if(type_safe(*j,loff)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+loff].node);
	if(t != rhs && int_add_edge(rhs,t)) {
	  _stats.new_edge_count++;
	  _change=true;
	}
      }
    }
  }

  void do_sd_constraint(constraint const &c,
			S const &delta) {
    // this function deals with constraints of the
    // form x = *y
    unsigned int lhs(_varmap[c.lhs().var()].node);
    unsigned int roff(c.rhs().offset());

    for(typename S::const_iterator j(delta.begin()); j!=delta.end();++j) {
      if(type_safe(*j,roff)) {
	// *j != NULL && *j != UNKNOWN
	unsigned int t(_varmap[*j+roff].node);       
	if(t != lhs && int_add_edge(t,lhs)) {
	  _stats.new_edge_count++;
	  _change=true;
	}
      } 
    }
  }

  void do_complex_constraint(constraint c,
			     S const &delta) {
    _stats.complex_count++;

    if(c.lhs().dereferenced()) {
      if(c.rhs().addressof()) {
	// *x=&y
	do_da_constraint(c,delta);
      } else {
	// *x=y
	do_ds_constraint(c,delta);
      }
    } else {
      // x=*y
      do_sd_constraint(c,delta);
    }
  }

  void process_unifyqueue(void) { 
    S tmp;
    std::vector<unsigned int>::iterator j(_unifyqueue.begin());
    std::vector<unsigned int>::iterator jend(_unifyqueue.end());
    while(j != jend) {
      unsigned int n(_varmap[*j].node);
      
      int_set_union(tmp,_sol[*j]);	
      merge_simple(n,*j,_graph);
      merge_simple(n,*j,_transitive_edges);
      condense_nodes(n,*j,_varmap);
      _sol.clear(*j);
      
      ++j;
      // if we've finished updating the solution of a component
      if(j == jend || _varmap[*j].node != n) {
	S const &lsol(_sol[n]);
	
	if(int_set_union(tmp,lsol) || lsol.size() < tmp.size()) { 
	  _sol.update(n,tmp); 
	}
	tmp.clear();
	// eliminate self loops
	erase_edge(diedge(n,n),_graph);
      }	  
    }
    _unifyqueue.clear(); // empty the queue now!
  } 

  bool redundant_constraint(unsigned int n, constraint const &c) {	    
    // the point of this function is to reduce the number of
    // complex constraints floating around.
    if(c.rhs().dereferenced()) {	     
      assert(c.rhs().var() == n);
      unsigned int lhs(c.lhs().var());
      // a = *n
      if(_varmap[lhs].node != lhs) {
	constraint tmp(expression(SCALAR,_varmap[lhs].node,0),c.rhs());
	_varmap[n].complicates.erase(c);
	if(!_varmap[n].complicates.insert(tmp).second) {
	  return true;
	}
      }
    } else {
      assert(c.lhs().var() == n);
      if(!c.rhs().addressof()) {
	unsigned int rhs(c.rhs().var());
	if(_varmap[rhs].node != rhs) {
	  // *n = a
	  constraint tmp(c.lhs(),expression(SCALAR,_varmap[rhs].node,c.rhs().offset()));
	  _varmap[n].complicates.erase(c);
	  if(!_varmap[n].complicates.insert(tmp).second) {
	    return true;
	  }
	}
      }
    }
    return false;
  }

  bool int_add_edge(unsigned int t, unsigned int h) {
    assert(_varmap[t].node == t);
    assert(_varmap[h].node == h);
    diedge e(t,h);
    if(!valid_edge(e,_transitive_edges)) {
      return add_edge(e,_graph);
    } else {
      return false;
    }
  }

  bool type_safe(unsigned int n, unsigned int o) {
    return n > 1 && (n+o) < _varmap[n].end;
  }

  inline bool int_set_union(S &to, S const &from) {
    _stats.union_count++;
    _stats.union_size += to.size();
    _stats.union_size += from.size();
    return set_union(to,from);
  }
};

template<class G, class S>
void PKH03e_worklist_solve(G &g, 
			   std::vector<variable_info> &v,
			   store<S> &s,
			   checker_stats &cs) {

  PKH03e_worklist_algorithm<G,S> sc(g,v,s,cs);
  sc.solve();
}

#endif







