#ifndef PKH03B_WORKLIST_SOLVER_HPP
#define PKH03B_WORKLIST_SOLVER_HPP

#include <set>
#include <string>
#include <boost/tuple/tuple.hpp>
#include "../../worklist_strategy.hpp"
#include "../../topological_sort.hpp"

#include "../../factored_store.hpp"
#include "../../solver_common.hpp"
#include "../../set_union.hpp"
#include "../../diedge.hpp"
#include "../../store.hpp"
#include "../../pcs.hpp" 

template<class G, class S>
class PKH03b_worklist_algorithm {
private:
  G &_graph;
  std::vector<variable_info> &_varmap;
  store<S> &_sol;
  factored_store<vector_set<unsigned int> > _deltas;
  bool _change;
  std::vector<bool> _changed;
  // this holds the (topological) iteration order
  // in reverse
  std::vector<unsigned int> _order;      

  // the final few vectors are used
  // by the algorithm for detecting
  // strongly connected components.
  std::vector<unsigned int> _vindex; 
  std::vector<bool> _in_component; 
  std::vector<unsigned int> _stack; 
  std::vector<unsigned int> _unifyqueue;
  std::vector<bool> _visited;
  unsigned int _index; 

  // this final property is just used to obtain
  // useful information on the algorithms operation
  checker_stats &_stats;
public:
  PKH03b_worklist_algorithm(G &g, std::vector<variable_info> &v, 
			    store<S> &s,
			    checker_stats &stats) :
    _graph(g),
    _varmap(v),
    _sol(s),
    _deltas(size(g)),
    _changed(size(g),true),
    _order(size(g)),
    _vindex(size(g)), 
    _visited(size(g),false),
    _in_component(size(g),true), 
    _stats(stats) {

    S emptyset;
    
    for(unsigned int i=0;i!=size(g);++i) {
      if(_sol.is_stored(i)) {
	vector_set<unsigned int> sol;
	S const &osol(_sol[i]);
	for(typename S::const_iterator j(osol.begin());
	    j!=osol.end();++j) {
	  sol.insert(sol.end(),*j);
	}
	_deltas.update(i,sol);
      }
    }
  }

  void solve(void) {
    _change = true;
    vector_set<unsigned int> emptyset;

    while(_change) {
      _stats.iter_count++;
      _change = false;      

      // detect strongly connected components
      _index = 0;
      _order.clear(); 
      for(unsigned int i=0;i!=size(_graph);++i) {
	if(!_visited[i] && _varmap[i].node == i) { visit(i); }
      }
      fill(_visited.begin(),_visited.end(),false);

      process_unifyqueue();

      // process nodes in topological order
      for(std::vector<unsigned int>::reverse_iterator i(_order.rbegin());
	  i!=_order.rend();++i) {
	assert(_varmap[*i].node == *i);
	if(_changed[*i]) {
	  _changed[*i]=false;
	  _stats.visit_count++;
	  vector_set<unsigned int> delta(_deltas[*i]);
	  _deltas.update(*i,emptyset);

	  process_complex_constraints(*i,delta);	     
	  // propagate to all successors
	  typename G::const_siterator w, wend;
	  for(boost::tie(w,wend) = successors(*i,_graph);w!=wend;++w) {  
	    S wsol(_sol[*w]);
	    vector_set<unsigned int> wdelta(_deltas[*w]);
	    if(int_set_union(wdelta,delta,wsol)) {
	      _changed[*w]=true;
	      _change=true;
	      _deltas.update(*w,wdelta);
	      _sol.update(*w,wsol);
	    }	    
	  }
	}	
      }
    }     
  }

private:
  void visit(unsigned int n) {
    assert(_varmap[n].node == n);
    _visited[n]=true;
    _in_component[n]=false;
    _vindex[n]=_index++;
   
    // visit all successors
    typename G::const_siterator i,iend;
    for(boost::tie(i,iend) = successors(n,_graph);i!=iend;++i) {
      unsigned int w(*i);
      if(!_visited[w]) { visit(w); }
      if(!_in_component[w]) {
	unsigned int t(_varmap[w].node);
	if(_vindex[t] < _vindex[_varmap[n].node]) {
	  _varmap[n].node = t;
	}
      }
    }
    
    // at this point we must check to see
    // if any components have been identified
    if(_varmap[n].node == n) {
      _order.push_back(n);
      unsigned int t(_vindex[n]);
      _in_component[n] = true;
      while(!_stack.empty() && t < _vindex[_stack.back()]) {
	unsigned int w(_stack.back());
	_varmap[w].node = n;
	_in_component[w] = true;
	_unifyqueue.push_back(w);
	_stack.pop_back(); 
      }
    } else {
      _stack.push_back(n);
    }
  }

  void do_da_constraint(constraint const &c, 
			vector_set<unsigned int> const &delta) {
    // this function deals with constraints of the
    // form *x = &y
    unsigned int rhs(c.rhs().var());
    unsigned int off(c.lhs().offset());

    for(typename vector_set<unsigned int>::const_iterator j(delta.begin());j!=delta.end();++j) {
      if(type_safe(*j,off)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+off].node);
	S const &tmp(_sol[t]);
	
	if(tmp.find(rhs) == tmp.end()) { 
	  S tsol(tmp);
	  vector_set<unsigned int> tdelta(_deltas[t]);
	  tsol.insert(rhs);
	  tdelta.insert(rhs);
	  _sol.update(t,tsol); 
	  _deltas.update(t,tdelta);
	  _change = true;
	  _changed[t]=true;
	}
      }
    }
  }

  void do_ds_constraint(constraint const &c,
			vector_set<unsigned int> const &delta) {
    // this function deals with constraints of the
    // form *x = y
    unsigned int rhs(_varmap[c.rhs().var()].node);
    unsigned int loff(c.lhs().offset());
    S const &sol(_sol[rhs]);

    for(typename vector_set<unsigned int>::const_iterator j(delta.begin());j!=delta.end();++j) {      
      if(type_safe(*j,loff)) {
	// *x != NULL && *x != UNKNOWN
	unsigned int t(_varmap[*j+loff].node);
	if(t != rhs && add_edge(diedge(rhs,t),_graph)) {
	  S tsol(_sol[t]);
	  vector_set<unsigned int> tdelta(_deltas[t]);
	  if(int_set_union(tdelta,sol,tsol)) {
	    _stats.new_edge_count++;
	    _sol.update(t,tsol);
	    _deltas.update(t,tdelta);
	    _change=true;
	    _changed[t]=true;
	  }
	}
      }
    }
  }

  void do_sd_constraint(constraint const &c,
			vector_set<unsigned int> const &delta) {
    // this function deals with constraints of the
    // form x = *y
    unsigned int lhs(_varmap[c.lhs().var()].node);
    unsigned int roff(c.rhs().offset());
    bool flag(false);
    S lsol(_sol[lhs]);
    vector_set<unsigned int> ldelta(_deltas[lhs]);

    for(typename vector_set<unsigned int>::const_iterator j(delta.begin()); j!=delta.end();++j) {
      if(type_safe(*j,roff)) {
	// *j != NULL && *j != UNKNOWN
	unsigned int t(_varmap[*j+roff].node);       
	if(t != lhs && add_edge(diedge(t,lhs),_graph)) {
	  _stats.new_edge_count++;
	  flag |= int_set_union(ldelta,_sol[t],lsol);
	}
      } 
    }
    
    if(flag) {
      _sol.update(lhs,lsol);
      _deltas.update(lhs,ldelta);
      _change=true;
      _changed[lhs]=true;
    }
  }

  void do_complex_constraint(constraint c,
			     vector_set<unsigned int> const &delta) {
    _stats.complex_count++;

    if(c.lhs().dereferenced()) {
      if(c.rhs().addressof()) {
	// *x=&y
	do_da_constraint(c,delta);
      } else {
	// *x=y
	do_ds_constraint(c,delta);
      }
    } else {
      // x=*y
      do_sd_constraint(c,delta);
    }
  }

  void process_unifyqueue(void) { 
    vector_set<unsigned int> tmp;
    std::vector<unsigned int>::iterator j(_unifyqueue.begin());
    std::vector<unsigned int>::iterator jend(_unifyqueue.end());
    while(j != jend) {
      unsigned int n(_varmap[*j].node);
      vector_set<unsigned int> tsol;
      S const &sol(_sol[*j]);
      for(typename S::const_iterator i(sol.begin());
	  i!=sol.end();++i) {
	tsol.insert(tsol.end(),*i);
      }
      
      int_set_union(tmp,tsol);
      int_set_union(tmp,_deltas[*j]);
      merge_simple(n,*j,_graph);
      condense_nodes(n,*j,_varmap);
      if(_changed[*j]) { _changed[n]=true; }
      _sol.clear(*j);
      _deltas.clear(*j);

      ++j;
      // if we've finished updating the solution of a component
      if(j == jend || _varmap[*j].node != n) {
	bool flag(false);
	_deltas.update(n,tmp);	
	tmp.clear();
	_changed[n]=true;
	
	// eliminate self loops
	erase_edge(diedge(n,n),_graph);
      }	  
    }
    _unifyqueue.clear(); // empty the queue now!
  } 

  void process_complex_constraints(unsigned int n, vector_set<unsigned int> const &solution) {
    vector_set<constraint> complicates(_varmap[n].complicates);
    // process complex constraints
    for(vector_set<constraint>::const_iterator j(complicates.begin());
	j!=complicates.end();++j) {      
      // attempt to reduce redundant constraints as they are expensive
      if(!redundant_constraint(n,*j)) {
	do_complex_constraint(*j,solution);	  
      }
    }
  }

  bool redundant_constraint(unsigned int n, constraint const &c) {	    
    // the point of this function is to reduce the number of
    // complex constraints floating around.
    if(c.rhs().dereferenced()) {	     
      assert(c.rhs().var() == n);
      unsigned int lhs(c.lhs().var());
      // a = *n
      if(_varmap[lhs].node != lhs) {
	constraint tmp(expression(SCALAR,_varmap[lhs].node,0),c.rhs());
	_varmap[n].complicates.erase(c);
	if(!_varmap[n].complicates.insert(tmp).second) {
	  return true;
	}
      }
    } else {
      assert(c.lhs().var() == n);
      if(!c.rhs().addressof()) {
	unsigned int rhs(c.rhs().var());
	if(_varmap[rhs].node != rhs) {
	  // *n = a
	  constraint tmp(c.lhs(),expression(SCALAR,_varmap[rhs].node,c.rhs().offset()));
	  _varmap[n].complicates.erase(c);
	  if(!_varmap[n].complicates.insert(tmp).second) {
	    return true;
	  }
	}
      }
    }
    return false;
  }

  bool type_safe(unsigned int n, unsigned int o) {
    return n > 1 && (n+o) < _varmap[n].end;
  }

  bool int_set_intersect(S &to, S const &from, 
			 vector_set<unsigned int> &delta) {
    typename S::const_iterator f(from.begin());
    typename S::const_iterator t(to.begin());
    vector_set<unsigned int> tmp;
    
    while(t != to.end() && f != from.end()) {
      if(*f < *t) {
	tmp.insert(tmp.end(),*f);
	++f;
      } else if(*t < *f) { 
	to.erase(*t);
	tmp.insert(tmp.end(),*t);
	++t; 
      } else {
	++t;++f;
      }
    }
    while(f != from.end()) { 
      tmp.insert(tmp.end(),*f);++f;
    }
    while(t != to.end()) { 
      to.erase(*t);
      tmp.insert(tmp.end(),*t);++t;
    }

    set_union(delta,tmp);
    return tmp.size() > 0;
  }

  template<class T>
  inline bool int_set_union(vector_set<unsigned int> &delta, 
			    T const &from, 
			    S &sol) {
    _stats.union_count++;
    _stats.union_size += delta.size();
    _stats.union_size += from.size();

    vector_set<unsigned int> tmp;
    
    typename vector_set<unsigned int>::iterator d(delta.begin());
    typename T::const_iterator f(from.begin());
    
    while(d != delta.end() && f != from.end()) {
      if(*f < *d) {
	if(sol.insert(*f).second) { tmp.insert(tmp.end(),*f); }
	++f;
      } else if(*d < *f) { 
	tmp.insert(tmp.end(),*d);
	++d; 
      } else {
	tmp.insert(tmp.end(),*d);
	++d;++f;
      }
    }
    
    // note, only one of the following is possible.
    while(f != from.end()) { 
      if(sol.insert(*f).second) { tmp.insert(tmp.end(),*f); }
      ++f;
    }
    while(d != delta.end()) { 
      tmp.insert(tmp.end(),*d);++d;
    }
    delta.swap(tmp);
    return delta.size() != tmp.size();
  } 

  template<class T>
  inline bool int_set_union(T &to, T const &from) {
    _stats.union_count++;
    _stats.union_size += to.size();
    _stats.union_size += from.size();
    return set_union(to,from);
  }
};

template<class G, class S>
void PKH03b_worklist_solve(G &g, 
			   std::vector<variable_info> &v,
			   store<S> &s,
			   checker_stats &cs) {

  PKH03b_worklist_algorithm<G,S> sc(g,v,s,cs);
  sc.solve();
}

#endif







