#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <set>
#include <string>
#include <cmath>
#include <stdexcept>
#include <getopt.h>
#include <sys/time.h>

#include "pcs.hpp"
#include "pcs_front_end.hpp" // constraint language front end
#include "bidirectional_adjacency_list.hpp"
#include "store.hpp"
#include "topological_sort.hpp"
#include "find_strong_components.hpp"
#include "filter_graph.hpp"

// set implementations
#include "set_union.hpp" // stl set
#include "vector_set.hpp"
#include "bitset.hpp"
#include "invset.hpp"

// store implementations
#include "simple_store.hpp"
#include "factored_store.hpp"

// cycle detectors
#include "dummy_cycle_detector.hpp"
#include "MNR_online_cycle_detector.hpp"
#include "POSCC1_online_cycle_detector.hpp"
#include "POSCC2_online_cycle_detector.hpp"
#include "DFS_online_cycle_detector.hpp"

// iteration strategies
#include "lrf_strategy.hpp"
#include "fifo_strategy.hpp"
#include "lifo_strategy.hpp"

// core algorithms
#include "algorithm_HT.hpp"
#include "algorithm_HT_FDS.hpp"
#include "algorithm_PW1.hpp"
#include "algorithm_PWD.hpp"
#include "algorithm_PW2.hpp"
#include "algorithm_PW2_FDS.hpp"
#include "algorithm_PW3.hpp"
#include "algorithm_PW3_FDS.hpp"

using namespace std;
using namespace boost;

// ----------------------------------------
// Function Prototypes
// ----------------------------------------

// ----------------------------------------
// Custom Types and Macros
// ----------------------------------------

typedef bidirectional_adjacency_list<vector_set<unsigned int> > graph_t;

#define OPT_HELP 0
#define OPT_LIST 2
#define OPT_TIME 3
#define OPT_AXS 5
#define OPT_VERB 6
#define OPT_OUTPUT 7
#define OPT_NORC 8
#define OPT_UNFACTORED_STORE 10
#define OPT_FACTORED_STORE 11
#define OPT_VECSET 12
#define OPT_BITSET 13
#define OPT_STLSET 14
#define OPT_INVSET 15
#define OPT_NCD 16
#define OPT_SCD 17
#define OPT_DCD 18
#define OPT_FDS 19
#define OPT_PM 20
#define OPT_HT 25
#define OPT_HT_FDS 26
#define OPT_PW1_LRF 27
#define OPT_PW1_FIFO_MNR 28
#define OPT_PW1_LIFO_MNR 29
#define OPT_PW1_LRF_POSCC1 30
#define OPT_PW1_LRF_POSCC2 31
#define OPT_PW1_LRF_MNR 32
#define OPT_PWD_LRF 33
#define OPT_PWD_LRF_POSCC1 34
#define OPT_PWD_LRF_POSCC2 35
#define OPT_PWD_LRF_MNR 36

#define OPT_PW2_POSCC2 37
#define OPT_PW2_DFS 38
#define OPT_PW2_FDS 39
#define OPT_PW3_POSCC2 40
#define OPT_PW3_DFS 41
#define OPT_PW3_FDS 42

typedef struct {
  const char *pattern;
  const char *desc;
  unsigned int val;
} option_info;

class my_timer {
private:
  struct timeval _start;
public:
  my_timer(void) {
    gettimeofday(&_start,NULL);
  }

  double elapsed(void) {
    struct timeval tmp;
    gettimeofday(&tmp,NULL); 
    double end = tmp.tv_sec + (tmp.tv_usec / 1000000.0);
    double start = _start.tv_sec + (_start.tv_usec / 1000000.0);    
    return end - start;
  }  
};

struct execution_stats {
public:
  unsigned int vcount; // Visit Count
  unsigned int num_comp_proc; // number of complex constraints processed
  unsigned int ucount; // number of set union operations
  unsigned int iter_count; // number of iterations of outer loop
  unsigned int post_iter_count; // number of iterations after main loop
  unsigned int new_edge_count; // number of edges added to graph
  double avg_union_size; // average size of sets involved in set unions
};

struct structural_stats {
public:
  unsigned int nTriv; // number trivial constraints (e.g. a=&b)
  unsigned int nSimp; // number (initial) simple constraints (e.g. a=b)
  unsigned int nSimpAfter; // number simple constraints added during solving
  double SimpRatio; // ratio of p=q to p=q+1 edges
  unsigned int nComp; // number complex constraints (e.g. *p=q)
  unsigned int nVar;  // number of (unnormalised) variables
  unsigned int nHeap; // number of (unnormalised) head variables
  unsigned int nElim; // number of nodes eliminated in total.
  unsigned int nWeighted; // number of positive weight cycles
};

struct precision_stats {
public:
  double avg_set_size; // average set size across all variables 
  double avg_deref_size_norm; // average normalised set size at dereference sites 
  double sets_0;       // percentage of (normalised) dereference sites with 0 targets
  double sets_1;       // percentage of (normalised) dereference sites with 1 targets
  double sets_2;       // percentage of (normalised) dereference sites with 2 targets
  double sets_3_10;    // percentage of (normalised) dereference sites with 3-10 targets
  double sets_11_100;    // percentage of (normalised) dereference sites with 11-100 targets
  double sets_101_1000;  // percentage of (normalised) dereference sites with 101-1000 targets
  double sets_1001_;    // percentage of (normalised) dereference sites with 1001+ targets
  double alias_frequency; // er ...
};

// ----------------------------------------
// Global Variables
// ----------------------------------------

bool timing_mode(false);
bool quiet_mode(false);
bool full_axiomatic_solution(false);
bool summary_mode(false);
ostream *output_file(&cout);
unsigned int real_var_limit;
unsigned int domain_size;

// ----------------------------------------
// Program Text
// ----------------------------------------

void build_constraint_graph(graph_t &g, vector<variable_info> &varmap,
			    vector<constraint> const &constraints) {

  for(vector<constraint>::const_iterator i(constraints.begin());
      i!=constraints.end();++i) {
    expression const &lhs(i->lhs());
    expression const &rhs(i->rhs());
    if(lhs.dereferenced()) {
      // *x = y or *x = &y (complex)
      if(rhs.addressof() || rhs.var() > 1) {
	// NOT *x = UNKNOWN
	varmap[lhs.var()].complicates.insert(*i);
      } 
    } else if(rhs.dereferenced()) {
      // x = *y (complex)
      if(lhs.var() > 1) {
	// NOT UNKNOWN = *y
	varmap[rhs.var()].complicates.insert(*i);
      }
    } else if(rhs.addressof()) {
      // x = &y (trivial)
      varmap[lhs.var()].solution.insert(rhs.var());
    } else if(lhs.var() > 1 && rhs.var() > 1) {
      // x = y (simple)
      
      add_edge(diedge(rhs.var(),lhs.var()),g);
      get(diedge(rhs.var(),lhs.var()),g).insert(rhs.offset());
    }
  }
}

void collapse_nodes(unsigned int a, unsigned int b, graph_t &g, vector<variable_info> &varmap) {
  condense_nodes(a,b,varmap);
  set_union(varmap[a].solution,varmap[b].solution);
  merge<set_union_func<unsigned int> >(a,b,g);

  if(valid_edge(diedge(a,a),g)) {
    // eliminate self loops
    vector_set<unsigned int> &labels(get(diedge(a,a),g));
    labels.erase((unsigned int) 0);
    if(labels.empty()) { erase_edge(diedge(a,a),g); }
  }

  varmap[b].solution.clear();
  // this bit is basically for the rountev_chandra algorithm
  varmap[a].address_taken |= varmap[b].address_taken;
  varmap[a].indirect_target |= varmap[b].indirect_target;  
}

unsigned int eliminate_static_cycles(graph_t &g, vector<variable_info> &varmap) {  
  // Now, perform static cycle detection.
  unsigned int r(0);
  n2c_wrapper root(varmap);
  find_strong_components(filter_graph<graph_t,zero_edge_select>(g),root);
  
  for(unsigned int i=0;i!=varmap.size();++i) {
    unsigned int n(varmap[i].node);
    if(n != i) {
      collapse_nodes(n,i,g,varmap);
      ++r;
    } else {
      // eliminate self loops
      if(valid_edge(diedge(n,n),g)) {
	vector_set<unsigned int> &labels(get(diedge(n,n),g));
	labels.erase((unsigned int) 0);
	if(labels.empty()) { erase_edge(diedge(n,n),g); }
      }
    }
  }
  return r;
}

unsigned int perform_rountev_chandra_elimination(graph_t &g, vector<variable_info> &varmap) {
  // For information on what this function does See "Off-Line Variable Substitution 
  // for Scaling Points-to Analysis", A. Rountev and S Chandra, PLDI 200. 
  vector<unsigned int> top_order(size(g));
  topological_sort(g,top_order);

  unsigned int count=0;
  // remember topological_sort gives the order in reverse!
  for(vector<unsigned int>::reverse_iterator i(top_order.rbegin());
      i!=top_order.rend();++i) {
    if(!varmap[*i].address_taken &&
       !varmap[*i].indirect_target) {

      graph_t::piterator pb,pe;
      tie(pb,pe) = predecessors(*i,g);
      unsigned int root(varmap.size());
      for(;pb!=pe;++pb) {
	vector_set<unsigned int> &labels(*(pb.base()->second()));
	if(labels.size() != 1 || (*labels.begin()) != 0) { 
	  // basically, cannot combine nodes when there is
	  // a weighted edge between them
	  root = varmap.size(); break;
	}

	unsigned int w(varmap[*pb].node);
	if(root == varmap.size()) { root = w; }
	else if(w != root) {
	  // ok, nodes don't share common root.
	  root = varmap.size();break;
	} 

	if(!is_subset(varmap[w].solution,varmap[*i].solution)) {
	  // alas, solution makes this incompatible
	  root = varmap.size();break;
	}
      }
      if(root != varmap[*i].node && root < varmap.size()) { 
	// ok, equivalence detected
	varmap[*i].node = root;
	collapse_nodes(root,*i,g,varmap);
	++count;
	//	cerr << "Collapsing " << varmap[*i].name << " into " << varmap[root].name << endl;
      } 
    }
  }

  return count;
}

void perform_projection_merging(vector<constraint> &constraints,
				vector<variable_info> &varmap) {
  // this could certainly be more efficient!
  vector<set<constraint> > lderefs(varmap.size());
  vector<set<constraint> > rderefs(varmap.size());
  vector<constraint> tmp;

  for(vector<constraint>::iterator i(constraints.begin());
      i!=constraints.end();++i) {
    expression const &lhs(i->lhs());
    expression const &rhs(i->rhs());
    if(lhs.dereferenced() && lhs.offset() == 0 && rhs.offset() == 0) {
      // *x = y or *x = &y (complex)
      if(rhs.addressof()) {
	tmp.push_back(*i);
      } else {
	lderefs[rhs.var()].insert(*i);
      }
    } else if(rhs.dereferenced() && rhs.offset() == 0) {
      // x = *y (complex)
      rderefs[lhs.var()].insert(*i);
    } else {
      // x = y
      tmp.push_back(*i);
    }
  }

  // now do left projection merging:
  for(unsigned int i=0;i!=lderefs.size();++i) {
    set<constraint> &ld(lderefs[i]);
    if(ld.size() <= 1) {
      // nothing is on
      if(!ld.empty()) { tmp.push_back(*(ld.begin())); }
    } else {
      // ok, game on
      unsigned int nvar(varmap.size());
      varmap.push_back(variable_info("pm",nvar+1,nvar,0));
      constraint c(expression(DEREFERENCE,nvar,0),expression(SCALAR,i,0));
      tmp.push_back(c);
      for(set<constraint>::iterator j(ld.begin());j!=ld.end();++j) {
	c = constraint(expression(SCALAR,nvar,0),expression(SCALAR,j->lhs().var(),0));
	tmp.push_back(c);
      }
    }
  }

  // now do right projection merging:
  for(unsigned int i=0;i!=rderefs.size();++i) {
    set<constraint> &rd(rderefs[i]);
    if(rd.size() <= 1) {
      // nothing is on
      if(!rd.empty()) { tmp.push_back(*(rd.begin())); }
    } else {
      // ok, game on
      unsigned int nvar(varmap.size());
      varmap.push_back(variable_info("pm",nvar+1,nvar,0));
      constraint c(expression(SCALAR,i,0),expression(DEREFERENCE,nvar,0));
      tmp.push_back(c);
      for(set<constraint>::iterator j(rd.begin());j!=rd.end();++j) {
	c = constraint(expression(SCALAR,nvar,0),expression(SCALAR,j->rhs().var(),0));
	tmp.push_back(c);
      }
    }
  }

  constraints.swap(tmp);
}

// build_axiomatic_solution: determine the axioms and their dependences
//                           from the graph and final solution.
void build_axiomatic_solution(graph_t &graph,
			      vector<constraint> &constraints,
			      vector<set<unsigned int> > &deps,
			      vector<variable_info> const &varmap,
			      store<vector_set<unsigned int> > &solution) {
  
  set<constraint> axioms;           // used to prevent inferring things about the axioms
  map<constraint,unsigned int> c2n; // map constraints to their indices    
  unsigned int index(0);
  
  for(vector<constraint>::const_iterator i(constraints.begin());
      i!=constraints.end();++i,++index) {
    c2n[*i] = index;
    axioms.insert(*i);
  }
  
  // process simple constraints
  for(unsigned int i=0;i<varmap.size();++i) {
    graph_t::const_siterator k,kend;
    for(boost::tie(k,kend) = successors(i,graph);k!=kend;++k) {
      diedge d(i,*k);
      for(vector_set<unsigned int>::const_iterator l(get(d,graph).begin());
	  l!=get(d,graph).end();++l) {
	expression lhs(SCALAR,*k);
	expression rhs(SCALAR,i,*l);
	constraint c(lhs,rhs);
	if(axioms.find(c) == axioms.end()) {
	  c2n[c] = index++;
	  constraints.push_back(c);
	}
      }
    }
  }    
  
  // process all trivial constraints
  for(unsigned int i=0;i!=varmap.size();++i) {
    expression lhs(SCALAR,i);
    vector_set<unsigned int>::const_iterator jend(solution[i].end());
    for(vector_set<unsigned int>::const_iterator j(solution[i].begin());
	j!=jend;++j) {
      expression rhs(ADDRESSOF,*j);
      constraint c(lhs,rhs);
      if(axioms.find(c) == axioms.end()) {
	c2n[c] = index++;
	constraints.push_back(c);
      }
    }
  }
  
  deps.resize(constraints.size());
  // at this point we back compute all dependence information.
  // phew it's quite a task!
  index=0;
  for(vector<constraint>::iterator i(constraints.begin());
      i!=constraints.end();++i,++index) {
    expression const &lhs(i->lhs());
    expression const &rhs(i->rhs());
    
    if(lhs.dereferenced()) {
      for(vector_set<unsigned int>::const_iterator j(solution[lhs.var()].begin());
	  j!=solution[lhs.var()].end();++j) {
	if(*j > 1) {
	  expression a_lhs(SCALAR,*j + lhs.offset());
	  constraint c(a_lhs,rhs);
	  if(c2n.find(c) != c2n.end()) {
	    deps[c2n[c]].insert(index);
	    
	    expression lhs2(SCALAR,lhs.var());
	    expression rhs2(ADDRESSOF,*j);
	    constraint c2(lhs2,rhs2);
	    if(c2n.find(c2) != c2n.end()) {
	      deps[c2n[c]].insert(c2n[c2]);
	    } else {
	      cerr << "1 WARNING: check failure for constraint. ";
	      cerr << constraint2str(c2,varmap) << " is missing [" << index << "]." << endl;
	    }
	  } else {
	    cerr << "2 WARNING: check failure for constraint ";
	    cerr << constraint2str(c,varmap) << " is missing [" << index << "]." << endl;
	    
	  }
	}
      }	  
    } else if(rhs.dereferenced()) {
      for(vector_set<unsigned int>::const_iterator j(solution[lhs.var()].begin());
	  j!=solution[lhs.var()].end();++j) {
	if(*j > 1) {
	  expression a_rhs(SCALAR,*j+rhs.offset());
	  constraint c(lhs,a_rhs);
	  if(c2n.find(c) != c2n.end()) {
	    deps[c2n[c]].insert(index);
	    
	    expression lhs2(SCALAR,rhs.var());
	    expression rhs2(ADDRESSOF,*j);
	    constraint c2(lhs2,rhs2);
	    if(c2n.find(c2) != c2n.end()) {
	      deps[c2n[c]].insert(c2n[c2]);
	    } else {
	      cerr << "3 WARNING: check failure for constraint. ";
	      cerr << constraint2str(c2,varmap) << " is missing [" << index << "]." << endl;
	    }
	  } else {
	    cerr << "4 WARNING: check failure for constraint ";
	    cerr << constraint2str(c,varmap) << " is missing [" << index << "]." << endl;
	  }
	}
      }
    } else if(rhs.addressof() && index >= axioms.size()) {
      // to determine where this came from we look
      // in the predecessor solution sets for the lhs ...
      graph_t::const_piterator k,kend;
      for(boost::tie(k,kend) = predecessors(lhs.var(),graph);k!=kend;++k) {
	vector_set<unsigned int> tmp(solution[*k]);
	if(*k != lhs.var() && 
	   (tmp.find(rhs.var()) != tmp.end())) {
	  // found one possible derivation. i don't bother
	  // enumerating them all, btw.
	  expression lhs_1(SCALAR,*k);
	  expression rhs_1(ADDRESSOF,rhs.var());
	  constraint c1(lhs_1,rhs_1);
	  expression lhs_2(SCALAR,lhs.var());
	  expression rhs_2(SCALAR,*k);
	  constraint c2(lhs_2,rhs_2);
	  
	  if(c2n.find(c1) != c2n.end()) {
	    deps[index].insert(c2n[c1]);
	    
	    if(c2n.find(c2) != c2n.end()) {
	      deps[index].insert(c2n[c2]);
	    } else {
	      cerr << "5 WARNING: check failure for constraint. ";
	      cerr << constraint2str(c2,varmap) << " is missing." << endl;
	    }
	  } else {
	    cerr << "6 WARNING: check failure for constraint. ";
	    cerr << constraint2str(c1,varmap) << " is missing." << endl;
	  }
	}
      }
    }
  }  
}

// print_axiomatic_solution: output the solution in axiomatic format,
//                           along with dependences
void print_axiomatic_solution(vector<constraint> const &constraints, 
			      vector<set<unsigned int> > const &deps,
			      vector<variable_info> const &varmap) {
  unsigned int index(0);
  
  for(vector<constraint>::const_iterator i(constraints.begin());
      i!=constraints.end();++i,++index) {
    cout << index << ".\t";
    if(!deps[index].empty()) {
      cout << "(";
      bool flag(false);
      for(set<unsigned int>::iterator i(deps[index].begin());
	  i!=deps[index].end();++i) {
	if(!flag) { flag=true; } else { cout << "+"; }
	cout << *i;
      }
      cout << ")\t";
    } else cout << "\t";
    cout << constraint2str(*i,varmap);
    cout << endl;
  }
}

template<class T>
void print_partial_solution(vector<variable_info> const &varmap, 
			    store<T> &solution,
			    ostream &out) {
  unsigned int index(0);
  for(vector<variable_info>::const_iterator i(varmap.begin());
      i!=varmap.end();++i,++index) {
    // don't output for variables who were projection merged
    if(index == real_var_limit) { break; }
    out << i->name << " = { ";
    for(typename T::const_iterator j(solution[i->node].begin());
	j!=solution[i->node].end();++j) {
      if(j != solution[i->node].begin()) {
	out << ", ";
      }
      out << varmap[*j].name;
    }
    out << " }" << endl;
  }
}

bool potential_alias(vector_set<unsigned int> const &s1, 
		       vector_set<unsigned int> const &s2) {

  vector_set<unsigned int>::const_iterator s1i(s1.begin());
  vector_set<unsigned int>::const_iterator s2i(s2.begin());
  while(s1i != s1.end() && s2i != s2.end()) {
    // basically, I don't count NULL or UNKNOWN in
    // this metric
    if(*s1i == *s2i && *s1i > 1) { 
      return true;
    } else if(*s1i < *s2i) { ++ s1i; }
    else { ++s2i; }
  }
  return false;
}

#if 0
double compute_alias_frequency(vector<variable_info> &varmap
			       store<vector_set<unsigned int> > &sol) {
  unsigned int count(0);
  unsigned int aliases(0);
  
  for(unsigned int i=0;i!=varmap.size()-1;++i) {
    unsigned int in(varmap[i].node);
    if(lderefed[i] || rderefed[i]) {
      for(unsigned int j=i+1;j!=varmap.size();++j) {
	unsigned int jn(varmap[j].node);
	if(lderefed[j] || rderefed[j]) {
	  //	  count += varmap[i].rep * varmap[j].rep;
	  count++;
	  if(potential_alias(sol[in],sol[jn])) {
	    //	    aliases += varmap[i].rep * varmap[j].rep;
	    aliases ++;
	  }
	}
      }
    }
  }

  double r(aliases);
  r /= count;
  return r;
}
#endif

unsigned int count_positive_weight_cycles(vector<variable_info> &varmap,
					  graph_t &graph) {
  // assumption that no non-pos-weighted cycles exist in graph!!!
  vector<unsigned int> root(size(graph));
  find_strong_components(graph,root);
  set<unsigned int> comps;
  
  for(unsigned int i=0;i!=size(graph);++i) {
    if(root[i] != i) {
      comps.insert(root[i]);
    } else {
      if(valid_edge(diedge(i,i),graph)) {
	vector_set<unsigned int> &labels(get(diedge(i,i),graph));
	if(!labels.empty() && (labels.size() > 1 || *(labels.begin()) != 0)) {
	  comps.insert(i);
	}
      }
    }
  }

  return comps.size();
}

execution_stats compute_execution_stats(checker_stats &stats) {
  execution_stats r;
  r.vcount = stats.visit_count;
  r.num_comp_proc = stats.complex_count;
  r.ucount = stats.union_count;
  r.iter_count = stats.iter_count;
  r.post_iter_count = stats.post_iter_count;

  if(stats.union_count == 0) {
    r.avg_union_size = 0;
  } else {
    r.avg_union_size = ((double) stats.union_size) / ((double)(stats.union_count*2));
  }

  r.new_edge_count = stats.new_edge_count;
  return r;
}

structural_stats compute_structural_stats(vector<constraint> &constraints,
					  vector<variable_info> &varmap,
					  graph_t &graph) {
  structural_stats r;

  r.nTriv=0;
  r.nSimp=0;
  r.SimpRatio=0;
  r.nComp=0;
  r.nVar=varmap.size();
  r.nHeap=0;
  r.nElim=0;
  r.nWeighted=count_positive_weight_cycles(varmap,graph);
  
  for(vector<constraint>::const_iterator i(constraints.begin());
      i!=constraints.end();++i) {
    if(i->lhs().dereferenced()) {
      if(i->rhs().addressof())         { r.nComp++;
      } else                           { r.nComp++; }
    } else if(i->rhs().dereferenced()) { r.nComp++; }
    else if(!i->rhs().addressof())     { r.nSimp++; } 
    else                               { r.nTriv++; }
  }

  // we accumulate the total number of edges in the graph
  // into count
  unsigned int count(0);
  for(unsigned int i=0;i!=size(graph);++i) {
    graph_t::siterator sb,se;
    tie(sb,se) = successors(i,graph);
    unsigned int is(varmap[i].variables.size()+1);
    for(;sb!=se;++sb) {
      unsigned int sbs(varmap[*sb].variables.size()+1);
      vector_set<unsigned int> &labels(*(sb.base()->second()));
      count += labels.size();
      r.SimpRatio += labels.size();
      if(labels.empty() || *(labels.begin()) == 0) {
	r.SimpRatio --;
      }
    }

    if(varmap[i].name.find("HEAP") < varmap[i].name.size()) { r.nHeap++; }

    if(varmap[i].node != i) {
      r.nElim ++;
    }  
  }

  r.nSimpAfter = count;
  r.SimpRatio /= r.nSimpAfter;

  return r;
}

template<class T>
unsigned int normalised_size(T const &s, vector<variable_info> &varmap) {
  unsigned int r(0);
  for(typename T::const_iterator j(s.begin());
      j!=s.end();++j) {
    r += varmap[*j].rep;
  }

  return r;
}

template<class T>
precision_stats compute_precision_stats(vector<constraint> &constraints,
					vector<variable_info> &varmap,
					store<T> &sol) {
  precision_stats r;
  r.avg_set_size=0;
  r.avg_deref_size_norm=0; 
  r.sets_0 = 0;
  r.sets_1 = 0;
  r.sets_2 = 0;
  r.sets_3_10 = 0;
  r.sets_11_100 = 0;
  r.sets_101_1000 = 0;
  r.sets_1001_ = 0;

  //  r.alias_frequency=compute_alias_frequency(varmap,lderefed,rderefed,sol); 
  
  unsigned int vsize(0),index(0);

  for(vector<variable_info>::iterator i(varmap.begin());
      i!=varmap.end();++i,++index) {
    if(index == real_var_limit) { break; }
    T tmp(varmap.size());
    unsigned int n(i->node);
    if(sol.is_stored(n)) { tmp = sol[n]; }
    r.avg_set_size += tmp.size();
  }

  r.avg_set_size /= real_var_limit;

  // now compute solution breakdown
  unsigned int num_derefs(0);
  for(vector<constraint>::iterator i(constraints.begin());
      i!=constraints.end();++i) {
    if(i->lhs().dereferenced() || i->rhs().dereferenced()) {
      unsigned int n;
      num_derefs ++;
      
      if(i->rhs().dereferenced()) {
	n = varmap[i->rhs().var()].node;
      } else {
	n = varmap[i->lhs().var()].node;
      }

      unsigned int s(normalised_size(sol[n],varmap));

      if(s == 0) {
	r.sets_0 ++;
      } else if(s == 1) {
	r.sets_1 ++;
      } else if(s == 2) {
	r.sets_2 ++;
      } else if(s >= 3 && s <= 10) {
	r.sets_3_10 ++;
      } else if(s >= 11 && s <= 100) {
	r.sets_11_100 ++;
      } else if(s >= 101 && s <= 1000) {
	r.sets_101_1000 ++;
      } else {
	r.sets_1001_ ++;
      }
      r.avg_deref_size_norm += s;
    }
  }

  r.avg_deref_size_norm /= num_derefs;
  r.sets_0 = (r.sets_0 * 100) / num_derefs;
  r.sets_1 = (r.sets_1 * 100) / num_derefs;  
  r.sets_2 = (r.sets_2 * 100) / num_derefs;
  r.sets_3_10 = (r.sets_3_10 * 100) / num_derefs;
  r.sets_11_100 = (r.sets_11_100 * 100) / num_derefs;
  r.sets_101_1000 = (r.sets_101_1000 * 100) / num_derefs;
  r.sets_1001_ = (r.sets_1001_ * 100) / num_derefs;
  return r;
}

template<class S> 
void do_solve(unsigned int algorithm,
	      unsigned int store_mode,
	      vector<constraint> &constraints,
	      vector<variable_info> &varmap,
	      graph_t &graph) {    

  //  typedef filter_graph<graph_t,zero_edge_select> fg_t;
  //  fg_t fg(graph);

  store<S> *sol;

  // construct the store to be used
  switch(store_mode) {
  case OPT_UNFACTORED_STORE:
    sol = new simple_store<S>(size(graph),domain_size);
    break;    
  case OPT_FACTORED_STORE:
    sol = new factored_store<S>(size(graph),domain_size);
    break;    
  default:
    throw runtime_error("invalid store selected");
  }

  // build the solution  
  for(unsigned int i=0;i!=varmap.size();++i) {
    if(varmap[i].node == i) {
      S tmp(domain_size);
      tmp.insert(varmap[i].solution.begin(),varmap[i].solution.end());
      sol->update(i,tmp);
      varmap[i].solution.clear();
    }
  }  

  // begin timing and start solving!
  my_timer timer;
  checker_stats stats;

  switch(algorithm) {
  case OPT_HT:
    solve_HT<graph_t,S>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_HT_FDS:
    solve_HT_FDS<graph_t,S>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW1_LRF:   
    solve_PW1<graph_t,S,lrf_strategy,dummy_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW1_LRF_POSCC1:   
    solve_PW1<graph_t,S,lrf_strategy,POSCC1_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW1_LRF_POSCC2:   
    solve_PW1<graph_t,S,lrf_strategy,POSCC2_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW1_LRF_MNR:   
    solve_PW1<graph_t,S,lrf_strategy,MNR_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW1_FIFO_MNR:   
    solve_PW1<graph_t,S,fifo_strategy,MNR_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW1_LIFO_MNR:   
    solve_PW1<graph_t,S,lifo_strategy,MNR_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PWD_LRF:   
    solve_PWD<graph_t,S,lrf_strategy,dummy_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PWD_LRF_POSCC1:   
    solve_PWD<graph_t,S,lrf_strategy,POSCC1_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PWD_LRF_POSCC2:   
    solve_PWD<graph_t,S,lrf_strategy,POSCC2_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PWD_LRF_MNR:   
    solve_PWD<graph_t,S,lrf_strategy,MNR_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW2_DFS:   
    solve_PW2<graph_t,S,DFS_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW2_POSCC2:   
    solve_PW2<graph_t,S,POSCC2_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW2_FDS:   
    solve_PW2_FDS<graph_t,S>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW3_POSCC2:   
    solve_PW3<graph_t,S,POSCC2_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW3_DFS:   
    solve_PW3<graph_t,S,DFS_online_cycle_detector>(graph,varmap,*sol,domain_size,stats);    
    break;
  case OPT_PW3_FDS:   
    solve_PW3_FDS<graph_t,S>(graph,varmap,*sol,domain_size,stats);    
    break;
  default:
    cerr << "unknown algorithm" << endl;
  }

  if(timing_mode) { cout << setprecision(3) << timer.elapsed() << endl; }

  if(!quiet_mode) {
    if(full_axiomatic_solution) {
      //	vector<set<unsigned int> > deps(constraints.size());
      //	build_axiomatic_solution(graph,constraints,deps,varmap,*solution);
      //	print_axiomatic_solution(constraints,deps,varmap);
      throw runtime_error("axiomatic solution not completed");
    } else {
      print_partial_solution(varmap,*sol,*output_file);
    }
  }  
  
  if(summary_mode) {
    if(full_axiomatic_solution) {
      cout << "WARNING: summary maybe incorrect when used with --axiomatic-solution" << endl;
    }

    // compute summary information 
    execution_stats estats = compute_execution_stats(stats);
    structural_stats sstats = compute_structural_stats(constraints,varmap,graph);
    precision_stats pstats = compute_precision_stats(constraints,varmap,*sol);
    // print summary information
    cout << "1.  Visit Count                                    : " << estats.vcount << endl;
    cout << "2.  Number Complex Constraints Processed           : " << estats.num_comp_proc << endl;
    cout << "3.  Number of set union ops                        : " << estats.ucount << endl;
    cout << "4.  Average set size involved in union             : " << setprecision(3) << estats.avg_union_size << endl;
    cout << "5.  Number edges added to graph                    : " << estats.new_edge_count << endl;   
    if(estats.post_iter_count > 0) { 
      cout << "6.  Number iterations of outer loop                : " << estats.iter_count - estats.post_iter_count;
      cout << "/" << estats.post_iter_count; 
    } else {
      cout << "6.  Number iterations of outer loop                : " << estats.iter_count;
    }
    cout << endl;
    cout << "7.  Number Trivial Constraints                     : " << sstats.nTriv << endl;
    cout << "8.  Number Simple Constraints (before)             : " << sstats.nSimp << endl;
    cout << "9.  Number Complex Constraints                     : " << sstats.nComp << endl;
    cout << "10. Number Variables (unnormalised)                : " << sstats.nVar << endl;
    cout << "11. Number Variables (address taken, unnormalised) : " << domain_size << endl;
    cout << "12. Number Heap Variables (unnormalised)           : " << sstats.nHeap << endl;
    cout << "13. Number Variables Eliminated                    : " << sstats.nElim << endl;    
    cout << "14. Number of positive weight cycles               : " << sstats.nWeighted << endl;
    cout << "15. Average Set Size (across all variables)        : " << setprecision(4) << pstats.avg_set_size << endl;
    cout << "16. Average Normalised Deref Size                  : " << pstats.avg_deref_size_norm << endl;    
    cout << "17. Percent Normalised Deref Size = 0              : " << setprecision(2) << pstats.sets_0 << endl;    
    cout << "18. Percent Normalised Deref Size = 1              : " << setprecision(2) << pstats.sets_1 << endl;    
    cout << "19. Percent Normalised Deref Size = 2              : " << setprecision(2) << pstats.sets_2 << endl;    
    cout << "20. Percent Normalised Deref Size = 3 - 10         : " << setprecision(2) << pstats.sets_3_10 << endl;    
    cout << "21. Percent Normalised Deref Size = 11 - 100       : " << setprecision(2) << pstats.sets_11_100 << endl;    
    cout << "22. Percent Normalised Deref Size = 101 - 1000     : " << setprecision(2) << pstats.sets_101_1000 << endl;    
    cout << "23. Percent Normalised Deref Size = 1001 - inf     : " << setprecision(2) << pstats.sets_1001_ << endl;    
    cout << "24. Alias Frequency                                : " << pstats.alias_frequency << endl;
  }       
}

string expr2str(expression const &e, 
		vector<variable_info> const &varmap) {
  string v(varmap[e.var()].name);

  if(e.dereferenced()) {
    if(e.offset() != 0) {
      return string("*(") + v + "+" + boost::lexical_cast<string>(e.offset()) + ")";
    } else {
      return string("*") + v;
    }
  }
  
  string r;    
  
  if(e.addressof()) { r += "&"; }
  if(e.offset() != 0) {
    return r + v + "+" + boost::lexical_cast<string>(e.offset());
  } else {
    return r + v;
  }
}

string constraint2str(constraint const &c, 
		      vector<variable_info> const &varmap) {
  return expr2str(c.lhs(),varmap) + " = " + expr2str(c.rhs(),varmap);
}

int main(int argc, char *argv[]) {
  struct option long_options[]={
    {"help",no_argument,NULL,OPT_HELP},
    {"time",no_argument,NULL,OPT_TIME},
    {"axiomatic-solution",no_argument,NULL,OPT_AXS},
    {"verbose",no_argument,NULL,OPT_VERB},
    {"output",required_argument,NULL,OPT_OUTPUT},
    {"no-rountev-chandra",no_argument,NULL,OPT_NORC},
    {"projection-merging",no_argument,NULL,OPT_PM},
    {"HT",no_argument,NULL,OPT_HT},
    {"HT-FDS",no_argument,NULL,OPT_HT_FDS},
    {"PW1-LRF",no_argument,NULL,OPT_PW1_LRF},
    {"PW1-LRF-POSCC1",no_argument,NULL,OPT_PW1_LRF_POSCC1},
    {"PW1-LRF-POSCC2",no_argument,NULL,OPT_PW1_LRF_POSCC2},
    {"PW1-LRF-MNR",no_argument,NULL,OPT_PW1_LRF_MNR},
    {"PW1-LIFO-MNR",no_argument,NULL,OPT_PW1_LIFO_MNR},
    {"PW1-FIFO-MNR",no_argument,NULL,OPT_PW1_FIFO_MNR},
    {"PWD-LRF",no_argument,NULL,OPT_PWD_LRF},
    {"PWD-LRF-POSCC1",no_argument,NULL,OPT_PWD_LRF_POSCC1},
    {"PWD-LRF-POSCC2",no_argument,NULL,OPT_PWD_LRF_POSCC2},
    {"PWD-LRF-MNR",no_argument,NULL,OPT_PWD_LRF_MNR},
    {"PW2-POSCC2",no_argument,NULL,OPT_PW2_POSCC2},
    {"PW2-DFS",no_argument,NULL,OPT_PW2_DFS},
    {"PW2-FDS",no_argument,NULL,OPT_PW2_FDS},
    {"PW3-DFS",no_argument,NULL,OPT_PW3_DFS},
    {"PW3-POSCC2",no_argument,NULL,OPT_PW3_POSCC2},
    {"PW3-FDS",no_argument,NULL,OPT_PW3_FDS},
    {"PW3-FDS",no_argument,NULL,OPT_PW3_FDS},
    {"Vecset",no_argument,NULL,OPT_VECSET},
    {"Bitset",no_argument,NULL,OPT_BITSET},
    {"Stlset",no_argument,NULL,OPT_STLSET},
    {"Invset",no_argument,NULL,OPT_INVSET},
    NULL
  };
  
  char *descriptions[]={
    "        --help                    display this information",
    " -q                               suppress printing the solution",
    " -s                               print out summary information",
    " -t     --time                    measure convergence time",
    "        --axiomatic-solution      print out full solution in axiomatic form with rule dependences",
    "        --verbose                 print out useful information about what's going on",
    " -o<x>  --output=<x>              write solution to file <x>.",
    "        --no-rountev-chandra      do not the Rountev-Chandra Elimination method",
    "        --projection-merging      use projection merging",
    "        --HT                      use original heintze tardieu solver",
    "        --HT-FDS                  use field-sensitive heintze tardieu solver",
    "        --PW1-LRF                 use solver PW1 with LRF strategy and no cycle detection",
    "        --PW1-LRF-POSCC1          use solver PW1 with LRF strategy and algorithm POSCC1",
    "        --PW1-LRF-POSCC2          use solver PW1 with LRF strategy and algorithm POSCC2",
    "        --PW1-LRF-MNR             use solver PW1 with LRF strategy and algorithm MNR",
    "        --PW1-LIFO-MNR            use solver PW1 with LIFO strategy and algorithm MNR",
    "        --PW1-FIFO-MNR            use solver PW1 with FIFO strategy and algorithm MNR",
    "        --PWD-LRF                 use solver PWD with LRF strategy and no cycle detection",
    "        --PWD-LRF-POSCC1          use solver PWD with LRF strategy and algorithm POSCC1",
    "        --PWD-LRF-POSCC2          use solver PWD with LRF strategy and algorithm POSCC2",
    "        --PWD-LRF-MNR             use solver PWD with LRF strategy and algorithm POSCC2",
    "        --PW2-DFS                 use solver PW2 and Tarjans Algorithm",
    "        --PW2-POSCC2              use solver PW2 and POSCC2 algorithm",
    "        --PW3-DFS                 use solver PW3 and Tarjans Algorithm",
    "        --PW3-POSCC2              use solver PW3 and POSCC2 algorithm",
    "        --PW2-FDS                 use field-sensitive PW2 solver",
    "        --PW3                     use solver PW3",
    "        --PW3-FDS                 use field-sensitive PW3 solver",
    "        --Vecset                  use Vecset representation",
    "        --Bitset                  use Bitset representation",
    "        --Stlset                  use STL set<> representation",
    "        --Invset                  use invset representation",
    "        --fds                     use field sensitive solver",
    NULL
  };

  unsigned int v;
  bool verbose_mode=false;
  bool no_rountev_chandra(false);
  bool projection_merging(false);
  unsigned int algorithm(OPT_PW3_POSCC2);
  unsigned int setrep(OPT_VECSET);
  unsigned int store_mode(OPT_FACTORED_STORE);

  while((v=getopt_long(argc,argv,"qsv:a:s:to:",long_options,NULL)) != -1) {
    switch(v) {      
    case OPT_HELP:
      cout << "usage: " << argv[0] << " [options] <constraint-file>" << endl;
      cout << "options:" << endl;
      for(char **ptr=descriptions;*ptr != NULL; ptr++) {
	cout << *ptr << endl;
      }    
      exit(1);
    case 'q':
      quiet_mode=true;
      break;
    case 's':
      summary_mode=true;
      break;
    case 't':
    case OPT_TIME:
      timing_mode=true;
      break;
    case OPT_AXS:
      full_axiomatic_solution=true;
      break;
    case OPT_VERB:
      verbose_mode=true;
      break;
    case 'o':
    case OPT_OUTPUT:
      output_file = new ofstream(optarg);
      break;
    case OPT_NORC:
      no_rountev_chandra=true;
      break;
    case OPT_PM:
      projection_merging=true;
      break;
    case OPT_HT:
    case OPT_HT_FDS:
    case OPT_PW1_LRF:
    case OPT_PW1_LRF_POSCC1:
    case OPT_PW1_LRF_POSCC2:
    case OPT_PW1_LRF_MNR:
    case OPT_PW1_LIFO_MNR:    
    case OPT_PW1_FIFO_MNR:
    case OPT_PWD_LRF:
    case OPT_PWD_LRF_POSCC1:
    case OPT_PWD_LRF_POSCC2:
    case OPT_PWD_LRF_MNR:
    case OPT_PW2_DFS:
    case OPT_PW2_POSCC2:
    case OPT_PW2_FDS:
    case OPT_PW3_DFS:
    case OPT_PW3_POSCC2:
    case OPT_PW3_FDS:
      algorithm=v;
      break;
    case OPT_VECSET:
      setrep = OPT_VECSET;
      break;
    case OPT_STLSET:
      setrep = OPT_STLSET;
      break;
    case OPT_BITSET:
      setrep = OPT_BITSET;
      break;
    case OPT_INVSET:
      setrep = OPT_INVSET;
      break;
    default:
      cerr << "unrecognised option" << endl;
      exit(2);
    }
  }

  if(optind >= argc) {
    cout << "usage: " << argv[0] << " [options] <constraint-file>" << endl;
    cout << "options:" << endl;
    for(char **ptr=descriptions;*ptr != NULL; ptr++) {
      cout << *ptr << endl;
    }    
    exit(1);
  }

  try {
    vector<constraint> constraints;
    vector<variable_info> varmap;
        
    // First, parse input file
    if(verbose_mode) { cerr << "Parsing ... " << flush; }

    pcs_front_end fe(argv[optind],varmap, constraints, false);
    while(fe.yyparse() != 1) {
    }

    if(verbose_mode) { 
      cerr << "OK [" << constraints.size() << " constraints, ";
      cerr << varmap.size() << " variables]" << endl; 
      cerr << "last variable is " << varmap[varmap.size()-1].name << endl;
    }

    
    // determine the "domain" size
    for(unsigned int i=0;i!=varmap.size();++i) {
      if(varmap[i].address_taken) {
	domain_size = i+1;
      }
    }    

    real_var_limit = varmap.size();
    if(projection_merging) {
      perform_projection_merging(constraints,varmap);
    }

    // now, build the constraint graphs
    graph_t graph;
    store<vector_set<unsigned int> > *solution;
    resize(varmap.size(),graph);
    if(verbose_mode) { cerr << "Building constraing graph ... " << flush; }    
    build_constraint_graph(graph,varmap,constraints);
    if(verbose_mode) { cerr << "OK" << endl; }

    if(!no_rountev_chandra) {
      if(verbose_mode) { cerr << "Performing Rountev-Chandra Elimination ... " << flush; }
      unsigned int t = eliminate_static_cycles(graph,varmap);      
      t += perform_rountev_chandra_elimination(graph,varmap);
      if(verbose_mode) { cerr << "OK [eliminated " << t << " nodes]" << endl; }
    }
    
    if(verbose_mode) { cerr << "Starting solver" << endl; }
    
    switch(setrep) {
    case OPT_VECSET:
      do_solve<domain_set_adaptor<vector_set<unsigned int> > >(algorithm,store_mode,constraints,varmap,graph);
      break;
    case OPT_BITSET:
      do_solve<domain_set_adaptor<bitset> >(algorithm,store_mode,constraints,varmap,graph);
      break;
    case OPT_STLSET:
      do_solve<domain_set_adaptor<std::set<unsigned int> > >(algorithm,store_mode,constraints,varmap,graph);
      break;
    case OPT_INVSET:
      do_solve<invset<vector_set<unsigned int> > >(algorithm,store_mode,constraints,varmap,graph);
      break;
    }
    
    if(verbose_mode) { cerr << "Solving finished" << endl; }
    
  } catch(runtime_error &e) {
    cerr << "pcs: " << e.what() << endl;
  } catch(...) {
    cerr << "pcs: internal failure" << endl;
  }
}







