#ifndef PCS_HPP
#define PCS_HPP

#include <utility>
#include <vector>
#include <set>
#include <string>
#include <stdexcept>
#include <vector_set.hpp>

// here we have all the various algorithmic options

// -- Main Algorithm types --
#define WORKLIST_ALGO 0
#define HEINTZE_TARDIEU_ALGO 1
#define PEARCE_ALGO 2
// -- Store Options --
#define FACTORED_STORE 3
#define SIMPLE_STORE 4
// -- Set Implementation options --
#define BITSET 5
#define VECTOR_SET 6
#define STLSET 7
// -- Cycle detection --
#define NO_CYCLE_DETECTION 8
#define STATIC_CYCLE_DETECTION 9
#define FULL_VARIABLE_SUBST 10
#define DFS_ONLINE_CYCLE_DETECTION 11
#define PKH_CYCLE_DETECTION 12
#define MNR_CYCLE_DETECTION 13
#define BATCH_MNR_CYCLE_DETECTION 14
// -- Worklist Iteration Order --
#define LIFO 15
#define FIFO 16
#define LRF 17
#define OTO1 18
#define OTO2 19
// -- Worklist Form Factor --
#define NEW_FORM_FACTOR 20
#define OLD_FORM_FACTOR 21
// -- Worklist Difference Propagation --
#define NO_DIFFERENCE_PROPAGATION 22
#define DIFFERENCE_PROPAGATION 23
// -- TRANSITIVE REDUCTION
#define NO_TRANSITIVE_REDUCTION 24
#define TRANSITIVE_REDUCTION 25

typedef enum {ADDRESSOF,DEREFERENCE,SCALAR} expr_type;

class expression {
protected:
  expr_type _type;
  unsigned int _val;
  unsigned int _offset;
public:
  expression(unsigned int v = 0) 
    : _type(SCALAR), _val(v), _offset(0) {    
  }
  
  expression(expr_type t, unsigned int v, unsigned int off = 0)
    : _type(t), _val(v), _offset(off) {
  }

  unsigned int var(void) const {
    return _val;
  }

  void set_type(expr_type t) {
    _type = t;
  }

  void set_offset(unsigned int o) {
    _offset = o;
  }

  unsigned int offset(void) const {
    return _offset;
  }

  bool addressof(void) const {
    return _type == ADDRESSOF;
  }

  bool dereferenced(void) const {
    return _type == DEREFERENCE;
  }

  bool scalar(void) const {
    return _type == SCALAR;
  }

  bool operator==(expression const &src) const {
    return _type == src._type &&
      _val == src._val &&
      _offset == src._offset;
  }

  bool operator<(expression const &src) const {
    if(_type == src._type) {
      if(_val == src._val) {
	return _offset < src._offset;
      } else {
	return _val < src._val;
      }
    } else {
      return _type < src._type;
    }
  } 
};

class constraint {
 protected:
  expression _lhs;
  expression _rhs;
 public:
  constraint(expression const &l, expression const &r)
    : _lhs(l), _rhs(r) {      
  }
  
  expression const &lhs(void) const {
    return _lhs;
  }

  expression const &rhs(void) const {
    return _rhs;
  }

  bool operator==(constraint const &src) const {
    return _lhs == src._lhs &&
      _rhs == src._rhs;
  }

  bool operator<(constraint const &src) const {
    if(_lhs < src._lhs) {
      return true;
    } else if(src._lhs < _lhs) {
      return false;
    } else {
      return _rhs < src._rhs;
    }
  }
};

class internal_contradiction : public std::runtime_error {
private:
  std::vector<constraint> _contradictions;
public:
  internal_contradiction(void) : runtime_error("default") {
  }
  internal_contradiction(std::vector<constraint> const &c)
    : runtime_error("internal contradiction"), _contradictions(c) {    
  }

  virtual ~internal_contradiction(void) throw() {
  }
  
  std::vector<constraint> const &contradictions(void) const {
    return _contradictions;
  }
};

class external_contradiction : public std::runtime_error {
private:
  std::vector<unsigned int> _contradictions;
public:
  external_contradiction(std::vector<unsigned int> c)
    : runtime_error("external contradiction"), _contradictions(c) {    
  }

  virtual ~external_contradiction(void) throw() {
  }
  
  std::vector<unsigned int> const &contradictions(void) const {
    return _contradictions;
  }
};

class variable_info {
public:
  std::string name;
  unsigned int end;  // one past last field of struct
  unsigned int node; // node representing this variable [confusing eh?]
  unsigned int rep;  // the number of variables this actually represents
  unsigned int address_taken : 1;
  unsigned int indirect_target : 1;
  vector_set<unsigned int> solution;
  vector_set<constraint> complicates;  // complex constraints involving this variable
  std::vector<unsigned int> variables;  // variables represented by this node.
public:
  variable_info(std::string const &s, unsigned int e, unsigned int n, unsigned int r)
    : name(s), end(e), node(n), rep(r), address_taken(false), indirect_target(false) {
  }

  variable_info(std::string const &s, unsigned int e, unsigned int n, unsigned int r,
		vector_set<constraint> const &c)
    : name(s), end(e), node(n), rep(r), address_taken(false), indirect_target(false), 
      complicates(c) {
  }
};

void condense_nodes(unsigned int, unsigned int, std::vector<variable_info> &);

std::string constraint2str(constraint const &, 
		      std::vector<variable_info> const &);

template<class T>
void convert_solution(std::vector<variable_info> &varmap, 		      
		      std::vector<T> &sol) {
  sol.resize(varmap.size());
  unsigned int index(0);
  for(std::vector<variable_info>::iterator i(varmap.begin());
      i!=varmap.end();++i,++index) {
    sol[index].insert(i->solution.begin(),i->solution.end());
    i->solution.clear(); // save memory
  }
}

template<class T>
void unconvert_solution(std::vector<T> &sol,
			std::vector<variable_info> &varmap) {
  assert(sol.size() == varmap.size());
  unsigned int index(0);
  for(typename std::vector<T>::iterator i(sol.begin());
      i!=sol.end();++i,++index) {
    varmap[index].solution.insert(i->old().begin(),i->old().end());
    varmap[index].solution.insert(i->delta().begin(),i->delta().end());
    i->clear(); // save memory
  }
}

class zero_edge_select {
public:
  bool operator()(vector_set<unsigned int> const &labels) {
    // return true if the labels contain zero.
    // note, the field-insensitive analyses just add
    // edges without any labels, so we also must account for that!
    return labels.empty() || *labels.begin() == 0;
  }
};

template<class T>
class set_union_func {
public:
  void operator()(vector_set<T> &to, 
		  vector_set<T> const &from) const {
    set_union(to,from);
  }
};

class n2c_wrapper {
private:
  std::vector<variable_info> &_varmap;
public:
  n2c_wrapper(std::vector<variable_info> &v)
    : _varmap(v) {
  }

  unsigned int &operator[](unsigned int i) {
    return _varmap[i].node;
  }
};

class checker_stats {
public:
  unsigned int visit_count;
  unsigned int complex_count;
  unsigned int union_count;
  unsigned int union_size;
  unsigned int iter_count;
  unsigned int post_iter_count;
  unsigned int new_edge_count;
public:
  checker_stats(void) 
    : visit_count(0), complex_count(0),
      union_count(0), union_size(0),
      iter_count(0), post_iter_count(0),
      new_edge_count(0) {
  }
};

template<class T> 
void sol_set_add(T &tmp, unsigned int label,
		 std::vector<variable_info> &varmap) {
  typename T::iterator iend(tmp.end());
  typename T::iterator j(tmp.begin());
  for(typename T::iterator i(j);
      i!=iend;++i) {
    unsigned int r(*i);
    // segfaults at next line when using vector_set.
    if(varmap[r].end <= (r+label)) {
      // otherwise ignore 
    } else {    
      const_cast<unsigned int &>(*j) = r+label; 
      ++j;
    }
  }

  tmp.erase(j,tmp.end());
}

#include "bitset.hpp"

template<> 
void sol_set_add(bitset &, unsigned int,
		 std::vector<variable_info> &);

#include "domain_set_adaptor.hpp"

template<> 
void sol_set_add(domain_set_adaptor<bitset> &s, unsigned int l,
		 std::vector<variable_info> &v); 

#include "invset.hpp"

template<> 
void sol_set_add(invset<vector_set<unsigned int> > &, unsigned int,
		 std::vector<variable_info> &);
#endif







