#ifndef PCS_FRONT_END_HPP
#define PCS_FRONT_END_HPP

#include <iostream>
#include <utility>
#include <map>
#include <vector>
#include <string>
#include <set>
#include <stdexcept>
#include <boost/lexical_cast.hpp>
#include "pcs.hpp"
#include "pcs_scanner.hpp"
#include "pcs_parser.hpp"

class pcs_front_end : public pcs_parser {
private:
  pcs_scanner _scanner;
  std::map<std::string,unsigned int> _lookup_map;
  std::vector<variable_info> &_varmap;
  std::vector<constraint> &_constraints;
  std::string _filename;
  bool _ftc;
public:
  pcs_front_end(std::string const &name, std::vector<variable_info> &vm, 
		std::vector<constraint> &cs, bool full_type_checking) 
    : _filename(name),_varmap(vm), _constraints(cs),
      _ftc(full_type_checking) {
    _scanner.line_no = 1;
    _scanner.col_no = 0;
    _scanner.yyin = fopen(name.c_str(),"r");
    
    if(_scanner.yyin == NULL) {
      throw std::runtime_error(std::string("file \"") +
			  name + "\" does not exist.");
    }
  }

  ~pcs_front_end(void) {
    fclose(_scanner.yyin);
  }

  virtual void yyerror(char *error) {
    throw std::runtime_error(_filename + ":" + boost::lexical_cast<std::string>(_scanner.line_no)
			+ ": " + error);    
  }
  virtual int yylex(void) {
    int token=_scanner.yylex(&yylval);
    return token;
  }
private:
  virtual bool lookup_var(std::string const &v, std::pair<unsigned int, unsigned int> &r) {
    std::map<std::string,unsigned int>::iterator i(_lookup_map.find(v));
    if(i == _lookup_map.end()) { return false; }

    r.first = i->second;
    r.second = _varmap[i->second].end;

    return true;
  }
  virtual void invalid_tsel(std::string const &err) {
    if(_ftc) {
      yyerror(const_cast<char*>(err.c_str()));
    }
  }
  virtual void parsed_constraint(constraint const &c) {
    expression const &rhs(c.rhs());
    expression const &lhs(c.lhs());
    if(rhs.addressof()) {
      // need to normalise this constraint
      unsigned int v(rhs.var() + rhs.offset());
      if(_varmap[rhs.var()].end <= v) {
	if(_ftc) {
	  yyerror("invalid use of type selector");
	}
      } else {
	constraint n(lhs,expression(ADDRESSOF,v));
	_constraints.push_back(n);

	while(v < _varmap[rhs.var()].end) {
	  _varmap[v++].address_taken = true;
	}
      }
    } else {
      if(!lhs.dereferenced() && rhs.dereferenced()) {
	_varmap[lhs.var()].indirect_target = true;
      }
      _constraints.push_back(c);
    }
  }
  virtual void parsed_type(std::string const &s, unsigned int i, unsigned int r) {
    unsigned int index(_varmap.size());
    if(i == 0) { i = 1; }
    unsigned int end(index + i);
    _varmap.push_back(variable_info(s,end,index,r));
    unsigned int j(0);
    _lookup_map[s] = index;
    while(--i > 0) {
      _varmap.push_back(variable_info(s + "#" + boost::lexical_cast<std::string>(++j),end,++index,r));
    }
  }
};

#endif








