// These functions simply implement inplace set union operations
// the set_union_diff function performs an inplace set union and
// produces the difference between the two

#ifndef SET_UNION_HPP
#define SET_UNION_HPP

#include <iostream>
#include <set>
#include "vector_set.hpp" // make sure that this is not used over vector_set's
#include "bitset.hpp"     // likewise

template<class T, class K>
bool set_union(std::set<T,K> &to, std::set<T,K> const &from) {
  if(static_cast<void const *>(&to) != static_cast<void const *>(&from)) {
    // quick sanity check
    unsigned int os(to.size());
    
    typename std::set<T,K>::iterator t(to.begin());
    typename std::set<T,K>::iterator tend(to.end());
    typename std::set<T,K>::const_iterator f(from.begin());
    typename std::set<T,K>::const_iterator fend(from.end());

    while(t != tend && f != fend) {
      if(*f < *t) {
	t = to.insert(t,*f);
	++f;
	tend = to.end(); // is this necessary ?
      } else if(!(*t < *f)) { ++f; }
      ++t;
    }

    while(f != fend) {
      // hinted set insertion is better
      // than range set insertion i think ...
      t = to.insert(t,*f);
      ++t;++f;
    }

    return to.size() != os;
  } else {
    return false;
  }
}  

#include "domain_set_adaptor.hpp"

template<class S>
void set_subtract(S &to, S const &from) {
  S res;
  set_difference(to.begin(),to.end(),from.begin(),from.end(),std::inserter(res,res.begin()));
  to = res;
}

template<class S>
bool set_union_diff(S &to, S const &from, S &diff) {
  if(&to != &from) {
    // quick sanity check
    unsigned int os(to.size());

    typename S::iterator t(to.begin());
    typename S::iterator tend(to.end());
    typename S::const_iterator f(from.begin());
    typename S::const_iterator fend(from.end());
    typename S::iterator d(diff.begin());
    
    while(t != tend && f != fend) {
      if(*f < *t) {
	t = to.insert(t,*f);
	d = diff.insert(d,*f);	
	++f;++d;
	tend = to.end(); // is this necessary ?
      } else if(!(*t < *f)) { ++f; }
      ++t;
    }

    while(f != fend) {
      // hinted set insertion is better
      // than range set insertion i think ...
      t = to.insert(t,*f);
      d = diff.insert(d,*f);	
      ++t;++f;++d;
    }

    return to.size() != os;
  } else {
    return false;
  }
}  

template<class T>
bool is_subset(T const &a, T const &b) {
  if(b.size() > a.size()) { return false; }
  else if(b.size() == a.size()) {
    return a == b;
  } else {
    if((b.size() * 2) < a.size()) {
      for(typename T::const_iterator i(b.begin());
	  i!=b.end();++i) {
	if(a.find(*i) == a.end()) {
	  return false;
	}
      }
      return true;
    } else {
      typename T::const_iterator i(a.begin());
      typename T::const_iterator j(b.begin());

      while(i!=a.end() && j!=b.end()) {
	if(*i < *j) { ++i; }
	else if(*j < *i) { 
	  return false;
	} else {
	  ++i;++j;
	}
      }
      if(j!=b.end()) { return false; }
      return true;
    }
  }
}

#endif
