#include <stdexcept>
#include <boost/lexical_cast.hpp>
#include "suif_helpers.h"

#define ERROR(x) throw runtime_error(string(x) + " (debug point " + __FILE__ + ":" + boost::lexical_cast<string,int>(__LINE__) + ")")

bool is_pointer_type(Type *t) {
  if(is_a<PointerType>(t)) {
    return true;
  } else if(is_a<QualifiedType>(t)) {
    QualifiedType *qt(to<QualifiedType>(t));
    return is_pointer_type(qt->get_base_type());
  } else {
    return false;
  }
}

Type *dereference_type(Type *t) {
  if(is_a<PointerType>(t)) {
    PointerType *ptr(to<PointerType>(t));
    return ptr->get_reference_type();
  } else if(is_a<QualifiedType>(t)) {
    QualifiedType *qt(to<QualifiedType>(t));
    return dereference_type(qt->get_base_type());
  } else {
    return NULL;
  }
}

bool is_struct_type(Type *t, StructType **st) {
  if(is_a<StructType>(t)) {
    if(st != NULL) { *st = to<StructType>(t); }
    return true;
  } else if(is_a<QualifiedType>(t)) {
    QualifiedType *qt(to<QualifiedType>(t));
    return is_struct_type(qt->get_base_type(), st);
  }
  return false;
}

bool is_pointer_to_struct(Type *t, StructType **st) {
  if(is_a<PointerType>(t)) {
    Type *ref_t(to<PointerType>(t)->get_reference_type());   
    return is_struct_type(ref_t,st);
  } else if(is_a<QualifiedType>(t)) {
    QualifiedType *qt(to<QualifiedType>(t));
    return is_pointer_to_struct(qt->get_base_type(), st);
  } else {
    return false;
  }
}

unsigned int get_bit_offset(FieldSymbol *ptr) {
  Expression *e(ptr->get_bit_offset());
  if(!is_a<IntConstant>(e)) {
    ERROR("Unable to decode bit offset");
  } else {
    IntConstant *ic(to<IntConstant>(e));
    if(!ic->get_value().is_c_unsigned_int()) {
      ERROR("Problem with IInteger (1)");
    } else {  
      return ic->get_value().c_unsigned_int();
    }
  }
}

unsigned int get_type_size_in_bytes(Type const *t) {
  if(is_a<QualifiedType>(t)) {
    return get_type_size_in_bytes(to<QualifiedType>(t)->get_base_type());
  } else if(is_kind_of<DataType>(t)) {
    DataType *ptr=to<DataType>(t);
    IInteger bs(ptr->get_bit_size());    
    if(bs.is_undetermined()) { ERROR("Cannot determine type size"); }
    if(!bs.is_c_unsigned_int()) { ERROR("Problem with IInteger (2)"); }
    return bs.c_unsigned_int() >> 3;
  } else {
    ERROR("Cannot handle type");
  }
}

unsigned int count_fields(StructType *s_type) {
  GroupSymbolTable *gst(s_type->get_group_symbol_table());
  unsigned int number(0),end;
  
  for(Iter<SymbolTableObject*> i=
	gst->get_symbol_table_object_iterator();
      i.is_valid();i.next()) {
    FieldSymbol *ptr(to<FieldSymbol>(i.current()));    
    StructType *s_type;
    unsigned int start(get_bit_offset(ptr));
    end = start + (get_type_size_in_bytes(ptr->get_type()) * 8);
    if(is_struct_type(ptr->get_type(),&s_type)) {
      number += count_fields(s_type);
    } else {
      ++number;
    }
  }

  // this is a little hack, as it seems that sometimes
  // the structure is larger than the number of fields
  // so I add a special extra field to account for this.

  if(end < (get_type_size_in_bytes(s_type)*8)) {
    number ++;
  }

  return number;
}

unsigned int get_field_number_from_bit_offset(unsigned int offset, StructType *s_type) {
  GroupSymbolTable *gst(s_type->get_group_symbol_table());
  unsigned int number(0);

  for(Iter<SymbolTableObject*> i=
	gst->get_symbol_table_object_iterator();
      i.is_valid();i.next()) {
    FieldSymbol *ptr(to<FieldSymbol>(i.current()));
    StructType *tmp;
    unsigned int start(get_bit_offset(ptr));
    unsigned int end(start + (get_type_size_in_bytes(ptr->get_type()) * 8));
 
    if(start <= offset && offset < end) {
      // definitely within this structure.
      if(is_struct_type(ptr->get_type(),&tmp)) {
	// this is a struct
	return number + get_field_number_from_bit_offset(offset - start,tmp);
      } 
      return number;
    } 

    if(is_struct_type(ptr->get_type(),&tmp)) {
      number += count_fields(tmp);
    } else {
      ++number;
    }
  }
  return number + 1;
}

bool contains_pointer_type(Type *t) {
  if(is_a<QualifiedType>(t)) {
    QualifiedType *qt(to<QualifiedType>(t));
    return contains_pointer_type(qt->get_base_type());
  } else if(is_kind_of<GroupType>(t)) {
    // now, we just iterate through members and
    // check each one ...
    GroupType *ptr(to<GroupType>(t));
    GroupSymbolTable *gst(ptr->get_group_symbol_table());

    for(Iter<SymbolTableObject*> i=
	  gst->get_symbol_table_object_iterator();
	i.is_valid();i.next()) {
      FieldSymbol *ptr(to<FieldSymbol>(i.current()));
      if(is_pointer_type(ptr->get_type())) {
	return true;
      } else if(contains_pointer_type(ptr->get_type())) {
	return true;
      }
    }
  } 

  return false;
}

bool is_char_type(Type *t) {
  if(is_a<QualifiedType>(t)) {
    QualifiedType *qt(to<QualifiedType>(t));
    return is_char_type(qt->get_base_type());
  } else if(is_a<IntegerType>(t)) {
    IntegerType *ptr(to<IntegerType>(t));
    // yeah, i've hardwired the size of char.
    if(ptr->get_is_signed() && ptr->get_bit_size() == 8) {
      return true;
    }
  }
  return false;
}

bool is_constant_bound(Expression *e1) {
  IntConstant *int1 = to<IntConstant>(e1);
  IInteger t1val(int1->get_value());

  if(t1val.is_undetermined()) {
    return false;
  } else {
    IntConstant *int1 = to<IntConstant>(e1);
    if(!int1->get_value().is_undetermined()) {
      return true;
    }
  } 
  return false;

}

bool is_char_array_type(Type *t) {
  if(is_a<QualifiedType>(t)) {
    QualifiedType *qt(to<QualifiedType>(t));
    return is_char_array_type(qt->get_base_type());
  } else if(is_a<ArrayType>(t)) {
    ArrayType *ptr(to<ArrayType>(t));
    return is_char_type(ptr->get_element_type()) &&
      is_constant_bound(ptr->get_lower_bound()) &&
      is_constant_bound(ptr->get_upper_bound());
  }
  return false;
}

ProcedureSymbol *is_static_call(Expression *e) {
  if(is_a<SymbolAddressExpression>(e)) {
    SymbolAddressExpression *ptr(to<SymbolAddressExpression>(e));
    return to<ProcedureSymbol>(ptr->get_addressed_symbol());
  }

  return NULL;
}

line_number get_line_number(AnnotableObject *s) {
  line_number r(UNKNOWN);
  if(s != NULL) {
    Annote *pAn = s->peek_annote("line");
    
    if(pAn != NULL) {
      IntegerBrick *tmp=to<IntegerBrick>(to<BrickAnnote>(pAn)->get_brick(0));
      IInteger v=tmp->get_value();
      r.type = EXACT;
      if(!v.is_c_unsigned_int()) { ERROR("Problem with IInteger (3)"); }
      r.value = v.c_unsigned_int();
    } 
  }

  return r;
}

int add_opcode=LString("add").get_ordinal();
int sub_opcode=LString("subtract").get_ordinal();
int mul_opcode=LString("multiply").get_ordinal();
int div_opcode=LString("divide").get_ordinal();
int rem_opcode=LString("remainder").get_ordinal();
int band_opcode=LString("bitwise_and").get_ordinal();
int bxor_opcode=LString("bitwise_xor").get_ordinal();
int bor_opcode=LString("bitwise_or").get_ordinal();
int ls_opcode=LString("left_shift").get_ordinal();
int rs_opcode=LString("right_shift").get_ordinal();
int net_opcode=LString("is_not_equal_to").get_ordinal();
int et_opcode=LString("is_equal_to").get_ordinal();
int get_opcode=LString("is_greater_than_or_equal_to").get_ordinal();
int let_opcode=LString("is_less_than_or_equal_to").get_ordinal();
int gt_opcode=LString("is_greater_than").get_ordinal();
int lt_opcode=LString("is_less_than").get_ordinal();
int lor_opcode=LString("logical_or").get_ordinal();
int land_opcode=LString("logical_and").get_ordinal();

/* unary opcodes */
int con_opcode=LString("convert").get_ordinal();
int lnot_opcode=LString("logical_not").get_ordinal();
int bnot_opcode=LString("bitwise_not").get_ordinal();
int neg_opcode=LString("negate").get_ordinal();

