#include <iostream>
#include <stdexcept>
#include <fstream>
#include <vector>
#include <deque>
#include <map>
#include <set>
#include <getopt.h>
#include <algorithm>
#include <boost/lexical_cast.hpp>

#include <suifkernel/suif_env.h>
#include <basicnodes/basic.h>
#include <suifnodes/suif.h>

#include "suif_helpers.h"

using namespace std;
#if __GNUC__ >= 3
using namespace __gnu_cxx; // needed for random_sample_n
#endif

#define ERROR(x) throw runtime_error(string(x) + " (debug point " + __FILE__ + ":" + boost::lexical_cast<string,int>(__LINE__) + ")")

#ifdef DEBUG
#define ASSERT(x) if(!(x)) ERROR("assertion failure")
#else
#define ASSERT(x)
#endif

using boost::lexical_cast;

typedef enum {ADDRESSOF,DEREFERENCE,EMPTY,OFFSET,SCALAR,INDIRECT_CALL} expr_type;

// Note. a DEREFERENCE with s_type != NULL is not the same as a FIELD_DEREFERENCE. Basically, this former
//       is used to describe the assigment of a struct to another.

// global variables - yes i know, put them into the generator class ...

bool merge_string_constants(false);
bool small_heap_model(false);
bool field_sensitive(true);
bool field_based(false);
bool partial_flow_sensitive(true);
bool fully_flow_sensitive(false);

// classes

class program_point {
public:
  string file;
  line_number line;  
public:
  program_point(void)
    : file("unknown"), line(UNKNOWN) {
  }
};

class var_type {
public:
  string var;
  unsigned int type;
  unsigned int rep;
public:
  var_type(string const &v, unsigned int t, unsigned int r) 
    : var(v), type(t), rep(r) {
  }

  bool operator<(var_type const &src) const {
    return var < src.var;
  }
};

class expression {
public:
  expr_type type;
  string var;          // variable name
  unsigned int offset; // add offset
  unsigned int subloc; // static selection
public:
  expression(string const &v, unsigned int off = 0, unsigned int sub = 0) 
    : type(SCALAR), var(v), offset(off), subloc(sub) {}

  expression(expr_type t, string const &v = "", unsigned int off = 0, unsigned int sub = 0)
    : type(t), var(v), offset(off), subloc(0) {
  }
  
  expression(expr_type t, unsigned int off)
    : type(t), offset(off), subloc(0) {
  }

  string str(void) const {
    if(type == ADDRESSOF) {
      string r(string("&") + var);
      if(offset != 0) { r += "+" + boost::lexical_cast<string>(offset); }
      return r;
    } else if(type == DEREFERENCE) {
      string r(string("*(") + var);
      if(subloc != 0) { r += "#" + boost::lexical_cast<string>(subloc); }
      if(offset != 0) {
	r += "+" + boost::lexical_cast<string>(offset);
      }
      return r + ")";
    } else if(type != SCALAR) {
      // this happens when assigning 
      // an integer to a pointer via casting.
      return string("&UNKNOWN");
    }

    string r(var);
    if(subloc != 0) {
      r += "#" + boost::lexical_cast<string>(subloc);
    } 
    if(offset != 0) {
      r += "+" + boost::lexical_cast<string>(offset);
    } 
    return r;
  }

  bool operator==(expression const &src) const {
    return type == src.type && var == src.var && 
      offset == src.offset && subloc == src.subloc;
  }
};

class constraint {
public:
  unsigned int block;
  expression lhs;
  expression rhs;
public:
  constraint(expression const &l, expression const &r) 
    : lhs(l), rhs(r) {
  }

  string str(void) const {
    return lhs.str() + " = " + rhs.str() + ";";
  }

  bool operator==(constraint const &src) const {
    return lhs == src.lhs && rhs == src.rhs;
  }
};

void update_line_number(line_number &l, AnnotableObject *obj) {
  line_number n(get_line_number(obj));
  if(n.type == EXACT) { l = n; }
  else if(l.type != UNKNOWN) {
    l.type = NEAR;
  }
}
   
string gen_temporary_name(void) {
  static unsigned int _tmpcounter(0);
  string r("tmp");
  r += lexical_cast<string,int>(_tmpcounter++);
  if(_tmpcounter == 0) { ERROR("too many temporaries created"); }
  return r;
}

void add_type(var_type const &e, set<var_type> &types) {
  // basically, it we try and allocate the same variable
  // twice then it corresponds to a duplicate allocation of a
  // field-based structure. this is recorded so that we 
  // know how many actual variables a struct variable 
  // represents
  pair<set<var_type>::iterator,bool> r(types.insert(e));
  if(!r.second) {
    const_cast<var_type&>(*r.first).rep += e.rep;
  }
}
  
class generator {
public:
  typedef vector<string>::const_iterator error_iterator;
protected:
  set<unsigned int> _lvars;  // local variables
  set<unsigned int> _fvars;  // static variables with file scope.
  set<unsigned int> _sfns;   // static functions
  map<StructType*,unsigned int> _structs_map; // for the field-based mode
  set<string> allocators;
  vector<string> _errors;
  string _file;
  ProcedureSymbol *_fn;
  unsigned int _fn_argcount;
  unsigned int malloc_counter;
  unsigned int max_struct_size;
  program_point _ppoint;
public:
  generator(set<string> &allocs, unsigned int mss) 
    : allocators(allocs), malloc_counter(0), max_struct_size(mss) {
  }

  void add_filescope(unsigned int f) {
    _fvars.insert(f);
  }

  void add_local(unsigned int l) {
    _lvars.insert(l);
  }

  void add_static_fn(ProcedureSymbol *fn) {
    _sfns.insert(fn->get_name().get_ordinal());
  }

  void new_file(string const &f) {
    _file=f;

    unsigned int pos(0);
    while((pos=_file.find(".",pos)) < _file.size()) {
      _file[pos]='_';
    }

    _file+="_";

    _fvars.clear();
    _lvars.clear();
    _sfns.clear();
  }

  error_iterator ebegin(void) const {
    return _errors.begin();
  }

  error_iterator eend(void) const {
    return _errors.end();
  }

  void new_function(ProcedureSymbol *f) {
    _fn=f;
    _lvars.clear();
    malloc_counter=0;
  }

  void evaluate(Statement *stmt, 
		vector<constraint> &cons,
		set<var_type> &types) {

    update_line_number(_ppoint.line,stmt);

    if(is_a<StatementList>(stmt)) {
      StatementList *ptr(to<StatementList>(stmt));
      for(Iter<Statement*> i(ptr->get_statement_iterator());
			     i.is_valid();i.next()) {
	evaluate(i.current(),cons,types);
      }
    } else if(is_a<StoreVariableStatement>(stmt)) {
      // StoreVariableStatements correspond to 
      // SCALAR assignments.
      StoreVariableStatement *ptr(to<StoreVariableStatement>(stmt));
      Type *type(ptr->get_destination()->get_type());
      if(is_pointer_type(type) || contains_pointer_type(type)) {
	if(!field_based || !is_pointer_to_struct(type,NULL)) {
	  // in field based mode it doesn't make sense
	  // to talk about pointers to structs.
	  expression rhs(evaluate(ptr->get_value(),cons,types));
	  expression lhs(variable_name(ptr->get_destination()));
	  do_assign(lhs,rhs,type,cons,types);
	}
      }
    } else if(is_a<StoreStatement>(stmt)) {
      // StoreStatements correspond to indirect assignments
      StoreStatement *ptr(to<StoreStatement>(stmt));
      Type *type(ptr->get_value()->get_result_type());
      Type *addr_type(ptr->get_destination_address()->get_result_type());

      if(is_pointer_type(type) || contains_pointer_type(type)) {
	if(!field_based || !is_pointer_to_struct(type,NULL)) {
	  // in field based mode it doesn't make sense
	  // to talk about pointers to structs.
	  expression address(evaluate(ptr->get_destination_address(),cons,types));
	  if(address.type != EMPTY && address.type != OFFSET) {
	    // basically, for the moment I ignore all
	    // pointer information generated from non-pointer
	    // variables. Which is certainly unsound!
	    expression value(evaluate(ptr->get_value(),cons,types));
	    address = do_dereference(address,cons,types);		  
	    
	    do_assign(address,value,type,cons,types);
	  }
	}
      }
    } else if(is_a<ReturnStatement>(stmt)) {
      ReturnStatement *ptr(to<ReturnStatement>(stmt));
      if(ptr->get_return_value() != NULL) {
	Type *type(ptr->get_return_value()->get_result_type());
	if(is_pointer_type(type) || contains_pointer_type(type)) {
	  if(!field_based || !is_pointer_to_struct(type,NULL)) {
	    expression rhs(evaluate(ptr->get_return_value(),cons,types));
	    expression lhs(function_name(_fn) + "_FN", 0, _fn_argcount);
	    do_assign(lhs,rhs,type,cons,types);
	  }
	}
      }
    } else if(is_a<CallStatement>(stmt)) {
      CallStatement *ptr(to<CallStatement>(stmt));
      ProcedureSymbol *fn=is_static_call(ptr->get_callee_address());
      expression fn_addr("junk");
      if(fn != NULL) {
	string fnname(fn->get_name().c_str());
	// quick check for the internal functions
	if((fnname == "__internal_alloc")
	   || (!small_heap_model && allocators.find(fnname) != allocators.end())) {
	  // this is a special case

	  string heap_name(function_name(_fn)+"_HEAP"+
			   boost::lexical_cast<string>(malloc_counter++));
	  
	  add_type(var_type(heap_name,field_sensitive ? max_struct_size : 1, field_sensitive ? 1 : max_struct_size),types);
	  if(ptr->get_destination() != NULL) {
	    expression lhs(variable_name(ptr->get_destination()));
	    expression rhs(ADDRESSOF,heap_name);
	    cons.push_back(constraint(lhs,rhs));
	  }
	  return; 
	} 
      } 
      
      fn_addr = evaluate(ptr->get_callee_address(),cons,types);

      Type *tmp = dereference_type(ptr->get_callee_address()->get_result_type());
      if(!is_kind_of<ProcedureType>(tmp)) { _errors.push_back(ppoint2str() + "problem with callee address type"); }
      else if(!is_a<CProcedureType>(tmp)) { _errors.push_back(ppoint2str() + "cannot cope with non-c procedures"); }
      CProcedureType *fn_type = to<CProcedureType>(tmp);
      string fn_name("unknown"); // for debugging purposes only
      if(fn_addr.type == ADDRESSOF) { fn_name = fn_addr.var; }
      if(fn_addr.type != EMPTY && fn_addr.type != EMPTY) {
	// basically, for the moment I ignore all
	// pointer information generated from non-pointer
	// variables. Which is certainly unsound!
	
	// first, do parameters 
	unsigned int index(0);
	unsigned int argnum(0);
	
	// a quick word about why I use argnum. Basically, I would like
	// to obtain the argument index from the function type.  see
	// get_argument_number for more information on this.  However
	// the SUIF 2 front end has a bug which means that function
	// types are incorrectly defined when K&R style parameter
	// declarations are used.
	
	for(Iter<Expression*> i(ptr->get_argument_iterator());
	    i.is_valid();i.next(),index++) {
	  Type *type(i.current()->get_result_type());
	  
	  if(is_pointer_type(type)
	     || contains_pointer_type(type)) {
	    
	    if(!field_based || !is_pointer_to_struct(type,NULL)) {
	      // again, it doesn't make sense to talk about pointers
	      // to structs in field_based mode.
	      expression lhs(do_dereference(fn_addr,cons,types));
	      lhs = do_select_subloc(lhs,argnum,cons);
	      expression rhs(evaluate(i.current(),cons,types));
	      do_assign(lhs,rhs,type,cons,types);
	    
	      StructType *s_type;
	      if(field_sensitive && is_struct_type(type,&s_type) && contains_pointer_type(type)) {
		argnum += count_fields(s_type);
	      } else  {
		++argnum;
	      }      	
	    }
	  }
	}
	
	// now, do return value
	if(ptr->get_destination() != NULL && 
	   (is_pointer_type(ptr->get_destination()->get_type())
	    || contains_pointer_type(ptr->get_destination()->get_type()))) {
	  Type *type(ptr->get_destination()->get_type());
	  if(!field_based || !is_pointer_to_struct(type,NULL)) {
	    // again, it doesn't make sense to talk about pointers
	    // to structs in field_based mode.	    
	    expression lhs(variable_name(ptr->get_destination()));
	    expression rhs(do_dereference(fn_addr,cons,types));
	    rhs = do_select_subloc(rhs,argnum,cons);
	    do_assign(lhs,rhs,type,cons,types);
	  }
	}
      }
    } else if(is_a<EvalStatement>(stmt)) {
      // don't do anything ...
    } else if(is_a<VaStartStatement>(stmt)) {
      // shoud do something ?
    } else if(is_a<VaEndStatement>(stmt)) {
      // shoud do something ?
    } else if(is_a<LabelLocationStatement>(stmt)) {
      // don't do anything ...
    } else if(is_a<BranchStatement>(stmt)) {
      // don't do anything ...
    } else if(is_a<JumpStatement>(stmt)) {
      // don't do anything ...
    } else {
      _errors.push_back(ppoint2str() + string( "Unknown statement \"") + 
			stmt->getClassName().c_str() + "\" encountered");
    }
  }
  
  void evaluate(expression const &lhs, ValueBlock *v, vector<constraint> &cons,
		set<var_type> &types) {
    // convert value block into constraints
    if(v != NULL && !is_a<UndefinedValueBlock>(v)) {
      // as this is a flow insensitive analysis
      // we must ignore undefined initialisation
      if(is_a<ExpressionValueBlock>(v)) {
	ExpressionValueBlock *ptr(to<ExpressionValueBlock>(v));
	expression rhs(evaluate(ptr->get_expression(),cons,types));
	if(rhs.type != EMPTY && rhs.type != OFFSET) {
	  cons.push_back(constraint(lhs,rhs));
	}
      } else if(is_a<MultiValueBlock>(v)) {
	MultiValueBlock *ptr(to<MultiValueBlock>(v));
	unsigned int index(0);
	for(Iter<indexed_list<IInteger,ValueBlock*>::pair> i=
	      ptr->get_sub_block_iterator();i.is_valid();i.next(),index++) {
	  expression tmp(do_select_subloc(lhs,(field_sensitive || field_based) ? index : 0,cons));
	  evaluate(tmp,i.current().second,cons,types);
	}

      } else {
	_errors.push_back(ppoint2str() + " Unrecognized value block");
      }
    }
  }

  expression evaluate(Expression *e, vector<constraint> &cons,
		      set<var_type> &types) {
    ASSERT(e != NULL);
    StructType *s_type;
    if(field_based && is_pointer_to_struct(e->get_result_type(),&s_type)) {
      return get_struct_address(s_type);
    } else if(is_a<IntConstant>(e)) {
      if(is_pointer_type(e->get_result_type()) 
	 && to<IntConstant>(e)->get_value() == 0) {
	// this corresponds to the NULL pointer
	return expression(ADDRESSOF,"NULL");
      } else {
	IInteger v(to<IntConstant>(e)->get_value());
	if((field_sensitive || field_based) && v.is_c_unsigned_int()) { return expression(OFFSET,v.c_unsigned_int()); }
	else { return expression(EMPTY); }
      }
    } else if(is_a<SymbolAddressExpression>(e)) {
      SymbolAddressExpression *ptr(to<SymbolAddressExpression>(e));
      // here is the point that we must 
      if(merge_string_constants && 
	 is_char_array_type(ptr->get_addressed_symbol()->get_type())) {
	// ok, the addressed location has fixed size char array type.
	// now, check that it's assigned some multivals and contains
	// Tmp in the name. this is a bit of hack, i know. but it works
	// the better alternative would be to use an initial suif pass
	// which marked these variables which have no name with a special
	// qualification
	string name(ptr->get_addressed_symbol()->get_name().c_str());
	if(name.find("Tmp",0) < name.size()) {
	  VariableSymbol *sym(to<VariableSymbol>(ptr->get_addressed_symbol()));
	  VariableDefinition *vdef(sym->get_definition());
	  if(vdef != NULL 
	     && is_a<MultiValueBlock>(vdef->get_initialization())) {
	    // yeah, ok i'm satisfied. it IS a string constant
	    return expression(ADDRESSOF,"MERGED_STRING_CONSTANTS");
	  }
	}
      }
      Symbol *sym(ptr->get_addressed_symbol());
      if(is_a<ProcedureSymbol>(sym)) {
	return expression(ADDRESSOF,function_name
			  (to<ProcedureSymbol>(sym)) + "_FN");
      } else {
	return expression(ADDRESSOF,variable_name(sym));
      }
    } else if(is_a<LoadVariableExpression>(e)) {    
      LoadVariableExpression *ptr(to<LoadVariableExpression>(e));
      VariableSymbol *sym(ptr->get_source());
      if(is_pointer_type(sym->get_type()) 
	 || contains_pointer_type(sym->get_type())) {		
	return expression(variable_name(sym));
      }
    } else if(is_a<LoadExpression>(e)) {
      LoadExpression *ptr(to<LoadExpression>(e));
      Type *type(ptr->get_result_type());
      if(is_pointer_type(type) || contains_pointer_type(type)) {
	Type *addr_type(ptr->get_source_address()->get_result_type());
	expression rhs(evaluate(ptr->get_source_address(),cons,types));
	if(rhs.type != EMPTY && rhs.type != OFFSET) {	  
	  // basically, for the moment I ignore all
	  // pointer information generated from non-pointer
	  // variables. Which is certainly unsound!
	  
	  return do_dereference(rhs,cons,types);
	}
      }
    } else if(is_a<UnaryExpression>(e)) {
      UnaryExpression *ptr(to<UnaryExpression>(e));
      int opcode(ptr->get_opcode().get_ordinal());
      if(is_pointer_type(e->get_result_type())) {
	expression rhs(evaluate(ptr->get_source(),cons,types));      
	if(con_opcode == opcode) { 
	  return rhs;
	}  else {
	  _errors.push_back(ppoint2str() + string(" Unsupported opcode (")
			    + ptr->get_opcode().c_str() + ")");
	}
      }
    } else if(is_a<BinaryExpression>(e)) {
      BinaryExpression *ptr(to<BinaryExpression>(e));
      int opcode(ptr->get_opcode().get_ordinal());
      expression lhs(evaluate(ptr->get_source1(),cons,types));
      expression rhs(evaluate(ptr->get_source2(),cons,types));
      
      if(is_pointer_type(e->get_result_type())) {
	Type *lhs_type(ptr->get_source1()->get_result_type());
	Type *rhs_type(ptr->get_source2()->get_result_type());
	if(opcode == add_opcode) { 
	  if(lhs.type == OFFSET || lhs.type == EMPTY) {
	    StructType *s_type;
	    // now, final check to make sure the point
	    // increment is to within the type pointed to
	    if(is_pointer_to_struct(rhs_type,&s_type) &&
	       lhs.offset < get_type_size_in_bytes(s_type)) {
	      // int + ptr
	      return do_add_offset(rhs,get_field_number_from_bit_offset(lhs.offset*8,s_type),cons,types);
	    } else {
	      return rhs;
	    }
	  } else if(rhs.type == OFFSET || rhs.type == EMPTY) {
	    StructType *s_type;
	    // now, final check to make sure the point
	    // increment is to within the type pointed to
	    if(is_pointer_to_struct(lhs_type,&s_type) &&
	       rhs.offset < get_type_size_in_bytes(s_type)) {
	      // ptr + int
	      return do_add_offset(lhs,get_field_number_from_bit_offset(rhs.offset*8,s_type),cons,types);
	    } else {
	      return lhs;
	    }
	  } else { ERROR("Internal failure"); }
	} else if(opcode == sub_opcode) { 
	  if(lhs.type != EMPTY && rhs.type != EMPTY && rhs.type != OFFSET) {
	    ERROR("Internal failure");
	  } else {
	    return lhs;
	  }
	} else {
	  _errors.push_back(ppoint2str() + string(" Unsupported opcode (")
			    + ptr->get_opcode().c_str() + ")");
	}
      } else if(lhs.type == OFFSET && rhs.type == OFFSET) {
	if(opcode == add_opcode) {
	  lhs.offset += rhs.offset;
	} else if(opcode == sub_opcode) {
	  lhs.offset -= rhs.offset; 
	} else if(opcode == mul_opcode) {
	  lhs.offset *= rhs.offset;
	} 
      }
    } else if(is_a<ArrayReferenceExpression>(e)) {
      ArrayReferenceExpression *ptr(to<ArrayReferenceExpression>(e));
      expression lhs(evaluate(ptr->get_base_array_address(),cons,types));
      return lhs;
    }
    
    return expression(EMPTY);
  }

  void do_procedure_defn(ProcedureDefinition *pd, 
			 vector<constraint> &cons,
			 set<var_type> &types) {
    
    SymbolTable *local_symtab(pd->get_symbol_table());
    
    // first add elements to symbol table
    new_function(pd->get_procedure_symbol());  
    for(Iter<SymbolTableObject*> i=local_symtab->get_symbol_table_object_iterator()
	  ;i.is_valid();i.next()) {
      if(is_a<VariableSymbol>(i.current())) {
	VariableSymbol *pSym=to<VariableSymbol>(i.current());
	add_local(pSym->get_name().get_ordinal());
	if(pSym->get_is_address_taken() || is_pointer_type(pSym->get_type()) 
	   || contains_pointer_type(pSym->get_type())) {
	  add_type(var_type(variable_name(pSym),get_type(pSym->get_type()),get_rep(pSym->get_type())),types);
	}
      }
    }
    
    unsigned int index(0);
    unsigned int argnum(0);
    // now, add the parameters
    for(Iter<ParameterSymbol*> i(pd->get_formal_parameter_iterator());
	i.is_valid();i.next(),index++) {
      ParameterSymbol *pSym=i.current();
      Type *type(i.current()->get_type());
      add_local(pSym->get_name().get_ordinal());
      
      if(is_pointer_type(type) || contains_pointer_type(type)) {
	if(!field_based || !is_pointer_to_struct(type,NULL)) {
	  // again, it doesn't make sense to talk about pointers
	  // to structs in field_based mode.

	  expression rhs(function_name(_fn) + "_FN",0,argnum);
	  expression lhs(variable_name(pSym));
	  do_assign(lhs,rhs,type,cons,types);
	  add_type(var_type(variable_name(pSym),get_type(pSym->get_type()),get_rep(pSym->get_type())),types); 
	
	  StructType *s_type;
	  if(field_sensitive && is_struct_type(type,&s_type) && contains_pointer_type(type)) {
	    argnum += count_fields(s_type);
	  } else  {
	    ++argnum;
	  }      	
	}
      } else if(pSym->get_is_address_taken()) {
	// non pointer variable whose address is taken
	add_type(var_type(variable_name(pSym),get_type(pSym->get_type()),get_rep(pSym->get_type())),types);
      }
    }
    
    _fn_argcount = argnum;
    
    // the whole point of the next chunk
    // is to determine whether this function has a 
    // return value and, if so, to increment the argnum
    // by the right amount amount ...

    Type *t = pd->get_procedure_symbol()->get_type();
    if(!is_a<CProcedureType>(t)) {
      _errors.push_back(ppoint2str() + " cannot cope with non-c procedures");
    } else {
      CProcedureType *tmp(to<CProcedureType>(t));
      unsigned int count(get_argument_count(tmp,field_sensitive));
      if(tmp->get_result_type() != NULL &&
	 !is_a<VoidType>(tmp->get_result_type())
	 && (is_pointer_type(tmp->get_result_type()) || contains_pointer_type(tmp->get_result_type()))) {

	if(!field_based || !is_pointer_to_struct(tmp->get_result_type(),NULL)) {
	  // again, it doesn't make sense to talk about pointers
	  // to structs in field_based mode.	  
	  argnum += get_type(tmp->get_result_type());
	}
      }
    }
    
    // I set the rep here to be zero because otherwise I cannot correctly
    // expand the number of variables representing a function definition.
    add_type(var_type(function_name(pd->get_procedure_symbol()) + "_FN",			       
			     argnum,0),types);
    
    // now, do the local var definitions
    DefinitionBlock *db(pd->get_definition_block());
    for(Iter<VariableDefinition*> i(db->get_variable_definition_iterator());
	i.is_valid();i.next()) {
      Symbol *sym(i.current()->get_variable_symbol());
      
      if(is_pointer_type(sym->get_type()) 
	 || contains_pointer_type(sym->get_type())) {
	update_line_number(_ppoint.line,i.current()->get_initialization());	
	evaluate(expression(variable_name(sym)),
		 i.current()->get_initialization(),cons,types);
      }
    }
    
    // finally, evaluate the function body
    evaluate(to<Statement>(pd->get_body()),cons,types);
  }

  void do_fileblock(FileBlock *fb,  
		    vector<constraint> &cons,
		    set<var_type> &types) {
    
    new_file(fb->get_source_file_name().c_str());
    _ppoint.file = string(fb->get_source_file_name().c_str());
    
    SymbolTable *table = fb->get_symbol_table();  
    // first, we add the variables to the generator
    for(Iter<SymbolTableObject*> i=table->get_symbol_table_object_iterator();i.is_valid();i.next()) {
      if(is_a<VariableSymbol>(i.current())) {
	VariableSymbol *sym(to<VariableSymbol>(i.current()));
	update_line_number(_ppoint.line,sym);
	add_filescope(sym->get_name().get_ordinal());
	if(is_pointer_type(sym->get_type()) || 
	   contains_pointer_type(sym->get_type()) ||
	   sym->get_is_address_taken()) {
	  add_type(var_type(variable_name(sym),get_type(sym->get_type()),get_rep(sym->get_type())),types);
	}
      } else if(is_a<ProcedureSymbol>(i.current())) {
	update_line_number(_ppoint.line,i.current());
	add_static_fn(to<ProcedureSymbol>(i.current()));
      }
    }
    // second we evaluate the variable definitions
    DefinitionBlock *db(fb->get_definition_block());
    for(Iter<VariableDefinition*> i(db->get_variable_definition_iterator());
	i.is_valid();i.next()) {
      Symbol *sym(i.current()->get_variable_symbol());
      if(is_pointer_type(sym->get_type()) 
	 || contains_pointer_type(sym->get_type())) {
	update_line_number(_ppoint.line,sym);
	StructType *s_type;
	evaluate(expression(variable_name(sym)),
		 i.current()->get_initialization(),cons,types);
      }
    }
    // third, we evaluate the procedure definitions
    for(Iter<ProcedureDefinition*> i(db->get_procedure_definition_iterator());
	i.is_valid();i.next()) {
      do_procedure_defn(i.current(),cons,types);
  } 
    
    _ppoint.file = "UNKNOWN";
    _ppoint.line = line_number(UNKNOWN);
  }

  unsigned int get_type(Type *t) {
    StructType *s_type;
    Type *elem;
    
    if(is_a<QualifiedType>(t)) {
      QualifiedType *ptr(to<QualifiedType>(t));
      return get_type(ptr->get_base_type());
    } else if(is_a<ArrayType>(t)) {
      ArrayType *ptr(to<ArrayType>(t));
      return get_type(ptr->get_element_type());
    } else if(is_struct_type(t,&s_type)) {
      return (field_sensitive || field_based) ? count_fields(s_type) : 1;
    } else if(is_kind_of<ProcedureType>(t)) {
      ERROR("internal failure");
      if(!is_a<CProcedureType>(t)) {
	_errors.push_back(ppoint2str() + " cannot cope with non-c procedures");
	return 1;
      }
      CProcedureType *tmp(to<CProcedureType>(t));
      unsigned int count(get_argument_count(tmp,field_sensitive));
      if(tmp->get_result_type() != NULL &&
	 !is_a<VoidType>(tmp->get_result_type())) {
	return count + 1;
      }
      return count;
    }
    
    return 1;
  }

  unsigned int get_rep(Type *t) {
    StructType *s_type;
    Type *elem;
    
    if(is_a<QualifiedType>(t)) {
      QualifiedType *ptr(to<QualifiedType>(t));
      return get_rep(ptr->get_base_type());
    } else if(is_a<ArrayType>(t)) {
      ArrayType *ptr(to<ArrayType>(t));
      return get_rep(ptr->get_element_type());
    } else if(is_struct_type(t,&s_type)) {
      return !field_sensitive && !field_based ? count_fields(s_type) : 1;
    } else if(is_kind_of<ProcedureType>(t)) {
      ERROR("internal failure");
      if(!is_a<CProcedureType>(t)) {
	_errors.push_back(ppoint2str() + " cannot cope with non-c procedures");
	return 1;
      }
      CProcedureType *tmp(to<CProcedureType>(t));
      unsigned int count(get_argument_count(tmp,field_sensitive));
      if(tmp->get_result_type() != NULL &&
	 !is_a<VoidType>(tmp->get_result_type())) {
	return count + 1;
      }
      return count;
    }
    
    return 1;
  }

  expression get_struct_address(StructType *ptr) {
    map<StructType*,unsigned int>::iterator i;
    i = _structs_map.insert(std::make_pair(ptr,_structs_map.size())).first;	
    return expression(ADDRESSOF,string("STRUCT_") + boost::lexical_cast<string>(i->second));
  }

  string variable_name(Symbol *sym) {
    assert(sym != NULL);
    StructType *ptr; 

    if(is_struct_type(sym->get_type(),&ptr) &&
       field_based) {
      // basically in field_based mode each 
      // field of a given struct is represented
      // by a single variable which covers all
      // instances of that variable.
      map<StructType*,unsigned int>::iterator i;
      i = _structs_map.insert(std::make_pair(ptr,_structs_map.size())).first;	
      
      return string("STRUCT_") + boost::lexical_cast<string>(i->second);
    } else {    
      unsigned int ord(sym->get_name().get_ordinal());
      string r;
      
      if(_lvars.find(ord) != _lvars.end()) {
	// variable has function scope
	r = function_name(_fn) + "_" + sym->get_name().c_str();
      } else if(_fvars.find(ord) != _fvars.end()) {
	// variable has filescope
	r = _file + "_" + sym->get_name().c_str();
      } else {
	// variable is global
	r = string(sym->get_name().c_str());
      }
      return r; 
    }
  }
  
  string function_name(ProcedureSymbol *fn) {
    if(_sfns.find(fn->get_name().get_ordinal()) != _sfns.end()) {
      return _file + "_" + fn->get_name().c_str();
    } else {
      return string(fn->get_name().c_str());
    }
  }  
private:
  expression do_add_offset(expression expr, unsigned int offset, vector<constraint> &cons,
			   set<var_type> &types) {
    if(expr.type == DEREFERENCE) {
      expression tmp(gen_temporary_name());
      cons.push_back(constraint(tmp,expr));      
      expr = tmp;
      add_type(var_type(tmp.var,1,0),types);    
    }
    expr.offset += offset;
    return expr;
  }
  
  expression do_select_subloc(expression expr, unsigned int offset, vector<constraint> &cons) {
    if(expr.type == DEREFERENCE) {  
      //      if(expr.subloc != 0) { _errors.push_back(ppoint2str() + string(" invalid use of do_select_subloc")); }
      expr.offset += offset;
    } else {
      if(expr.offset != 0) { _errors.push_back(ppoint2str() + string(" invalid use of do_select_subloc")); }
      expr.subloc += offset;
    }
    return expr;
  }
  
  expression do_dereference(expression expr, vector<constraint> &cons, 
			    set<var_type> &types) {
    switch(expr.type) {
    case ADDRESSOF: 
	// deref(&a) => a
	expr.type = SCALAR;
	// covers deref(&a+1) case ...
	expr.subloc += expr.offset;
	expr.offset = 0;
      break;
    case SCALAR:
      // deref(a) => *a    
      expr.type = DEREFERENCE;
      break;
    case DEREFERENCE:
      {
	// deref(*a) => tmp=*a, *tmp
	expression tmp_lhs(gen_temporary_name());
	add_type(var_type(tmp_lhs.var,1,0),types);
	
	cons.push_back(constraint(tmp_lhs,expr));
	expr.var = tmp_lhs.var;
	expr.type = DEREFERENCE;
	expr.offset = 0;
	expr.subloc = 0;
      }
      break;
    case OFFSET:
      _errors.push_back(ppoint2str() + " Unknown expression encountered (type OFFSET)");
      break;
    case EMPTY:
      _errors.push_back(ppoint2str() + " Unknown expression encountered (type EMPTY)");
      break;
    }	
    
    return expr;
  }
  
  void do_assign(expression lhs, expression rhs, Type *type, 
		 vector<constraint> &cons, set<var_type> &types) {
    StructType *s_type;

    if(field_sensitive && is_struct_type(type,&s_type)) {
      GroupSymbolTable *gst(s_type->get_group_symbol_table());
      
      unsigned int num(0);
      for(Iter<SymbolTableObject*> i=
	    gst->get_symbol_table_object_iterator();
	  i.is_valid();i.next()) {
	FieldSymbol *ptr(to<FieldSymbol>(i.current()));
	Type *field_type(ptr->get_type());
	StructType *field_s_type;
	
	if(is_pointer_type(field_type) || contains_pointer_type(field_type)) {
	  do_assign(lhs,rhs,field_type,cons,types);
	} 
	
	if(is_struct_type(field_type,&field_s_type)) {
	  // should cause problems if lhs or rhs
	  // is a pointer .. easy to fix though
	  lhs = do_select_subloc(lhs,count_fields(field_s_type),cons);
	  rhs = do_select_subloc(rhs,count_fields(field_s_type),cons);
	} else {
	  lhs = do_select_subloc(lhs,1,cons);
	  rhs = do_select_subloc(rhs,1,cons);
	}
      }
    } else if(lhs.type == DEREFERENCE && rhs.type == DEREFERENCE) {
      expression tmp(gen_temporary_name());
      add_type(var_type(tmp.var,1,0),types);
      cons.push_back(constraint(tmp,rhs));
      cons.push_back(constraint(lhs,tmp));
    } else {
      if(!field_sensitive) {
	if(rhs.type == SCALAR && rhs.offset != 0) {
	  // positive weights constraints are only allowed
	  // for the field_sensitive system
	  string msg(" internal failure");
	  _errors.push_back(ppoint2str() + msg);
	}
      } 
      
      if(!field_based || !is_struct_type(type,NULL)) {
	cons.push_back(constraint(lhs,rhs));
      }
    }
  }

  unsigned int get_argument_number(ProcedureType *_type, unsigned int index, bool expand_structs, 
				   string fn_name) {
    if(!is_a<CProcedureType>(_type)) {
      ERROR("Cannot cope with non-c procedures");
    } else {
      CProcedureType *type(to<CProcedureType>(_type));
      if(type->get_argument_count() <= index) { 
	string msg(string(" problem with parameters in call to \"" + fn_name + "\" - i=") + 
		   boost::lexical_cast<string>(index) + 
		   ", count=" + boost::lexical_cast<string>(type->get_argument_count()));
	_errors.push_back(ppoint2str() + msg);
	return 0;
      }
      unsigned int num(0);
      
      for(unsigned int i=0;i!=index;++i) {
	Type *tmp = type->get_argument(i);
	StructType *s_type;
	
	if(expand_structs && is_struct_type(tmp,&s_type) && contains_pointer_type(tmp)) {
	  num += count_fields(s_type);
	} else if(is_pointer_type(tmp) || contains_pointer_type(tmp)) {
	  ++num;
	}      
      }
      
      return num;
    }    
  }
  
  unsigned int get_argument_count(ProcedureType *_type, bool expand_structs) {
    if(!is_a<CProcedureType>(_type)) {
      ERROR("Cannot cope with non-c procedures");
    } else {
      CProcedureType *type(to<CProcedureType>(_type));
      unsigned int num(0);
      
      for(unsigned int i=0;i!=type->get_argument_count();++i) {
	Type *tmp = type->get_argument(i);
	StructType *s_type;
	
	if(expand_structs && is_struct_type(tmp,&s_type) && contains_pointer_type(tmp)) {
	  num += count_fields(s_type);
	} else if(is_pointer_type(tmp) || contains_pointer_type(tmp)) {
	  ++num;
	}      
      }
      
      return num;
    }    
  }
  

  string ppoint2str(void) {
    string r(_ppoint.file); 
    r += ":";
    if(_ppoint.line.type == EXACT) {
      r += boost::lexical_cast<string>(_ppoint.line.value);
    } else if(_ppoint.line.type == NEAR) {
      r += "~" +  boost::lexical_cast<string>(_ppoint.line.value);
    } else {
      r += "?";
    }

    return r;
  }
};

#define OPT_HELP 0
#define OPT_MSC 1
#define OPT_SHM 2
#define OPT_FI 3
#define OPT_FB 4
#define OPT_ADDALLOCS 5
#define OPT_MSSOVER 6

int main(int argc, char *argv[]) {
  struct option long_options[]={
    {"help",no_argument,NULL,OPT_HELP},
    {"merge-string-constants",no_argument,NULL,OPT_MSC},
    {"small-heap-model",no_argument,NULL,OPT_SHM},
    {"field-insensitive",no_argument,NULL,OPT_FI},
    {"field-based",no_argument,NULL,OPT_FB},
    {"add-allocators",required_argument,NULL,OPT_ADDALLOCS},
    {"heap-object-size",required_argument,NULL,OPT_MSSOVER},
    NULL
  };
  
  char *descriptions[]={
    "        --help                    display this information",
    "        --merge-string-constants  merge string constants into one object",
    "        --small-heap-model        use a single heap object only as model",
    "        --field-insensitive       generate field-insensitive constraints",
    "        --field-based             generate field-based constraints",
    "        --add-allocators=\"<x>\"  where x is comma seperated list of functions",
    "                                  to use along with the usual malloc etc as functions",
    "                                  which return heap allocated memory",
    "        --heap-object-size=<x>    override number of fields for heap objects",
    NULL
  };

  set<string> allocators;
  unsigned int v,mss_value;
  bool mss_override=false;

  // now add the set of default allocator functions
  allocators.insert("malloc");
  allocators.insert("calloc");
  allocators.insert("alloca");

  while((v=getopt_long(argc,argv,"",long_options,NULL)) != -1) {
    switch(v) {      

    case OPT_HELP:
      cout << "usage: " << argv[0] << " [options] <suif file>" << endl;
      cout << "options:" << endl;
      for(char **ptr=descriptions;*ptr != NULL; ptr++) {
	cout << *ptr << endl;
      }    
      exit(1);
    case OPT_MSC:
      merge_string_constants=true;
      break;
    case OPT_SHM:
      small_heap_model=true;
      break;
    case OPT_FI:
      field_sensitive=false;
      field_based=false;
      break;
    case OPT_FB:
      field_based=true;
      field_sensitive=false;
      break;
    case OPT_ADDALLOCS:
      {
	char *tmp = optarg;
	char *tok;
	while((tok = strtok(tmp,",")) != NULL) {
	  tmp=NULL;
	  allocators.insert(string(tok));
	}
      }
      break;
    case OPT_MSSOVER:
      mss_override=true;
      mss_value = atoi(optarg);
      break;
    default:
      exit(2);
    }
  }

  if((argc-optind) < 1) {
    cout << "usage: suifcgen [options] <suifile>" << endl;
    exit(1);
  }

  SuifEnv suif;

  suif.init();
  suif.require_module("basicnodes");
  suif.require_module("suifnodes"); 
  suif.require_module("cfenodes");  
  suif.require_module("dcalcfgnodes"); // for the graphnodes
  suif.require_module("transforms");
  suif.require_module("suifcloning");
  suif.require_module("usefulpasses");
  string suiffile(argv[optind]);
  
  ifstream f(suiffile.c_str(),ios::in);

  if(!f.is_open()) {
    ERROR("cannot open " + suiffile);
  }

  f.close();

  // ok, so it does exist ... 
  suif.read(suiffile.c_str());

  try {
    FileSetBlock *fsb(suif.get_file_set_block());
    unsigned int max_struct_size(0);
    // first, compute the maximum struct size

    SymbolTable *table = fsb->get_external_symbol_table();      
    for(Iter<SymbolTableObject*> i=table->get_symbol_table_object_iterator();
	i.is_valid();i.next()) {
      if(is_a<StructType>(i.current())) {
	StructType *s_type(to<StructType>(i.current()));
	GroupSymbolTable *ptr(s_type->get_group_symbol_table());
	unsigned int tmp(ptr->get_symbol_table_object_count());
	max_struct_size = max(tmp,max_struct_size);
      }
    }

    if(mss_override) {
      max_struct_size = mss_value;
    }

    generator gen(allocators,max_struct_size);
    vector<constraint> cons;
    set<var_type> types;

    // now, process global variables
    for(Iter<SymbolTableObject*> i=table->get_symbol_table_object_iterator();
	i.is_valid();i.next()) {
      if(is_a<VariableSymbol>(i.current())) {
	VariableSymbol *sym(to<VariableSymbol>(i.current()));
	if(is_pointer_type(sym->get_type()) || 
	   contains_pointer_type(sym->get_type()) ||
	   sym->get_is_address_taken()) {
	  
	  add_type(var_type(gen.variable_name(sym),
			    gen.get_type(sym->get_type()),gen.get_rep(sym->get_type())),types);
	}
      } 
    }
    
    if(merge_string_constants) {
      add_type(var_type("MERGED_STRING_CONSTANTS",1,1),types);
    }

    for(Iter<FileBlock*> j=fsb->get_file_block_iterator();j.is_valid();j.next()) {
      gen.do_fileblock(j.current(),cons,types);
    }

    if(gen.ebegin() != gen.eend()) {
      unsigned int count(0);
      for(generator::error_iterator i(gen.ebegin());
	  i!=gen.eend();++i,++count) {
	cerr << *i << endl;
      }
      cerr << endl << count << " errors." << endl;
      exit(1);
    } else {
      set<string> addr_taken;
      for(vector<constraint>::iterator i(cons.begin());
	  i!=cons.end();++i) {
	if(i->rhs.type == ADDRESSOF) {
	  addr_taken.insert(i->rhs.var);
	}
      }
     
      deque<var_type> sorted_types;
      //      sorted_types.reserve(types.size());
      
      for(set<var_type>::const_iterator i(types.begin());
	  i!=types.end();++i) {
	if(addr_taken.find(i->var) != addr_taken.end()) {
	  sorted_types.push_front(*i);
	} else {
	  sorted_types.push_back(*i);
	}
      }

      // NULL and UNKNOWN _must_ come first

      sorted_types.push_front(var_type("NULL",1,1));
      sorted_types.push_front(var_type("UNKNOWN",1,1));
           
      for(deque<var_type>::const_iterator i(sorted_types.begin());
	  i!=sorted_types.end();++i) {
	cout << (*i).var << " : " << (*i).type << " (" << (*i).rep << ");" << endl;
      }
      
      // crude algorithm for removing duplicates
      vector<constraint>::iterator end(cons.end());
      for(vector<constraint>::iterator i(cons.begin());i!=end;++i) {
	end=remove(i+1,end,*i);
      }
      cons.erase(end,cons.end());
      
      for(vector<constraint>::const_iterator i(cons.begin());i!=cons.end();++i) {
	cout << i->str() << endl;
      }
    }
  } catch(runtime_error &e) {
    cerr << e.what() << endl;
    exit(1);
  } catch(...) {
    cerr << "internal error" << endl;
    exit(1);
  }
}





