#include <algorithm>
#include "bitset.hpp"

using namespace std;

// I generally find that setting this to 1
// gives the optimal solution. I believe it's because
// the entire lookup table fits into cache in this case.
#define USE_CHAR_TABLE 1

bitset::bitset(void) 
  : _bits(NULL), _size(0), _bits_size(0) {
  _bits = new uint32_t[1];
  *_bits = 1U;
}
  
bitset::bitset(bitset const &src) { copy(src); }
bitset::~bitset(void) { destroy(); }

bitset const &bitset::operator=(bitset const &src) {
  if(this != &src) {
    destroy();
    copy(src);
  }
  
  return *this;
}

bool bitset::operator==(bitset const &src) const {
  if(src._size != _size || src._bits_size != _bits_size) {
    return false;
  } else {
    uint32_t *p(_bits);
    uint32_t *q(src._bits);
    for(unsigned int i=0;i!=_bits_size;++i,++p,++q) {
      if(*p != *q) { return false; }
    }
  }
  return true;
}

bitset::iterator bitset::begin(void) { return bitset_iterator(_bits,1U,0); }
bitset::const_iterator bitset::begin(void) const { return bitset_iterator(_bits,1U,0); }
bitset::iterator bitset::end(void) { return bitset_iterator(_bits+_bits_size,1U,_bits_size << 5U); }
bitset::const_iterator bitset::end(void) const { return bitset_iterator(_bits+_bits_size,1U,_bits_size << 5U); }

std::pair<bitset::iterator, bool> bitset::insert(const value_type &x) {
  resize(x); // always
  assert((_bits_size*32) > x);
  unsigned int offset(x >> 5U);
  unsigned int boffset(x & 31U);
  uint32_t mask(1U << boffset);
  
  bool r((_bits[offset] & mask) == 0);
  if(r) { _size++; }
  _bits[offset] |= mask;  
  assert((_bits[offset]&mask) != 0);
  return std::make_pair(iterator(_bits+offset,mask,x),r);
}

bitset::iterator bitset::insert(iterator start, const value_type &x) {
  resize(x); // always
  assert((_bits_size*32) > x);
  unsigned int offset(x >> 5U);
  unsigned int boffset(x & 31U);
  uint32_t mask(1U << boffset);
  
  if((_bits[offset] & mask) == 0) { ++_size; }
  _bits[offset] |= mask;
  return iterator(_bits+offset,mask,x);
}

void bitset::erase(iterator pos) {
  //  assert(*(pos.base()) == true);
  --_size;
  pos.set(false);
}

bitset::size_type bitset::erase(value_type const &x) {
  unsigned int offset(x >> 5U);    
  uint32_t mask(1U << (x & 31U));
  
  if(offset >= _bits_size || (_bits[offset] & mask) == 0) { 
    return 0;
  }
  --_size;
  _bits[offset] &= ~mask;
  return 1;
}

void bitset::erase(iterator beg, iterator end) {
  while(beg != end) {
    erase(*beg);beg++;
  }
}

void bitset::clear(void) {
  _size=0;
  memset(_bits,0,_bits_size* sizeof(uint32_t));
}

void bitset::swap(bitset &src) {
  std::swap(_size,src._size);
  std::swap(_bits_size,src._bits_size);
  std::swap(_bits,src._bits);
}

bitset::iterator bitset::find(value_type const &x) {
  unsigned int offset(x >> 5U);
  unsigned int boffset(x & 31U);
  uint32_t mask(1U << boffset);
  
  if(offset >= _bits_size || (_bits[offset] & mask) == 0) { 
    return end();
  }
  return iterator(_bits+offset,mask,x);
}

bitset::const_iterator bitset::find(value_type const &x) const {
  unsigned int offset(x >> 5U);
  unsigned int boffset(x & 31U);
  uint32_t mask(1U << boffset);
  
  if(offset >= _bits_size || (_bits[offset] & mask) == 0) { 
    return end();
  }
  return const_iterator(_bits+offset,mask,x);
}

bitset::iterator bitset::lower_bound(value_type const &x) {
  unsigned int offset(x >> 5U);
  unsigned int boffset(x & 31U);
  uint32_t mask(1U << boffset);
  
  if(offset >= _bits_size) { return end(); }
  // note, this will auto increment into the
  // correct place...
  return iterator(_bits+offset,mask,x);
}

bitset::iterator bitset::upper_bound(value_type const &x) {
  return lower_bound(x);
}

void bitset::destroy(void) { delete [] _bits; }

void bitset::copy(bitset const &src) {
  _size = src._size;
  _bits_size = src._bits_size;
  _bits = new uint32_t[_bits_size+1];
  memcpy(_bits,src._bits,(_bits_size+1) * sizeof(uint32_t));
}

void bitset::resize(unsigned int setsize) {
  // resize so as to include atleast
  // setsize elements
  unsigned int new_bits_size = (setsize >> 5U) + 1;
  if(_bits_size >= new_bits_size) { return; }
  //  new_bits_size *= 2; 
  // plus one here for the terminator
  uint32_t *new_bits(new uint32_t[new_bits_size+1]);    
  memcpy(new_bits,_bits,_bits_size* sizeof(uint32_t));
  memset(new_bits+_bits_size,0,(new_bits_size-_bits_size) * sizeof(uint32_t));
  new_bits[new_bits_size] = 1U; // add terminator.
  delete [] _bits;
  _bits = new_bits;
  _bits_size = new_bits_size;
}

unsigned char bitset_bit_count_table[65536] = {
  // don't worry! I auto generated this here table ...
  0x00,0x01,0x01,0x02,0x01,0x02,0x02,0x03,0x01,0x02,0x02,0x03,0x02,0x03,0x03,0x04,0x01,0x02,0x02,0x03,
  0x02,0x03,0x03,0x04,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x01,0x02,0x02,0x03,0x02,0x03,0x03,0x04,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x01,0x02,0x02,0x03,0x02,0x03,0x03,0x04,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x01,0x02,0x02,0x03,0x02,0x03,0x03,0x04,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x01,0x02,0x02,0x03,
  0x02,0x03,0x03,0x04,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x01,0x02,0x02,0x03,0x02,0x03,0x03,0x04,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x01,0x02,0x02,0x03,0x02,0x03,0x03,0x04,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x01,0x02,0x02,0x03,0x02,0x03,0x03,0x04,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x01,0x02,0x02,0x03,
  0x02,0x03,0x03,0x04,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x01,0x02,0x02,0x03,0x02,0x03,0x03,0x04,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x01,0x02,0x02,0x03,0x02,0x03,0x03,0x04,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x0c,0x0d,0x0d,0x0e,0x0d,0x0e,0x0e,0x0f,0x01,0x02,0x02,0x03,0x02,0x03,0x03,0x04,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x02,0x03,0x03,0x04,
  0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x0c,0x0d,0x0d,0x0e,0x0d,0x0e,0x0e,0x0f,0x02,0x03,0x03,0x04,0x03,0x04,0x04,0x05,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x03,0x04,0x04,0x05,
  0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0c,0x0d,0x0d,0x0e,
  0x0d,0x0e,0x0e,0x0f,0x03,0x04,0x04,0x05,0x04,0x05,0x05,0x06,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x04,0x05,0x05,0x06,
  0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0c,0x0d,0x0d,0x0e,0x0d,0x0e,0x0e,0x0f,
  0x04,0x05,0x05,0x06,0x05,0x06,0x06,0x07,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x05,0x06,0x06,0x07,
  0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x0c,0x0d,0x0d,0x0e,0x0d,0x0e,0x0e,0x0f,0x05,0x06,0x06,0x07,0x06,0x07,0x07,0x08,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x06,0x07,0x07,0x08,
  0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x0c,0x0d,0x0d,0x0e,0x0d,0x0e,0x0e,0x0f,0x06,0x07,0x07,0x08,0x07,0x08,0x08,0x09,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x07,0x08,0x08,0x09,
  0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0c,0x0d,0x0d,0x0e,
  0x0d,0x0e,0x0e,0x0f,0x07,0x08,0x08,0x09,0x08,0x09,0x09,0x0a,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x08,0x09,0x09,0x0a,
  0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,
  0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0c,0x0d,0x0d,0x0e,0x0d,0x0e,0x0e,0x0f,
  0x08,0x09,0x09,0x0a,0x09,0x0a,0x0a,0x0b,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x09,0x0a,0x0a,0x0b,
  0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x0c,0x0d,0x0d,0x0e,0x0d,0x0e,0x0e,0x0f,0x09,0x0a,0x0a,0x0b,0x0a,0x0b,0x0b,0x0c,0x0a,0x0b,0x0b,0x0c,
  0x0b,0x0c,0x0c,0x0d,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,
  0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0b,0x0c,0x0c,0x0d,
  0x0c,0x0d,0x0d,0x0e,0x0c,0x0d,0x0d,0x0e,0x0d,0x0e,0x0e,0x0f,0x0a,0x0b,0x0b,0x0c,0x0b,0x0c,0x0c,0x0d,
  0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0c,0x0d,0x0d,0x0e,
  0x0d,0x0e,0x0e,0x0f,0x0b,0x0c,0x0c,0x0d,0x0c,0x0d,0x0d,0x0e,0x0c,0x0d,0x0d,0x0e,0x0d,0x0e,0x0e,0x0f,
  0x0c,0x0d,0x0d,0x0e,0x0d,0x0e,0x0e,0x0f,0x0d,0x0e,0x0e,0x0f,0x0e,0x0f,0x0f,0x010
};

// Now, specialise my set_union function
// for the case when we are merging
// two bitsets
bool set_union(bitset &dest, bitset const &src) {
  unsigned int size(0);
  dest.resize(src._bits_size << 5U);
  uint32_t *dptr(dest._bits);
  uint32_t *sptr(src._bits);
  assert(src._bits_size <= dest._bits_size);
  unsigned int i=0;
  for(;i!=src._bits_size;++i,++dptr,++sptr) {
    *dptr |= *sptr;
#if USE_CHAR_TABLE
    size += bitset_bit_count_table[*((unsigned char*)dptr)];
    size += bitset_bit_count_table[*(((unsigned char*)dptr)+1)];
    size += bitset_bit_count_table[*(((unsigned char*)dptr)+2)];
    size += bitset_bit_count_table[*(((unsigned char*)dptr)+3)];
#else
    size += bitset_bit_count_table[*((uint16_t*)dptr)];
    size += bitset_bit_count_table[*(((uint16_t*)dptr)+1)];
#endif
  }
  for(;i!=dest._bits_size;++i,++dptr) {
#if USE_CHAR_TABLE
    size += bitset_bit_count_table[*((unsigned char*)dptr)];
    size += bitset_bit_count_table[*(((unsigned char*)dptr)+1)];
    size += bitset_bit_count_table[*(((unsigned char*)dptr)+2)];
    size += bitset_bit_count_table[*(((unsigned char*)dptr)+3)];
#else
    size += bitset_bit_count_table[*((uint16_t*)dptr)];
    size += bitset_bit_count_table[*(((uint16_t*)dptr)+1)];
#endif
  }

  bool r(size != dest._size);
  dest._size = size;
  return r;
}







